/*
 *  1541fs.cpp - 1541 emulation in host file system
 *
 *  Frodo (C) 1996 Christian Bauer

 *
 * Notes:
 * ------
 *
 *  - If the directory is opened (file name "$"), a temporary file
 *    with the structure of a 1541 directory file is created and
 *    opened. It can then be accessed in the same way as all other
 *    files.
 *
 * Incompatibilities:
 * ------------------
 *
 *  - No "raw" directory reading
 *  - No relative files
 *  - Only "I" and "UJ" commands implemented
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/stat.h>

#include "1541fs.h"
#include "IEC.h"
#include "main.h"
#include "Prefs.h"


// Access modes
enum {
	FMODE_READ, FMODE_WRITE, FMODE_APPEND
};

// File types
enum {
	FTYPE_PRG, FTYPE_SEQ
};

// Prototypes
CBOOL match(char *p, char *n);


/*
 *  Constructor: Prepare emulation, create directory object
 *   and set its ref
 */

FSDrive::FSDrive(IEC *iec, char *path) : Drive(iec)
{
	DIR *dir;

	Ready = FALSE;
	strcpy(dir_path, path);

	if (dir = opendir(dir_path)) {
		closedir(dir);
		strncpy(dir_title, dir_path, 16);

		for (int i=0; i<16; i++)
			file[i] = NULL;

		cmd_len = 0;	
		Ready = TRUE;
	}
}


/*
 *  Destructor
 */

FSDrive::~FSDrive()
{
	if (Ready) {
		close_all_channels();
		Ready = FALSE;
	}
}


/*
 *  Open channel
 */

UBYTE FSDrive::Open(int channel, char *filename)
{
	set_error(ERR_OK);

	// Channel 15: Execute file name as command
	if (channel == 15) {
		execute_command(filename);
		return ST_OK;
	}

	// Close previous file if still open
	if (file[channel]) {
		fclose(file[channel]);
		file[channel] = NULL;
	}

	if (filename[0] == '$')
		return open_directory(channel, filename+1);

	if (filename[0] == '#') {
		set_error(ERR_NOCHANNEL);
		return ST_OK;
	}

	return open_file(channel, filename);
}


/*
 *  Open file
 */

UBYTE FSDrive::open_file(int channel, char *filename)
{
	char plainname[NAMEBUF_LENGTH];
	int filemode = FMODE_READ;
	int filetype = FTYPE_PRG;
	CBOOL wildflag = FALSE;
	char *mode;
	
	convert_filename(filename, plainname, &filemode, &filetype, &wildflag);

	// Channel 0 is READ PRG, channel 1 is WRITE PRG
	if (!channel) {
		filemode = FMODE_READ;
		filetype = FTYPE_PRG;
	}
	if (channel == 1) {
		filemode = FMODE_WRITE;
		filetype = FTYPE_PRG;
	}

	// Wildcards are only allowed on reading
	if (wildflag) {
		if (filemode != FMODE_READ) {
			set_error(ERR_SYNTAX33);
			return ST_OK;
		}
		find_first_file(plainname);
	}

	// Select fopen() mode according to file mode
	switch (filemode) {
		case FMODE_READ:
			mode = "rb";
			break;
		case FMODE_WRITE:
			mode = "wb";
			break;
		case FMODE_APPEND:
			mode = "ab";
			break;
	}

	// Open file
	chdir(dir_path);
	if (file[channel] = fopen(plainname, mode)) {
		if (filemode == FMODE_READ)	// Read and buffer first byte
			read_char = fgetc(file[channel]);
	} else
		set_error(ERR_FILENOTFOUND);
	chdir(AppDirPath);

	return ST_OK;
}


/*
 *  Analyze file name, get access mode and type
 */

void FSDrive::convert_filename(char *srcname, char *destname, int *filemode, int *filetype, CBOOL *wildflag)
{
	char *p, *q;
	int i;

	// Search for ':', p points to first character after ':'
	if (p = strchr(srcname, ':'))
		p++;
	else
		p = srcname;

	// Convert char set of the remaining string -> destname
	q = destname;
	for (i=0; i<NAMEBUF_LENGTH && (*q++ = conv_from_64(*p++)); i++) ;

	// Look for mode parameters seperated by ','
	p = destname;
	while (p = strchr(p, ',')) {

		// Cut string after the first ','
		*p++ = 0;

		switch (*p) {
			case 'p':
				*filetype = FTYPE_PRG;
				break;
			case 's':
				*filetype = FTYPE_SEQ;
				break;
			case 'r':
				*filemode = FMODE_READ;
				break;
			case 'w':
				*filemode = FMODE_WRITE;
				break;
			case 'a':
				*filemode = FMODE_APPEND;
				break;
		}
	}

	// Search for wildcards
	*wildflag = (strchr(destname, '?') != NULL) || (strchr(destname, '*') != NULL);
}


/*
 *  Find first file matching wildcard pattern and get its real name
 */

// Return TRUE if name 'n' matches pattern 'p'
CBOOL match(char *p, char *n)
{
	if (!*p)		// Null pattern matches everything
		return TRUE;

	do {
		if (*p == '*')	// Wildcard '*' matches all following characters
			return TRUE;
		if ((*p != *n) && (*p != '?'))	// Wildcard '?' matches single character
			return FALSE;
		p++; n++;
	} while (*p);

	return TRUE;
}

void FSDrive::find_first_file(char *name)
{
	DIR *dir;
	struct dirent *de;

	// Open directory for reading and skip '.' and '..'
	if (!(dir = opendir(dir_path)))
		return;
	de = readdir(dir);
	while(de && (0 == strcmp(".", de->d_name) || 0 == strcmp("..", de->d_name))) 
		de = readdir(dir);

	while (de) {

		// Match found? Then copy real file name
		if (match(name, de->d_name)) {
			strncpy(name, de->d_name, NAMEBUF_LENGTH);
			closedir(dir);
			return;
		}

		// Get next directory entry
		de = readdir(dir);
	}

	closedir(dir);
}


/*
 *  Open directory, create temporary file
 */

UBYTE FSDrive::open_directory(int channel, char *filename)
{
	char buf[] = "\001\004\001\001\0\0\022\042                \042 00 2A";
	char str[NAMEBUF_LENGTH];
	char pattern[NAMEBUF_LENGTH];
	char *p, *q;
	int i;
	int filemode;
	int filetype;
	CBOOL wildflag;
	DIR *dir;
	struct dirent *de;
	struct stat statbuf;

	// Convert filename ('$' already stripped), filemode/type are ignored
	convert_filename(filename, pattern, &filemode, &filetype, &wildflag);

	// Open directory for reading and skip '.' and '..'
	if (!(dir = opendir(dir_path)))
		return ST_OK;
	de = readdir(dir);
	while(de && (0 == strcmp(".", de->d_name) || 0 == strcmp("..", de->d_name))) 
		de = readdir(dir);

	// Create temporary file
	if (!(file[channel] = fopen("/tmp/FrodoDirTmpFile", "wb"))) {
		closedir(dir);
		return ST_OK;
	}

	// Create directory title
	p = &buf[8];
	for (i=0; i<16 && dir_title[i]; i++)
		*p++ = conv_to_64(dir_title[i]);
	fwrite(buf, 1, 32, file[channel]);

	// Create and write one line for every directory entry
	while (de) {

		// Include only files matching the pattern
		if (match(pattern, de->d_name)) {

			// Get file statistics
			chdir(dir_path);
			stat(de->d_name, &statbuf);
			chdir(AppDirPath);

			// Clear line with spaces and terminate with null byte
			memset(buf, ' ', 31);
			buf[31] = 0;

			p = buf;
			*p++ = 0x01;	// Dummy line link
			*p++ = 0x01;

			// Calculate size in blocks (254 bytes each)
			i = (statbuf.st_size + 254) / 254;
			*p++ = i & 0xff;
			*p++ = (i >> 8) & 0xff;

			p++;
			if (i < 10) p++;	// Less than 10: add one space
			if (i < 100) p++;	// Less than 100: add another space

			// Convert and insert file name
			strcpy(str, de->d_name);
			*p++ = '\"';
			q = p;
			for (i=0; i<16 && str[i]; i++)
				if (str[i])
					*q++ = conv_to_64(str[i]);
			*q++ = '\"';
			p += 18;

			// File type
			if (S_ISDIR(statbuf.st_mode)) {
				*p++ = 'D';
				*p++ = 'I';
				*p++ = 'R';
			} else {
				*p++ = 'P';
				*p++ = 'R';
				*p++ = 'G';
			}

			// Write line
			fwrite(buf, 1, 32, file[channel]);
		}

		// Get next directory entry
		de = readdir(dir);
	}

	// Final line
	fwrite("\001\001\0\0BLOCKS FREE.             \0\0", 1, 32, file[channel]);
	fclose(file[channel]);

	// Open file for reading and read first character
	if (file[channel] = fopen("/tmp/FrodoDirTmpFile", "rb"))
		read_char = fgetc(file[channel]);

	// Close directory
	closedir(dir);

	return ST_OK;
}


/*
 *  Close channel
 */

UBYTE FSDrive::Close(int channel)
{
	if (channel == 15) {
		close_all_channels();
		return ST_OK;
	}

	if (file[channel]) {
		fclose(file[channel]);
		file[channel] = NULL;
	}

	return ST_OK;
}


/*
 *  Close all channels
 */

void FSDrive::close_all_channels(void)
{
	for (int i=0; i<15; i++)
		Close(i);

	cmd_len = 0;
}


/*
 *  Read from channel
 */

UBYTE FSDrive::Read(int channel, UBYTE *byte)
{
	int c;

	// Channel 15: Error channel
	if (channel == 15) {
		*byte = *error_ptr++;

		if (*byte != '\r')
			return ST_OK;
		else {	// End of message
			set_error(ERR_OK);
			return ST_EOF;
		}
	}

	if (!file[channel]) return ST_READ_TIMEOUT;

	// Get char from buffer and read next
	*byte = read_char;
	c = fgetc(file[channel]);
	if (c == EOF)
		return ST_EOF;
	else {
		read_char = c;
		return ST_OK;
	}
}


/*
 *  Write to channel
 */

UBYTE FSDrive::Write(int channel, UBYTE byte, CBOOL eoi)
{
	// Channel 15: Collect chars and execute command on EOI
	if (channel == 15) {
		if (cmd_len >= 40)
			return ST_TIMEOUT;
		
		cmd_buffer[cmd_len++] = byte;

		if (eoi) {
			cmd_buffer[cmd_len] = 0;
			cmd_len = 0;
			execute_command(cmd_buffer);
		}
		return ST_OK;
	}

	if (!file[channel]) {
		set_error(ERR_FILENOTOPEN);
		return ST_TIMEOUT;
	}

	if (fputc(byte, file[channel]) == EOF) {
		set_error(ERR_WRITEERROR);
		return ST_TIMEOUT;
	}

	return ST_OK;
}


/*
 *  Execute command string
 */

void FSDrive::execute_command(char *command)
{
	switch (command[0]) {
		case 'I':
			close_all_channels();
			set_error(ERR_OK);
			break;

		case 'U':
			if ((command[1] & 0x0f) == 0x0a) {
				close_all_channels();
				set_error(ERR_STARTUP);
			} else
				set_error(ERR_SYNTAX30);
			break;

		default:
			set_error(ERR_SYNTAX30);
	}
}


/*
 *  Conversion PETSCII->ASCII
 */

char FSDrive::conv_from_64(char c)
{
  if ((c >= 'A') && (c <= 'Z') || (c >= 'a') && (c <= 'z'))
    return c ^ 0x20;
  if ((c == '/') && ThePrefs.MapSlash)
    return '\\';
  return c;
}


/*
 *  Conversion ASCII->PETSCII
 */

char FSDrive::conv_to_64(char c)
{
  if ((c >= 'A') && (c <= 'Z') || (c >= 'a') && (c <= 'z'))
    return c ^ 0x20;
  if ((c == '\\') && ThePrefs.MapSlash)
    return '/';
  return c;
}
