/*
 *  1541job.cpp - Emulation of 1541 job loop
 *
 *  Frodo (C) 1996 Christian Bauer
 *

 *
 * Notes:
 * ------
 *
 *  - This is only used for processor-level 1541 emulation.
 *    It simulates the 1541 interrupt job loop that performs
 *    low-level disk accesses.
 *  - The preferences settings for drive 8 are used to
 *    specify the .d64 file
 */

#include <stdio.h>
#include <string.h>

#include "1541job.h"
#include "CPU.h"
#include "Prefs.h"


// Number of tracks
const int NUM_TRACKS = 35;

// Job return codes
const int RET_OK = 1;				// No error
const int RET_NOT_FOUND = 2;		// Block not found
const int RET_VERIFY_ERROR = 7;		// Verify error
const int RET_NOT_READY = 15;		// Drive not ready


/*
 *  Constructor: Open .d64 file if processor-level 1541
 *   emulation is enabled
 */

Job1541::Job1541(MOS6502_1541 *cpu)
{
	the_cpu = cpu;
	ram = cpu->RAM;
	the_file = NULL;

	if (ThePrefs.Emul1541Proc)
		open_d64_file(ThePrefs.DrivePath[0]);
}


/*
 *  Destructor: Close .d64 file
 */

Job1541::~Job1541()
{
	close_d64_file();
}


/*
 *  Preferences may have changed
 */

void Job1541::NewPrefs(Prefs *prefs)
{
	// 1541 emulation turned off?
	if (!prefs->Emul1541Proc)
		close_d64_file();

	// 1541 emulation turned on?
	else if (!ThePrefs.Emul1541Proc && prefs->Emul1541Proc)
		open_d64_file(prefs->DrivePath[0]);

	// .d64 file name changed?
	else if (strcmp(ThePrefs.DrivePath[0], prefs->DrivePath[0])) {
		close_d64_file();
		open_d64_file(prefs->DrivePath[0]);
	}
}


/*
 *  Open .d64 file
 */

void Job1541::open_d64_file(char *filepath)
{
	ULONG magic;
	UBYTE bam[256];

	if (the_file = fopen(filepath, "rb+")) {

		// Check length
		fseek(the_file, 0, SEEK_END);
		if (ftell(the_file) < 174848) {
			fclose(the_file);
			the_file = NULL;
			return;
		}

		// x64 image?
		fseek(the_file, 0, SEEK_SET);
		fread(&magic, 4, 1, the_file);
		image_header = (magic == 0x43154164 ? 64 : 0);

		// Read BAM and get ID
		read_sector(18, 0, bam);
		id1 = bam[162];
		id2 = bam[163];
	}
}


/*
 *  Close .d64 file
 */

void Job1541::close_d64_file(void)
{
	if (the_file) {
		fclose(the_file);
		the_file = NULL;
	}
}


/*
 *  Job loop: Check and execute job codes. Only called if
 *   processor-level 1541 emulation is enabled.
 */

void Job1541::JobLoop(void)
{
	// Check all 5 job buffers for valid jobs and execute them
	if (the_cpu->InterruptEnabled()) {
		for (int i=0; i<5; i++)
			if (ram[i] & 0x80)
				ram[i] = execute_job(ram + ((i+3) << 8), ram[i], ram[i*2+6], ram[i*2+7]);
	}
}


/*
 *  Execute job, return status code
 */

UBYTE Job1541::execute_job(UBYTE *buffer, int jobcode, int track, int sector)
{
	// Return "Drive Not Ready" if .d64 not open or job for drive 1
	if ((the_file == NULL) || (jobcode & 1))
		return RET_NOT_READY;

	// Store block header in RAM
	ram[0x16] = id1;
	ram[0x17] = id2;
	ram[0x18] = track;
	ram[0x19] = sector;

	switch (jobcode & 0x78) {

		case 0x00:	// Read sector
			return read_sector(track, sector, buffer) ? RET_OK : RET_NOT_FOUND;

		case 0x10:	// Write sector
			return write_sector(track, sector, buffer) ? RET_OK : RET_NOT_FOUND;

		case 0x20:{	// Verify sector
			UBYTE vbuf[256];
			if (read_sector(track, sector, vbuf)) {
				for (int i=0; i<0x100; i++)
					if (buffer[i] != vbuf[i]) return RET_VERIFY_ERROR;
				return RET_OK;
			} else
				return RET_NOT_FOUND;
		}
		case 0x30:	// Read sector header
			return (offset_from_ts(track, sector) >= 0) ? RET_OK : RET_NOT_FOUND;

		case 0x40:	// Realign head
			return RET_OK;

		case 0x50:	// Execute program
		case 0x60:
		default:
			printf("Job code %02x, track %d, sector %d\n", jobcode, track, sector);
			return RET_NOT_READY;
	};
	return RET_OK;
}


/*
 *  Read sector (256 bytes)
 *  TRUE: success, FALSE: error
 */

CBOOL Job1541::read_sector(int track, int sector, UBYTE *buffer)
{
	int offset;

	// Convert track/sector to byte offset in file
	if ((offset = offset_from_ts(track, sector)) < 0)
		return FALSE;

	fseek(the_file, offset + image_header, SEEK_SET);
	fread(buffer, 256, 1, the_file);
	return TRUE;
}


/*
 *  Write sector (256 bytes)
 *  TRUE: success, FALSE: error
 */

CBOOL Job1541::write_sector(int track, int sector, UBYTE *buffer)
{
	int offset;

	// Convert track/sector to byte offset in file
	if ((offset = offset_from_ts(track, sector)) < 0)
		return FALSE;

	fseek(the_file, offset + image_header, SEEK_SET);
	fwrite(buffer, 256, 1, the_file);
	return TRUE;
}


/*
 *  Convert track/sector to offset
 */

const int num_sectors[36] = {
	0,
	21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,
	19,19,19,19,19,19,19,
	18,18,18,18,18,18,
	17,17,17,17,17
};

const int sector_offset[36] = {
	0,
	0,21,42,63,84,105,126,147,168,189,210,231,252,273,294,315,336,
	357,376,395,414,433,452,471,
	490,508,526,544,562,580,
	598,615,632,649,666
};

int Job1541::offset_from_ts(int track, int sector)
{
	if ((track < 1) || (track > NUM_TRACKS) ||
		(sector < 0) || (sector >= num_sectors[track]))
		return -1;

	return (sector_offset[track] + sector) << 8;
}
