/*
 *  C64.cpp - Put the pieces together
 *
 *  Frodo (C) 1996 Christian Bauer
 */

#include <stdlib.h>
#include <string.h>

#include "C64.h"
#include "CPU.h"
#include "VIC.h"
#include "SID.h"
#include "CIA.h"
#include "IEC.h"
#include "1541job.h"
#include "Display.h"
#include "Prefs.h"


/*
 *  Constructor: Allocate objects and memory
 */

C64::C64()
{
	int i,j;
	UBYTE *p;

	// Open display
	TheDisplay = new C64Display;

	// Allocate RAM/ROM memory
	RAM = new UBYTE[0x10000];
	Basic = new UBYTE[0x2000];
	Kernal = new UBYTE[0x2000];
	Char = new UBYTE[0x1000];
	Color = new UBYTE[0x0400];
	RAM1541 = new UBYTE[0x0800];
	ROM1541 = new UBYTE[0x4000];

	// Create the chips
	TheCPU = new MOS6510;
	TheCPU->TheC64 = this;

	TheCPU1541 = new MOS6502_1541;
	TheCPU1541->TheC64 = this;

	TheVIC = TheCPU->TheVIC = new MOS6569(this, TheDisplay, TheCPU, RAM, Char, Color);
	TheSID = TheCPU->TheSID = new MOS6581;
	TheCIA1 = TheCPU->TheCIA1 = new MOS6526_1(TheCPU);
	TheCIA2 = TheCPU->TheCIA2 = TheCPU1541->TheCIA2 = new MOS6526_2(TheCPU, TheVIC, TheCPU1541);
	TheIEC = TheCPU->TheIEC = new IEC(TheDisplay);

	// Set memory pointers
	TheCPU->RAM = RAM;
	TheCPU->Basic = Basic;
	TheCPU->Kernal = Kernal;
	TheCPU->Char = Char;
	TheCPU->Color = Color;
	TheCPU1541->RAM = RAM1541;
	TheCPU1541->ROM = ROM1541;
	TheJob1541 = new Job1541(TheCPU1541);	// RAM must be valid

	// Initialize RAM with powerup pattern
	for (i=0, p=RAM; i<512; i++) {
		for (j=0; j<64; j++)
			*p++ = 0;
		for (j=0; j<64; j++)
			*p++ = 0xff;
	}

	// Initialize color RAM with random values
	for (i=0, p=Color; i<1024; i++)
		*p++ = rand() & 0x0f;

	// Clear 1541 RAM
	memset(RAM1541, 0, 0x800);

	// The thread is not yet running
	thread_running = FALSE;
	quit_thyself = FALSE;
	have_a_break = FALSE;

	// Open joystick drivers if required
	joy1fd = joy2fd = -1;
	open_close_joysticks(FALSE, FALSE, ThePrefs.Joystick1On, ThePrefs.Joystick2On);
	joykey = 0xff;

	// System-dependent things
	c64_ctor();
}


/*
 *  Destructor: Delete all objects
 */

C64::~C64()
{
	c64_dtor();

	open_close_joysticks(ThePrefs.Joystick1On, ThePrefs.Joystick2On, FALSE, FALSE);

	if (TheJob1541) delete TheJob1541;
	if (TheIEC) delete TheIEC;
	if (TheCIA2) delete TheCIA2;
	if (TheCIA1) delete TheCIA1;
	if (TheSID) delete TheSID;
	if (TheVIC) delete TheVIC;
	if (TheCPU1541) delete TheCPU1541;
	if (TheCPU) delete TheCPU;
	if (TheDisplay) delete TheDisplay;

	if (RAM) delete[] RAM;
	if (Basic) delete[] Basic;
	if (Kernal) delete[] Kernal;
	if (Char) delete[] Char;
	if (Color) delete[] Color;
	if (RAM1541) delete[] RAM1541;
	if (ROM1541) delete[] ROM1541;
}


/*
 *  Reset C64
 */

void C64::Reset(void)
{
	TheCPU->AsyncReset();
	TheSID->Reset();
	TheCIA1->Reset();
	TheCIA2->Reset();
	TheCPU1541->AsyncReset();
}


/*
 *  NMI C64
 */

void C64::NMI(void)
{
	TheCPU->AsyncNMI();
}


/*
 *  The preferences have changed. prefs is a pointer to the new
 *   preferences, ThePrefs still holds the previous ones
 */

void C64::NewPrefs(Prefs *prefs)
{
	open_close_joysticks(ThePrefs.Joystick1On, ThePrefs.Joystick2On, prefs->Joystick1On, prefs->Joystick2On);
	patch_kernal(prefs->FastReset, prefs->Emul1541Proc);
	TheIEC->NewPrefs(prefs);
	TheJob1541->NewPrefs(prefs);
	TheSID->NewPrefs(prefs);

	// Reset 1541 processor if turned on
	if (!ThePrefs.Emul1541Proc && prefs->Emul1541Proc)
		TheCPU1541->AsyncReset();
}


/*
 *  Patch kernal IEC routines
 */

void C64::patch_kernal(CBOOL fast_reset, CBOOL emul_1541_proc)
{
	if (fast_reset) {
		Kernal[0x1d84] = 0xa0;
		Kernal[0x1d85] = 0x00;
	} else {
		Kernal[0x1d84] = orig_kernal_1d84;
		Kernal[0x1d85] = orig_kernal_1d85;
	}

	if (emul_1541_proc) {
		Kernal[0x0d40] = 0x78;
		Kernal[0x0d41] = 0x20;
		Kernal[0x0d23] = 0x78;
		Kernal[0x0d24] = 0x20;
		Kernal[0x0d36] = 0x78;
		Kernal[0x0d37] = 0x20;
		Kernal[0x0e13] = 0x78;
		Kernal[0x0e14] = 0xa9;
		Kernal[0x0def] = 0x78;
		Kernal[0x0df0] = 0x20;
		Kernal[0x0dbe] = 0xad;
		Kernal[0x0dbf] = 0x00;
		Kernal[0x0dcc] = 0x78;
		Kernal[0x0dcd] = 0x20;
		Kernal[0x0e03] = 0x20;
		Kernal[0x0e04] = 0xbe;
	} else {
		Kernal[0x0d40] = 0xf2;	// IECOut
		Kernal[0x0d41] = 0x00;
		Kernal[0x0d23] = 0xf2;	// IECOutATN
		Kernal[0x0d24] = 0x01;
		Kernal[0x0d36] = 0xf2;	// IECOutSec
		Kernal[0x0d37] = 0x02;
		Kernal[0x0e13] = 0xf2;	// IECIn
		Kernal[0x0e14] = 0x03;
		Kernal[0x0def] = 0xf2;	// IECSetATN
		Kernal[0x0df0] = 0x04;
		Kernal[0x0dbe] = 0xf2;	// IECRelATN
		Kernal[0x0dbf] = 0x05;
		Kernal[0x0dcc] = 0xf2;	// IECTurnaround
		Kernal[0x0dcd] = 0x06;
		Kernal[0x0e03] = 0xf2;	// IECRelease
		Kernal[0x0e04] = 0x07;
	}
}


#ifdef __BEOS__
#include "C64_Be.i"
#endif

#ifdef AMIGA
#include "C64_Amiga.i"
#endif

#ifdef __unix
#include "C64_x.i"
#endif
