/*
 *  C64_Be.i - Put the pieces together, Be specific stuff
 *
 *  Frodo (C) 1996 Christian Bauer
 */

#include <KernelKit.h>
#include <unistd.h>
#include <joystick.h>


/*
 *  Constructor, system-dependent things
 */

void C64::c64_ctor(void)
{
	// Initialize semaphore (initially acquired)
	pause_sem = create_sem(0, "Frodo Pause Semaphore");

	// Preset speedometer start time
	start_time = system_time();
}


/*
 *  Destructor, system-dependent things
 */

void C64::c64_dtor(void)
{
	delete_sem(pause_sem);
}


/*
 *  Start main emulation thread
 */

void C64::Run(void)
{
	// Reset chips
	TheCPU->Reset();
	TheSID->Reset();
	TheCIA1->Reset();
	TheCIA2->Reset();
	TheCPU1541->Reset();

	// Patch kernal IEC routines
	orig_kernal_1d84 = Kernal[0x1d84];
	orig_kernal_1d85 = Kernal[0x1d85];
	patch_kernal(ThePrefs.FastReset, ThePrefs.Emul1541Proc);

	// Start the CPU thread
	the_thread = spawn_thread(thread_invoc, "Frodo 6510", B_REAL_TIME_PRIORITY, this);
	thread_running = TRUE;
	quit_thyself = FALSE;
	have_a_break = FALSE;
	resume_thread(the_thread);
}


/*
 *  Stop main emulation thread
 */

void C64::Quit(void)
{
	long ret;

	// Ask the thread to quit itself if it is running
	if (thread_running) {
		if (have_a_break) Resume();
		quit_thyself = TRUE;
		wait_for_thread(the_thread, &ret);
		thread_running = FALSE;
	}
}


/*
 *  Pause main emulation thread
 */

void C64::Pause(void)
{
	// Ask the thread to pause and wait for acknowledge
	if (thread_running) {
		have_a_break = TRUE;
		acquire_sem(pause_sem);
		TheSID->PauseSound();
	}
}


/*
 *  Resume main emulation thread
 */

void C64::Resume(void)
{
	if (thread_running) {
		have_a_break = FALSE;
		release_sem(pause_sem);
		TheSID->ResumeSound();
	}
}


/*
 *  Vertical blank: Poll keyboard and joysticks, update window
 */

void C64::VBlank(CBOOL draw_frame)
{
	double elapsed_time, speed_index;

	// To avoid deadlocks on quitting
	if (quit_thyself) return;

	// Poll keyboard
	TheDisplay->PollKeyboard(TheCIA1->KeyMatrix, TheCIA1->RevMatrix, &joykey);

	// Poll joysticks
	TheCIA1->Joystick1 = poll_joystick(joy1fd);
	TheCIA1->Joystick2 = poll_joystick(joy2fd);

	if (ThePrefs.JoystickSwap) {
		UBYTE tmp = TheCIA1->Joystick1;
		TheCIA1->Joystick1 = TheCIA1->Joystick2;
		TheCIA1->Joystick2 = tmp;
	}

	// Joystick keyboard emulation
	if (TheDisplay->NumLock())
		TheCIA1->Joystick1 &= joykey;
	else
		TheCIA1->Joystick2 &= joykey;

	// Count TOD clocks
	TheCIA1->CountTOD();
	TheCIA2->CountTOD();

	// Do 1541 job loop
	if (ThePrefs.Emul1541Proc)
		TheJob1541->JobLoop();

	// Update window if needed
	if (draw_frame) {
		TheDisplay->Update();

		// Calculate time between VBlanks, display speedometer
		elapsed_time = system_time() - start_time;
		speed_index = 20000 / (elapsed_time + 1) * ThePrefs.SkipFrames * 100;

		// Limit speed to 100% if desired
		if ((speed_index > 100) && ThePrefs.LimitSpeed) {
			snooze(ThePrefs.SkipFrames * 20000 - elapsed_time);
			speed_index = 100;
        }

		start_time = system_time();

		TheDisplay->Speedometer(speed_index);
	}
}


/*
 *  Open/close joystick drivers given old and new state of
 *  joystick preferences
 */

void C64::open_close_joysticks(CBOOL oldjoy1, CBOOL oldjoy2, CBOOL newjoy1, CBOOL newjoy2)
{
	if (oldjoy1 != newjoy1)
		if (newjoy1)
			joy1fd = open("/dev/joystick_1", O_RDONLY);
		else {
			close(joy1fd);
			joy1fd = -1;
		}

	if (oldjoy2 != newjoy2)
		if (newjoy2)
			joy2fd = open("/dev/joystick_3", O_RDONLY);
		else {
			close(joy2fd);
			joy2fd = -1;
		}
}


/*
 *  Poll joystick port, return CIA mask
 */

UBYTE C64::poll_joystick(int fd)
{
	joystick js;
	UBYTE j = 0xff;

	if (fd >= 0) {
		if (read(fd, &js, sizeof(js)) > 0) {
			if (js.horizontal < 2200)
				j &= 0xf7;				// Right
			else if (js.horizontal > 2800)
				j &= 0xfb;				// Left
			if (js.vertical < 2200)
				j &= 0xfd;				// Down
			else if (js.vertical > 2800)
				j &= 0xfe;				// Up
			if (!js.button1)
				j &= 0xef;				// Button
		}
	}
	return j;
}


/*
 * The emulation's main loop
 */

long C64::thread_invoc(void *obj)
{
	((C64 *)obj)->thread_func();
	return 0;
}

void C64::thread_func(void)
{
	int cycles, cycles_1541;

	while (!quit_thyself) {
		if (have_a_break) {
			release_sem(pause_sem);	// Acknowledge pause
			acquire_sem(pause_sem);	// Wait for resume
		}

		// The order of calls is important here
		cycles = TheVIC->EmulateLine();
		TheCIA1->EmulateLine();
		TheCIA2->EmulateLine();

		if (ThePrefs.Emul1541Proc) {

			// 1541 processor enabled, alternately execute
			//  6502 and 6510 instructions until both have
			//  used up their cycles
			TheCPU1541->CountVIATimers(cycles_1541 = ThePrefs.FloppyCycles);
			while (cycles >= 0 || cycles_1541 >= 0)
				if (cycles > cycles_1541)
					cycles -= TheCPU->EmulateLine(1);
				else
					cycles_1541 -= TheCPU1541->EmulateLine(1);
		} else

			// 1541 processor disabled, only emulate 6510
			TheCPU->EmulateLine(cycles);
	}
}
