/*
 *  C64_x.i - Put the pieces together, X specific stuff
 *
 *  Frodo (C) 1996 Christian Bauer
 */

#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
 
static struct timeval tv_start;

 
/*
 *  Constructor, system-dependent things
 */

void C64::c64_ctor(void)
{
	TheDisplay->TheC64 = this;
	gettimeofday(&tv_start, NULL);
}


/*
 *  Destructor, system-dependent things
 */

void C64::c64_dtor(void)
{
}


/*
 *  Start main emulation thread
 */

void C64::Run(void)
{
	// Reset chips
	TheCPU->Reset();
	TheSID->Reset();
	TheCIA1->Reset();
	TheCIA2->Reset();
	TheCPU1541->Reset();

	// Patch kernal IEC routines
	orig_kernal_1d84 = Kernal[0x1d84];
	orig_kernal_1d85 = Kernal[0x1d85];
	patch_kernal(ThePrefs.FastReset, ThePrefs.Emul1541Proc);

	quit_thyself = FALSE;
#if 0
	// Start the CPU thread
	the_thread = spawn_thread(thread_invoc, "Frodo 6510", B_NORMAL_PRIORITY, this);
	thread_running = TRUE;
	have_a_break = FALSE;
	resume_thread(the_thread);
#endif
	thread_func();
}


/*
 *  Vertical blank: Poll keyboard and joysticks, update window
 */

void C64::VBlank(CBOOL draw_frame)
{
	// Poll keyboard
	TheDisplay->PollKeyboard(TheCIA1->KeyMatrix, TheCIA1->RevMatrix, &joykey);

	// Poll joysticks
	TheCIA1->Joystick1 = poll_joystick(joy1fd);
	TheCIA1->Joystick2 = poll_joystick(joy2fd);

	if (ThePrefs.JoystickSwap) {
		UBYTE tmp = TheCIA1->Joystick1;
		TheCIA1->Joystick1 = TheCIA1->Joystick2;
		TheCIA1->Joystick2 = tmp;
	}

	// Joystick keyboard emulation
	if (TheDisplay->NumLock())
		TheCIA1->Joystick1 &= joykey;
	else
		TheCIA1->Joystick2 &= joykey;
       
	// Count TOD clocks
	TheCIA1->CountTOD();
	TheCIA2->CountTOD();
       
	// Do 1541 job loop
	if (ThePrefs.Emul1541Proc)
		TheJob1541->JobLoop();

	// Update window if needed
	if (draw_frame) {
    	TheDisplay->Update();
		struct timeval tv;
		gettimeofday(&tv, NULL);
		if ((tv.tv_usec -= tv_start.tv_usec) < 0) {
			tv.tv_usec += 1000000;
			tv.tv_sec -= 1;
		}
		tv.tv_sec -= tv_start.tv_sec;
		double elapsed_time = (double)tv.tv_sec * 1000000 + tv.tv_usec;
		double speed_index = 20000 / (elapsed_time + 1) * ThePrefs.SkipFrames * 100;
		static int count = 0;
		if (!(++count % 30))
			/*printf("%f\n", (float)speed_index)*/;
		gettimeofday(&tv_start, NULL);
	}
}


/*
 *  Open/close joystick drivers given old and new state of
 *  joystick preferences
 */

void C64::open_close_joysticks(CBOOL oldjoy1, CBOOL oldjoy2, CBOOL newjoy1, CBOOL newjoy2)
{
	joy1fd = open("/dev/js0", O_RDONLY);
	joy2fd = open("/dev/js1", O_RDONLY);
}


/*
 *  Poll joystick port, return CIA mask
 */

UBYTE C64::poll_joystick(int fd)
{
	UBYTE j = 0xff;
#if 0

	if (fd >= 0) {
		if (read(fd, &js, sizeof(js)) > 0) {
			if (js.horizontal < 2200)
				j &= 0xf7;                              // Right
			else if (js.horizontal > 2800)
				j &= 0xfb;                              // Left
			if (js.vertical < 2200)
				j &= 0xfd;                              // Down
			else if (js.vertical > 2800)
				j &= 0xfe;                              // Up
			if (!js.button1)
				j &= 0xef;                              // Button
		}
	}
#endif
	return j;
}


/*
 * The emulation's main loop
 */

void C64::thread_func(void)
{
	int cycles, cycles_1541;

	while (!quit_thyself) {
		// The order of calls is important here
		cycles = TheVIC->EmulateLine();
		TheCIA1->EmulateLine();
		TheCIA2->EmulateLine();

		if (ThePrefs.Emul1541Proc) {

			// 1541 processor enabled, alternately execute
			//  6502 and 6510 instructions until both have
			//  used up their cycles
			TheCPU1541->CountVIATimers(cycles_1541 = ThePrefs.FloppyCycles);
			while (cycles >= 0 || cycles_1541 >= 0)
				if (cycles > cycles_1541)
					cycles -= TheCPU->EmulateLine(1);
				else
					cycles_1541 -= TheCPU1541->EmulateLine(1);
		} else
			// 1541 processor disabled, only emulate 6510
			TheCPU->EmulateLine(cycles);
	}
}
