/*
 *  CPU.cpp - 6510/6502 emulation (line based)
 *
 *  Frodo (C) 1996 Christian Bauer
 *

 *
 * Notes:
 * ------
 *
 *  - The EmulateLine() function is called for every emulated
 *    raster line. It has a cycle counter that is decremented
 *    by every executed opcode and if the counter goes below
 *    zero, the function returns.
 *  - Memory configurations:
 *     $01  $a000-$bfff  $d000-$dfff  $e000-$ffff
 *     -----------------------------------------------
 *      0       RAM          RAM          RAM
 *      1       RAM       Char ROM        RAM
 *      2       RAM       Char ROM    Kernal ROM
 *      3    Basic ROM    Char ROM    Kernal ROM
 *      4       RAM          RAM          RAM
 *      5       RAM          I/O          RAM
 *      6       RAM          I/O      Kernal ROM
 *      7    Basic ROM       I/O      Kernal ROM
 *  - All memory accesses are done with the read_byte() and
 *    write_byte() functions which also do the memory address
 *    decoding. The read_zp() and write_zp() functions allow
 *    faster access to the zero page, the pop_byte() and
 *    push_byte() macros for the stack.
 *  - If a write occurs to addresses 0 or 1, new_config is
 *    called to check whether the memory configuration has
 *    changed
 *  - The PC is either emulated with a 16 bit address or a
 *    direct memory pointer (for faster access), depending on
 *    the PC_IS_POINTER #define. In the latter case, a second
 *    pointer, pc_base, is kept to allow recalculating the
 *    16 bit 6502 PC if it has to be pushed on the stack.
 *  - The possible interrupt sources are:
 *      INT_VICIRQ: I flag is checked, jump to ($fffe)
 *      INT_CIAIRQ: I flag is checked, jump to ($fffe)
 *      INT_NMI: Jump to ($fffa)
 *      INT_RESET: Jump to ($fffc)
 *  - Interrupts are not checked before every opcode but only
 *    at certain times:
 *      On entering EmulateLine()
 *      On CLI
 *      On PLP if the I flag was cleared
 *      On RTI if the I flag was cleared
 *  - The z_flag variable has the inverse meaning of the
 *    6502 Z flag
 *  - The $f2 opcode that would normally crash the 6502 is
 *    used to implement emulator-specific functions, mainly
 *    those for the IEC routines
 *  - The 1541 6502 emulation also includes a very simple VIA
 *    emulation (enough to make the IEC bus work). It's too
 *    small to move it to a source file of its own.
 *
 * Incompatibilities:
 * ------------------
 *
 *  - If PC_IS_POINTER is set, neither branches accross memory
 *    areas nor jumps to I/O space are possible
 *  - Extra cycles for crossing page boundaries are not
 *    accounted for
 *  - The cassette sense line is always closed
 *  - RRA and ISB have no decimal mode, RRA doesn't affect the
 *    V flag
 *  - A few very esoteric undocumented opcodes are missing
 */

#include <stdio.h>
#include <stdlib.h>

#include "CPU.h"
#include "C64.h"
#include "VIC.h"
#include "SID.h"
#include "CIA.h"
#include "IEC.h"
#include "Display.h"
#include "Version.h"


/*
 *  6502/6510 constructors: Initialize registers
 */

MOS6502::MOS6502()
{
	a = x = y = 0;
	sp = 0xff;
	n_flag = v_flag = d_flag = z_flag = c_flag = FALSE;
	i_flag = TRUE;
}

MOS6510::MOS6510() : MOS6502()
{
	dfff_byte = 0x55;
}

MOS6502_1541::MOS6502_1541() : MOS6502()
{
}


/*
 *  Reset CPU asynchronously
 */

void MOS6502::AsyncReset(void)
{
	Interrupt.intr[INT_RESET] = TRUE;
}


/*
 *  Raise NMI asynchronously
 */

void MOS6502::AsyncNMI(void)
{
	if (!NMIState)
		Interrupt.intr[INT_NMI] = TRUE;
}


/*
 *  Get 6502/1541 register state
 */

void MOS6502_1541::GetState(MOS6502State *state)
{
	state->a = a;
	state->x = x;
	state->y = y;

	state->p = 0x20;
	if (n_flag) state->p |= 0x80;
	if (v_flag) state->p |= 0x40;
	if (d_flag) state->p |= 0x08;
	if (i_flag) state->p |= 0x04;
	if (!z_flag) state->p |= 0x02;
	if (c_flag) state->p |= 0x01;
	
#if PC_IS_POINTER
	state->pc = pc - pc_base;
#else
	state->pc = pc;
#endif
	state->sp = sp | 0x0100;
}


/*
 *  Restore 6502/1541 state
 */

void MOS6502_1541::SetState(MOS6502State *state)
{
	a = state->a;
	x = state->x;
	y = state->y;

	n_flag = state->p & 0x80;
	v_flag = state->p & 0x40;
	d_flag = state->p & 0x08;
	i_flag = state->p & 0x04;
	z_flag = !(state->p & 0x02);
	c_flag = state->p & 0x01;

#if PC_IS_POINTER
	jump(state->pc);
#else
	pc = state->pc;
#endif

	sp = state->sp & 0xff;
}


/*
 *  Get 6510 register state
 */

void MOS6510::GetState(MOS6510State *state)
{
	state->a = a;
	state->x = x;
	state->y = y;

	state->p = 0x20;
	if (n_flag) state->p |= 0x80;
	if (v_flag) state->p |= 0x40;
	if (d_flag) state->p |= 0x08;
	if (i_flag) state->p |= 0x04;
	if (!z_flag) state->p |= 0x02;
	if (c_flag) state->p |= 0x01;
	
	state->ddr = RAM[0];
	state->pr = RAM[1] & 0x3f;

#if PC_IS_POINTER
	state->pc = pc - pc_base;
#else
	state->pc = pc;
#endif
	state->sp = sp | 0x0100;
}


/*
 *  Restore 6510 state
 */

void MOS6510::SetState(MOS6510State *state)
{
	a = state->a;
	x = state->x;
	y = state->y;

	n_flag = state->p & 0x80;
	v_flag = state->p & 0x40;
	d_flag = state->p & 0x08;
	i_flag = state->p & 0x04;
	z_flag = !(state->p & 0x02);
	c_flag = state->p & 0x01;

	RAM[0] = state->ddr;
	RAM[1] = state->pr;
	new_config();

#if PC_IS_POINTER
	jump(state->pc);
#else
	pc = state->pc;
#endif

	sp = state->sp & 0xff;
}


/*
 *  Read byte from 6502/1541 address space (used by SAM)
 */

UBYTE MOS6502_1541::ExtReadByte(UWORD adr)
{
	return read_byte(adr);
}


/*
 *  Write byte to 6502/1541 address space (used by SAM)
 */

void MOS6502_1541::ExtWriteByte(UWORD adr, UBYTE byte)
{
	write_byte(adr, byte);
}


/*
 *  Read byte from 6510 address space (used by SAM)
 */

UBYTE MOS6510::ExtReadByte(UWORD adr)
{
	// Save old memory configuration
	CBOOL bi = basic_in, ki = kernal_in, ci = char_in, ii = io_in;

	// Set new configuration
	basic_in = (ExtConfig & 3) == 3;
	kernal_in = ExtConfig & 2;
	char_in = (ExtConfig & 3) && ~(ExtConfig & 4);
	io_in = (ExtConfig & 3) && (ExtConfig & 4);

	// Read byte
	UBYTE byte = read_byte(adr);

	// Restore old configuration
	basic_in = bi; kernal_in = ki; char_in = ci; io_in = ii;

	return byte;
}


/*
 *  Write byte to 6510 address space (used by SAM)
 */

void MOS6510::ExtWriteByte(UWORD adr, UBYTE byte)
{
	// Save old memory configuration
	CBOOL bi = basic_in, ki = kernal_in, ci = char_in, ii = io_in;

	// Set new configuration
	basic_in = (ExtConfig & 3) == 3;
	kernal_in = ExtConfig & 2;
	char_in = (ExtConfig & 3) && ~(ExtConfig & 4);
	io_in = (ExtConfig & 3) && (ExtConfig & 4);

	// Write byte
	write_byte(adr, byte);

	// Restore old configuration
	basic_in = bi; kernal_in = ki; char_in = ci; io_in = ii;
}


/*
 *  Memory configuration has probably changed (6502/1541, no-op)
 */

void MOS6502_1541::new_config(void)
{
}


/*
 *  Memory configuration has probably changed (6510)
 */

void MOS6510::new_config(void)
{
	UBYTE port = ~RAM[0] | RAM[1];

	basic_in = (port & 3) == 3;
	kernal_in = port & 2;
	char_in = (port & 3) && ~(port & 4);
	io_in = (port & 3) && (port & 4);
}


/*
 *  Read a byte from the CPU's address space (6502/1541)
 */

UBYTE MOS6502_1541::read_byte(UWORD adr)
{
	if (adr >= 0xc000)
		return ROM[adr & 0x3fff];
	if (adr < 0x800)
		return RAM[adr & 0x07ff];

	if ((adr & 0xfff0) == 0x1800)	// VIA 1
		switch (adr & 0xf) {
			case 0:
				return via1_prb & 0x1a
					| ((DATAState && TheCIA2->DATAState) ? 0 : 0x01)
					| ((CLKState && TheCIA2->CLKState) ? 0 : 0x04)
					| (TheCIA2->ATNState ? 0 : 0x80);
			case 4:
				via1_ifr &= 0xbf;
				return via1_t1c;
			case 5:
				return via1_t1c >> 8;
			case 6:
				return via1_t1l;
			case 7:
				return via1_t1l >> 8;
			case 13:
				return via1_ifr;
			default:
				return 0;
		}

	if (adr == 0x1c00)				// VIA 2
		return via2_prb;
	return 0;
}


/*
 *  Read a byte from the CPU's address space (6510)
 */

UBYTE MOS6510::read_byte(UWORD adr)
{
	if (adr < 0xa000)
		return RAM[adr];
	else
		switch (adr >> 12) {
			case 0xa:
			case 0xb:
				if (basic_in)
					return Basic[adr & 0x1fff];
				else
					return RAM[adr];
			case 0xc:
				return RAM[adr];
			case 0xd:
				if (io_in)
					switch ((adr >> 8) & 0x0f) {
						case 0x0:	// VIC
						case 0x1:
						case 0x2:
						case 0x3:
							return TheVIC->ReadRegister(adr & 0x3f);
						case 0x4:	// SID
						case 0x5:
						case 0x6:
						case 0x7:
							return TheSID->ReadRegister(adr & 0x1f);
						case 0x8:	// Color RAM
						case 0x9:
						case 0xa:
						case 0xb:
							return Color[adr & 0x03ff] | rand() & 0xf0;
						case 0xc:	// CIA 1
							return TheCIA1->ReadRegister(adr & 0x0f);
						case 0xd:	// CIA 2
							return TheCIA2->ReadRegister(adr & 0x0f);
						case 0xe:	// Open I/O
						case 0xf:
							if (adr < 0xdfa0)
								return rand() & 0xff;
							else
								return read_emulator_id(adr & 0x7f);
					}
				else if (char_in)
					return Char[adr & 0x0fff];
				else
					return RAM[adr];
			case 0xe:
			case 0xf:
				if (kernal_in)
					return Kernal[adr & 0x1fff];
				else
					return RAM[adr];
		}

	return 0;
}



/*
 *  $dfa0-$dfff: Emulator identification (6510)
 */

const char frodo_id[0x5c] = "FRODO\r(C) 1994-1996 CHRISTIAN BAUER";

UBYTE MOS6510::read_emulator_id(UWORD adr)
{
	switch (adr) {
		case 0x7c:	// $dffc: revision
			return FRODO_REVISION << 4;
		case 0x7d:	// $dffd: version
			return FRODO_VERSION;
		case 0x7e:	// $dffe returns 'F' (Frodo ID)
			return 'F';
		case 0x7f:	// $dfff alternates between $55 and $aa
			dfff_byte = ~dfff_byte;
			return dfff_byte;
		default:
			return frodo_id[adr - 0x20];
	}
}


/*
 *  Read a word (little-endian) from the CPU's address space
 */

inline UWORD MOS6502::read_word(UWORD adr)
{
	return read_byte(adr) | (read_byte(adr+1) << 8);
}


/*
 *  Write a byte to the CPU's address space (6502/1541)
 */

void MOS6502_1541::write_byte(UWORD adr, UBYTE byte)
{
	if (adr < 0x800)
		RAM[adr & 0x07ff] = byte;
	else if ((adr & 0xfff0) == 0x1800)		// VIA 1
		switch (adr & 0xf) {
			case 0:
				via1_prb = byte;
				DATAState = !(via1_prb & 0x02)
					&& ((!(via1_prb & 0x10) && TheCIA2->ATNState) || ((via1_prb & 0x10) && !TheCIA2->ATNState));
				CLKState = !(byte & 0x08);
				break;
			case 4:
			case 6:
				via1_t1l = via1_t1l & 0xff00 | byte;
				break;
			case 5:
				via1_t1l = via1_t1l & 0xff | (byte << 8);
				via1_ifr &= 0xbf;
				via1_t1c = via1_t1l;
				break;
			case 7:
				via1_t1l = via1_t1l & 0xff | (byte << 8);
				break;
			case 13:
				via1_ifr &= ~byte;
				break;
		}
	else if (adr == 0x1c00) {		// Bit 4 of VIA 2 PRB controls LED
		via2_prb = byte;
		TheC64->TheDisplay->UpdateLEDs(byte & 8 ? 1 : 0, 0, 0, 0);
	}
}


/*
 *  Write a byte to the CPU's address space (6510)
 */

void MOS6510::write_byte(UWORD adr, UBYTE byte)
{
	if (adr < 0xd000 || adr >= 0xe000) {
		RAM[adr] = byte;
		if (adr < 2)
			new_config();
	} else
		if (io_in)
			switch ((adr >> 8) & 0x0f) {
				case 0x0:	// VIC
				case 0x1:
				case 0x2:
				case 0x3:
					TheVIC->WriteRegister(adr & 0x3f, byte);
					return;
				case 0x4:	// SID
				case 0x5:
				case 0x6:
				case 0x7:
					TheSID->WriteRegister(adr & 0x1f, byte);
					return;
				case 0x8:	// Color RAM
				case 0x9:
				case 0xa:
				case 0xb:
					Color[adr & 0x03ff] = byte & 0x0f;
					return;
				case 0xc:	// CIA 1
					TheCIA1->WriteRegister(adr & 0x0f, byte);
					return;
				case 0xd:	// CIA 2
					TheCIA2->WriteRegister(adr & 0x0f, byte);
					return;
				case 0xe:	// Open I/O
				case 0xf:
					return;
			}
		else
			RAM[adr] = byte;	
}


/*
 *  Read a byte from the zeropage
 */

inline UBYTE MOS6502::read_zp(UWORD adr)
{
	return RAM[adr];
}


/*
 *  Read a word (little-endian) from the zeropage
 */

inline UWORD MOS6502::read_zp_word(UWORD adr)
{
	return RAM[adr & 0xff] | (RAM[(adr+1) & 0xff] << 8);
}


/*
 *  Write a byte to the zeropage
 */

inline void MOS6502::write_zp(UWORD adr, UBYTE byte)
{
	RAM[adr] = byte;

	// Check if memory configuration may have changed.
	// This is implemented here for speed reasons, new_config() is a no-op on the 6502
	if (adr < 2)
		new_config();
}


#if PC_IS_POINTER
/*
 *  Jump to address (6502/1541)
 */

void MOS6502_1541::jump(UWORD adr)
{
	if (adr >= 0xc000) {
		pc = ROM + (adr & 0x3fff);
		pc_base = ROM - 0xc000;
	} else if (adr < 0x800) {
		pc = RAM + adr;
		pc_base = RAM;
	} else
		illegal_jump(adr);
}

/*
 *  Jump to address (6510)
 */

void MOS6510::jump(UWORD adr)
{
	if (adr < 0xa000) {
		pc = RAM + adr;
		pc_base = RAM;
	} else
		switch (adr >> 12) {
			case 0xa:
			case 0xb:
				if (basic_in) {
					pc = Basic + (adr & 0x1fff);
					pc_base = Basic - 0xa000;
				} else {
					pc = RAM + adr;
					pc_base = RAM;
				}
				break;
			case 0xc:
				pc = RAM + adr;
				pc_base = RAM;
				break;
			case 0xd:
				if (io_in)
					illegal_jump(adr);
				else if (char_in) {
					pc = Char + (adr & 0x0fff);
					pc_base = Char - 0xd000;
				} else {
					pc = RAM + adr;
					pc_base = RAM;
				}
				break;
			case 0xe:
			case 0xf:
				if (kernal_in) {
					pc = Kernal + (adr & 0x1fff);
					pc_base = Kernal - 0xe000;
				} else {
					pc = RAM + adr;
					pc_base = RAM;
				}
				break;
		}
}
#endif


/*
 *  Addressing mode macros
 */

// Read immediate operand
#if PC_IS_POINTER
#define read_byte_imm() (*pc++)
#else
#define read_byte_imm() read_byte(pc++)
#endif

// Read zeropage operand address
#define read_adr_zero() ((UWORD)read_byte_imm())

// Read zeropage operand
#define read_byte_zero() read_zp(read_adr_zero())

// Read zeropage x-indexed operand address
#define read_adr_zero_x() ((read_byte_imm() + x) & 0xff)

// Read zeropage x-indexed operand
#define read_byte_zero_x() read_zp(read_adr_zero_x())

// Read zeropage y-indexed operand address
#define read_adr_zero_y() ((read_byte_imm() + y) & 0xff)

// Read zeropage y-indexed operand
#define read_byte_zero_y() read_zp(read_adr_zero_y())

// Read absolute operand address
#if PC_IS_POINTER
inline UWORD MOS6502::read_adr_abs(void)
{
	UWORD adr;
	
	adr = ((*(pc+1)) << 8) | *pc;
	pc+=2;
	return adr;
}
#else
inline UWORD MOS6502::read_adr_abs(void)
{
	UWORD adr;
	
	adr = read_word(pc);
	pc+=2;
	return adr;
}
#endif

// Read absolute operand
#define read_byte_abs() read_byte(read_adr_abs())

// Read absolute x-indexed operand address
#define read_adr_abs_x() (read_adr_abs() + x)

// Read absolute x-indexed operand
#define read_byte_abs_x() read_byte(read_adr_abs_x())

// Read absolute y-indexed operand address
#define read_adr_abs_y() (read_adr_abs() + y)

// Read absolute x-indexed operand
#define read_byte_abs_y() read_byte(read_adr_abs_y())

// Read indexed indirect operand address
#define read_adr_ind_x() (read_zp_word(read_byte_imm() + x))

// Read indexed indirect operand
#define read_byte_ind_x() read_byte(read_adr_ind_x())

// Read indirect indexed operand address
#define read_adr_ind_y() (read_zp_word(read_byte_imm()) + y)

// Read indirect indexed operand
#define read_byte_ind_y() read_byte(read_adr_ind_y())


/*
 *  Stack macros
 */

// Pop a byte from the stack
#define pop_byte() RAM[(++sp) | 0x0100]

// Pop a word from the stack
inline UWORD MOS6502::pop_word(void)
{
	UBYTE tmp;

	tmp = RAM[(++sp) & 0xff | 0x0100];
	return tmp | (RAM[(++sp) & 0xff | 0x0100] << 8);
}

// Push a byte onto the stack
#define push_byte(byte) (RAM[(sp--) & 0xff | 0x0100] = byte)

// Pop processor flags from the stack
inline void MOS6502::pop_flags(void)
{
	UBYTE flags = pop_byte();

	n_flag = flags & 0x80;
	v_flag = flags & 0x40;
	d_flag = flags & 0x08;
	i_flag = flags & 0x04;
	z_flag = !(flags & 0x02);
	c_flag = flags & 0x01;
}

// Push processor flags onto the stack
inline void MOS6502::push_flags(CBOOL b_flag)
{
	UBYTE flags = 0x20;

	if (n_flag) flags |= 0x80;
	if (v_flag) flags |= 0x40;
	if (b_flag) flags |= 0x10;
	if (d_flag) flags |= 0x08;
	if (i_flag) flags |= 0x04;
	if (!z_flag) flags |= 0x02;
	if (c_flag) flags |= 0x01;

	push_byte(flags);
}


/*
 *  Set N and Z flags according to byte
 */

inline UBYTE MOS6502::set_nz(UBYTE byte)
{
	n_flag = (BYTE)byte < 0;
	return z_flag=byte;
}


/*
 *  Adc instruction
 */

inline void MOS6502::do_adc(UBYTE byte)
{
	if (!d_flag) {
		UWORD tmp;

		// Binary mode
		tmp = a + byte + (c_flag ? 1 : 0);
		c_flag = tmp > 0xff;
		v_flag = !((a ^ byte) & 0x80) && ((a ^ tmp) & 0x80);
		set_nz(a = tmp);

	} else {
		UWORD al, ah;

		// Decimal mode
		al = (a & 0x0f) + (byte & 0x0f) + (c_flag ? 1 : 0);		// Calculate lower nybble
		if (al > 9) al += 6;									// BCD fixup for lower nybble

		ah = (a >> 4) + (byte >> 4);							// Calculate upper nybble
		if (al > 0x0f) ah++;

		z_flag = (a + byte + (c_flag ? 1 : 0)) & 0xff;			// Set flags
		n_flag = ah & 8;
		v_flag = (((ah << 4) ^ a) & 0x80) && !((a ^ byte) & 0x80);

		if (ah > 9) ah += 6;									// BCD fixup for upper nybble
		c_flag = ah > 0x0f;										// Set carry flag
		a = (ah << 4) | (al & 0x0f);							// Compose result
	}
}


/*
 * Sbc instruction
 */

inline void MOS6502::do_sbc(UBYTE byte)
{
	UWORD tmp = a - byte - (c_flag ? 0 : 1);

	if (!d_flag) {

		// Binary mode
		c_flag = tmp < 0x100;
		v_flag = ((a ^ tmp) & 0x80) && ((a ^ byte) & 0x80);
		set_nz(a = tmp);

	} else {
		UWORD al, ah;

		// Decimal mode
		al = (a & 0x0f) - (byte & 0x0f) - (c_flag ? 0 : 1);		// Calculate lower nybble
		ah = (a >> 4) - (byte >> 4);							// Calculate upper nybble
		if (al & 0x10) {
			al -= 6;											// BCD fixup for lower nybble
			ah--;
		}
		if (ah & 0x10) ah -= 6;									// BCD fixup for upper nybble

		c_flag = tmp < 0x100;									// Set flags
		v_flag = ((a ^ tmp) & 0x80) && ((a ^ byte) & 0x80);
		set_nz(tmp);

		a = (ah << 4) | (al & 0x0f);							// Compose result
	}
}


/*
 *  Handle interrupt
 */

int MOS6502::handle_interrupt(void)
{
	if (Interrupt.intr[INT_RESET])
		Reset();

	else if (Interrupt.intr[INT_NMI]) {
		Interrupt.intr[INT_NMI] = FALSE;	// Simulate an edge-triggered input
#if PC_IS_POINTER
		push_byte((pc-pc_base) >> 8); push_byte(pc-pc_base);
#else
		push_byte(pc >> 8); push_byte(pc);
#endif
		push_flags(FALSE);
		i_flag = TRUE;
#if PC_IS_POINTER
		jump(read_word(0xfffa));
#else
		pc = read_word(0xfffa);
#endif
		return 7;

	} else if ((Interrupt.intr[INT_VICIRQ] || Interrupt.intr[INT_CIAIRQ]) && !i_flag) {
#if PC_IS_POINTER
		push_byte((pc-pc_base) >> 8); push_byte(pc-pc_base);
#else
		push_byte(pc >> 8); push_byte(pc);
#endif
		push_flags(FALSE);
		i_flag = TRUE;
#if PC_IS_POINTER
		jump(read_word(0xfffe));
#else
		pc = read_word(0xfffe);
#endif
		return 7;
	}
	return 0;
}


/*
 *  1541: Count VIA timers
 */

void MOS6502_1541::CountVIATimers(int cycles)
{
	unsigned long tmp;
	via1_t1c = tmp = via1_t1c - cycles;
	if (tmp > 0xffff) {
		via1_t1c = via1_t1l;
		via1_ifr |= 0x40;
	}
}


/*
 *  1541: ATN line probably changed state, recalc DATA
 */

void MOS6502_1541::NewATNState(void)
{
	DATAState = !(via1_prb & 0x02)
		&& ((!(via1_prb & 0x10) && TheCIA2->ATNState) || ((via1_prb & 0x10) && !TheCIA2->ATNState));
}


/*
 *  1541 interrupt by negative edge of ATN on IEC bus
 */

void MOS6502_1541::IECInterrupt(void)
{
	RAM[0x7c] = 1;
}


/*
 *  Test if interrupts are enabled (for job loop)
 */

CBOOL MOS6502_1541::InterruptEnabled(void)
{
	return !i_flag;
}


/*
 *  Reset CPU
 */

// Generic 6502
void MOS6502::Reset(void)
{
	// Clear all interrupt lines
	Interrupt.intr_any = 0;
	NMIState = FALSE;

	// Read reset vector
#if PC_IS_POINTER
	jump(read_word(0xfffc));
#else
	pc = read_word(0xfffc);
#endif
}

// 6502 (1541)
void MOS6502_1541::Reset(void)
{
	// IEC lines and VIA registers
	DATAState = TRUE;
	CLKState = TRUE;
	via1_prb = via2_prb = 0;
	via1_ifr = 0;

	// Same as 6502
	MOS6502::Reset();
}

// 6510 (C64)
void MOS6510::Reset(void)
{
	// Delete 'CBM80' if present
	if (RAM[0x8004] == 0xc3 && RAM[0x8005] == 0xc2 && RAM[0x8006] == 0xcd
	 && RAM[0x8007] == 0x38 && RAM[0x8008] == 0x30)
		RAM[0x8004] = 0;

	// Initialize extra 6510 registers and memory configuration
	RAM[0] = RAM[1] = 0;
	new_config();

	// Same as 6502
	MOS6502::Reset();
}


/*
 *  Illegal opcode encountered
 */

void MOS6502_1541::illegal_op(void)
{
	char illop_msg[80];

#if PC_IS_POINTER
	sprintf(illop_msg, "1541: Illegal opcode %02x at %04x.\n", *(pc-1), pc-pc_base-1);
#else
	sprintf(illop_msg, "1541: Illegal opcode %02x at %04x.\n", read_byte(pc-1), pc-1);
#endif
	if (ShowRequester(illop_msg, "Reset 1541", "Reset C64"))
	{
		TheC64->Reset();
	}
	Reset();
}

void MOS6510::illegal_op(void)
{
	char illop_msg[80];

#if PC_IS_POINTER
	sprintf(illop_msg, "Illegal opcode %02x at %04x.\n", *(pc-1), pc-pc_base-1);
#else
	sprintf(illop_msg, "Illegal opcode %02x at %04x.\n", read_byte(pc-1), pc-1);
#endif
	ShowRequester(illop_msg, "Reset");
	TheC64->Reset();
	Reset();
}


#if PC_IS_POINTER
/*
 *  Jump to illegal address space (PC_IS_POINTER only)
 */

void MOS6502_1541::illegal_jump(UWORD adr)
{
	char illop_msg[80];

	sprintf(illop_msg, "1541: Jump to I/O space at %04x to %04x.\n", pc-pc_base, adr);
	if (ShowRequester(illop_msg, "Reset 1541", "Reset C64")) {
		TheC64->Reset();
	}
	Reset();
}

void MOS6510::illegal_jump(UWORD adr)
{
	char illop_msg[80];

	sprintf(illop_msg, "Jump to I/O space at %04x to %04x.\n", pc-pc_base, adr);
	ShowRequester(illop_msg, "Reset");
	TheC64->Reset();
	Reset();
}
#endif


/*
 *  Emulate one line worth of 65xx instructions
 *  Returns number of cycles of last instruction
 */

// End of opcode, decrement cycles left
#define ENDOP(cyc) cycles_left -= last_cycles = cyc; break;

int MOS6502::EmulateLine(int cycles_left)
{
	UBYTE tmp, tmp2;
	UWORD adr;
	int last_cycles = 0;

	// Any pending interrupts?
	if (Interrupt.intr_any)
		cycles_left -= last_cycles = handle_interrupt();

	// Main opcode fetch/execute loop
	while (cycles_left >= 0) {
		switch (read_byte_imm()) {


		// Load group
		case 0xa9:	// LDA #imm
			set_nz(a = read_byte_imm());
			ENDOP(2);

		case 0xa5:	// LDA zero
			set_nz(a = read_byte_zero());
			ENDOP(3);

		case 0xb5:	// LDA zero,X
			set_nz(a = read_byte_zero_x());
			ENDOP(4);

		case 0xad:	// LDA abs
			set_nz(a = read_byte_abs());
			ENDOP(4);

		case 0xbd:	// LDA abs,X
			set_nz(a = read_byte_abs_x());
			ENDOP(4);

		case 0xb9:	// LDA abs,Y
			set_nz(a = read_byte_abs_y());
			ENDOP(4);

		case 0xa1:	// LDA (ind,X)
			set_nz(a = read_byte_ind_x());
			ENDOP(6);
		
		case 0xb1:	// LDA (ind),Y
			set_nz(a = read_byte_ind_y());
			ENDOP(5);

		case 0xa2:	// LDX #imm
			set_nz(x = read_byte_imm());
			ENDOP(2);

		case 0xa6:	// LDX zero
			set_nz(x = read_byte_zero());
			ENDOP(3);

		case 0xb6:	// LDX zero,Y
			set_nz(x = read_byte_zero_y());
			ENDOP(4);

		case 0xae:	// LDX abs
			set_nz(x = read_byte_abs());
			ENDOP(4);

		case 0xbe:	// LDX abs,Y
			set_nz(x = read_byte_abs_y());
			ENDOP(4);

		case 0xa0:	// LDY #imm
			set_nz(y = read_byte_imm());
			ENDOP(2);

		case 0xa4:	// LDY zero
			set_nz(y = read_byte_zero());
			ENDOP(3);

		case 0xb4:	// LDY zero,X
			set_nz(y = read_byte_zero_x());
			ENDOP(4);

		case 0xac:	// LDY abs
			set_nz(y = read_byte_abs());
			ENDOP(4);

		case 0xbc:	// LDY abs,X
			set_nz(y = read_byte_abs_x());
			ENDOP(4);


		// Store group
		case 0x85:	// STA zero
			write_byte(read_adr_zero(), a);
			ENDOP(3);

		case 0x95:	// STA zero,X
			write_byte(read_adr_zero_x(), a);
			ENDOP(4);

		case 0x8d:	// STA abs
			write_byte(read_adr_abs(), a);
			ENDOP(4);

		case 0x9d:	// STA abs,X
			write_byte(read_adr_abs_x(), a);
			ENDOP(5);

		case 0x99:	// STA abs,Y
			write_byte(read_adr_abs_y(), a);
			ENDOP(5);

		case 0x81:	// STA (ind,X)
			write_byte(read_adr_ind_x(), a);
			ENDOP(6);

		case 0x91:	// STA (ind),Y
			write_byte(read_adr_ind_y(), a);
			ENDOP(6);

		case 0x86:	// STX zero
			write_byte(read_adr_zero(), x);
			ENDOP(3);

		case 0x96:	// STX zero,Y
			write_byte(read_adr_zero_y(), x);
			ENDOP(4);

		case 0x8e:	// STX abs
			write_byte(read_adr_abs(), x);
			ENDOP(4);

		case 0x84:	// STY zero
			write_byte(read_adr_zero(), y);
			ENDOP(3);

		case 0x94:	// STY zero,X
			write_byte(read_adr_zero_x(), y);
			ENDOP(4);

		case 0x8c:	// STY abs
			write_byte(read_adr_abs(), y);
			ENDOP(4);


		// Transfer group
		case 0xaa:	// TAX
			set_nz(x = a);
			ENDOP(2);

		case 0x8a:	// TXA
			set_nz(a = x);
			ENDOP(2);

		case 0xa8:	// TAY
			set_nz(y = a);
			ENDOP(2);

		case 0x98:	// TYA
			set_nz(a = y);
			ENDOP(2);

		case 0xba:	// TSX
			set_nz(x = sp);
			ENDOP(2);

		case 0x9a:	// TXS
			sp = x;
			ENDOP(2);


		// Arithmetic group
		case 0x69:	// ADC #imm
			do_adc(read_byte_imm());
			ENDOP(2);

		case 0x65:	// ADC zero
			do_adc(read_byte_zero());
			ENDOP(3);

		case 0x75:	// ADC zero,X
			do_adc(read_byte_zero_x());
			ENDOP(4);

		case 0x6d:	// ADC abs
			do_adc(read_byte_abs());
			ENDOP(4);

		case 0x7d:	// ADC abs,X
			do_adc(read_byte_abs_x());
			ENDOP(4);

		case 0x79:	// ADC abs,Y
			do_adc(read_byte_abs_y());
			ENDOP(4);

		case 0x61:	// ADC (ind,X)
			do_adc(read_byte_ind_x());
			ENDOP(6);

		case 0x71:	// ADC (ind),Y
			do_adc(read_byte_ind_y());
			ENDOP(5);

		case 0xe9:	// SBC #imm
#if UNDOCUMENTED_OPCODES
		case 0xeb:
#endif
			do_sbc(read_byte_imm());
			ENDOP(2);

		case 0xe5:	// SBC zero
			do_sbc(read_byte_zero());
			ENDOP(3);

		case 0xf5:	// SBC zero,X
			do_sbc(read_byte_zero_x());
			ENDOP(4);

		case 0xed:	// SBC abs
			do_sbc(read_byte_abs());
			ENDOP(4);

		case 0xfd:	// SBC abs,X
			do_sbc(read_byte_abs_x());
			ENDOP(4);

		case 0xf9:	// SBC abs,Y
			do_sbc(read_byte_abs_y());
			ENDOP(4);

		case 0xe1:	// SBC (ind,X)
			do_sbc(read_byte_ind_x());
			ENDOP(6);

		case 0xf1:	// SBC (ind),Y
			do_sbc(read_byte_ind_y());
			ENDOP(5);


		// Increment/decrement group
		case 0xe8:	// INX
			set_nz(++x);
			ENDOP(2);

		case 0xca:	// DEX
			set_nz(--x);
			ENDOP(2);

		case 0xc8:	// INY
			set_nz(++y);
			ENDOP(2);

		case 0x88:	// DEY
			set_nz(--y);
			ENDOP(2);

		case 0xe6:	// INC zero
			adr = read_adr_zero();
			write_zp(adr, set_nz(read_zp(adr) + 1));
			ENDOP(5);

		case 0xf6:	// INC zero,X
			adr = read_adr_zero_x();
			write_zp(adr, set_nz(read_zp(adr) + 1));
			ENDOP(6);

		case 0xee:	// INC abs
			adr = read_adr_abs();
			write_byte(adr, set_nz(read_byte(adr) + 1));
			ENDOP(6);

		case 0xfe:	// INC abs,X
			adr = read_adr_abs_x();
			write_byte(adr, set_nz(read_byte(adr) + 1));
			ENDOP(7);

		case 0xc6:	// DEC zero
			adr = read_adr_zero();
			write_zp(adr, set_nz(read_zp(adr) - 1));
			ENDOP(5);

		case 0xd6:	// DEC zero,X
			adr = read_adr_zero_x();
			write_zp(adr, set_nz(read_zp(adr) - 1));
			ENDOP(6);

		case 0xce:	// DEC abs
			adr = read_adr_abs();
			write_byte(adr, set_nz(read_byte(adr) - 1));
			ENDOP(6);

		case 0xde:	// DEC abs,X
			adr = read_adr_abs_x();
			write_byte(adr, set_nz(read_byte(adr) - 1));
			ENDOP(7);


		// Logic group
		case 0x29:	// AND #imm
			set_nz(a &= read_byte_imm());
			ENDOP(2);

		case 0x25:	// AND zero
			set_nz(a &= read_byte_zero());
			ENDOP(3);

		case 0x35:	// AND zero,X
			set_nz(a &= read_byte_zero_x());
			ENDOP(4);

		case 0x2d:	// AND abs
			set_nz(a &= read_byte_abs());
			ENDOP(4);

		case 0x3d:	// AND abs,X
			set_nz(a &= read_byte_abs_x());
			ENDOP(4);

		case 0x39:	// AND abs,Y
			set_nz(a &= read_byte_abs_y());
			ENDOP(4);

		case 0x21:	// AND (ind,X)
			set_nz(a &= read_byte_ind_x());
			ENDOP(6);

		case 0x31:	// AND (ind),Y
			set_nz(a &= read_byte_ind_y());
			ENDOP(5);

		case 0x09:	// ORA #imm
			set_nz(a |= read_byte_imm());
			ENDOP(2);

		case 0x05:	// ORA zero
			set_nz(a |= read_byte_zero());
			ENDOP(3);

		case 0x15:	// ORA zero,X
			set_nz(a |= read_byte_zero_x());
			ENDOP(4);

		case 0x0d:	// ORA abs
			set_nz(a |= read_byte_abs());
			ENDOP(4);

		case 0x1d:	// ORA abs,X
			set_nz(a |= read_byte_abs_x());
			ENDOP(4);

		case 0x19:	// ORA abs,Y
			set_nz(a |= read_byte_abs_y());
			ENDOP(4);

		case 0x01:	// ORA (ind,X)
			set_nz(a |= read_byte_ind_x());
			ENDOP(6);

		case 0x11:	// ORA (ind),Y
			set_nz(a |= read_byte_ind_y());
			ENDOP(5);

		case 0x49:	// EOR #imm
			set_nz(a ^= read_byte_imm());
			ENDOP(2);

		case 0x45:	// EOR zero
			set_nz(a ^= read_byte_zero());
			ENDOP(3);

		case 0x55:	// EOR zero,X
			set_nz(a ^= read_byte_zero_x());
			ENDOP(4);

		case 0x4d:	// EOR abs
			set_nz(a ^= read_byte_abs());
			ENDOP(4);

		case 0x5d:	// EOR abs,X
			set_nz(a ^= read_byte_abs_x());
			ENDOP(4);

		case 0x59:	// EOR abs,Y
			set_nz(a ^= read_byte_abs_y());
			ENDOP(4);

		case 0x41:	// EOR (ind,X)
			set_nz(a ^= read_byte_ind_x());
			ENDOP(6);

		case 0x51:	// EOR (ind),Y
			set_nz(a ^= read_byte_ind_y());
			ENDOP(5);


		// Compare group
		case 0xc9:	// CMP #imm
			set_nz(adr = a - read_byte_imm());
			c_flag = adr < 0x100;
			ENDOP(2);

		case 0xc5:	// CMP zero
			set_nz(adr = a - read_byte_zero());
			c_flag = adr < 0x100;
			ENDOP(3);

		case 0xd5:	// CMP zero,X
			set_nz(adr = a - read_byte_zero_x());
			c_flag = adr < 0x100;
			ENDOP(4);

		case 0xcd:	// CMP abs
			set_nz(adr = a - read_byte_abs());
			c_flag = adr < 0x100;
			ENDOP(4);

		case 0xdd:	// CMP abs,X
			set_nz(adr = a - read_byte_abs_x());
			c_flag = adr < 0x100;
			ENDOP(4);

		case 0xd9:	// CMP abs,Y
			set_nz(adr = a - read_byte_abs_y());
			c_flag = adr < 0x100;
			ENDOP(4);

		case 0xc1:	// CMP (ind,X)
			set_nz(adr = a - read_byte_ind_x());
			c_flag = adr < 0x100;
			ENDOP(6);

		case 0xd1:	// CMP (ind),Y
			set_nz(adr = a - read_byte_ind_y());
			c_flag = adr < 0x100;
			ENDOP(5);

		case 0xe0:	// CPX #imm
			set_nz(adr = x - read_byte_imm());
			c_flag = adr < 0x100;
			ENDOP(2);

		case 0xe4:	// CPX zero
			set_nz(adr = x - read_byte_zero());
			c_flag = adr < 0x100;
			ENDOP(3);

		case 0xec:	// CPX abs
			set_nz(adr = x - read_byte_abs());
			c_flag = adr < 0x100;
			ENDOP(4);

		case 0xc0:	// CPY #imm
			set_nz(adr = y - read_byte_imm());
			c_flag = adr < 0x100;
			ENDOP(2);

		case 0xc4:	// CPY zero
			set_nz(adr = y - read_byte_zero());
			c_flag = adr < 0x100;
			ENDOP(3);

		case 0xcc:	// CPY abs
			set_nz(adr = y - read_byte_abs());
			c_flag = adr < 0x100;
			ENDOP(4);


		// Bit-test group
		case 0x24:	// BIT zero
			z_flag = a & (tmp = read_byte_zero());
			n_flag = tmp & 0x80;
			v_flag = tmp & 0x40;
			ENDOP(3);

		case 0x2c:	// BIT abs
			z_flag = a & (tmp = read_byte_abs());
			n_flag = tmp & 0x80;
			v_flag = tmp & 0x40;
			ENDOP(4);


		// Shift/rotate group
		case 0x0a:	// ASL A
			c_flag = a & 0x80;
			set_nz(a <<= 1);
			ENDOP(2);

		case 0x06:	// ASL zero
			tmp = read_zp(adr = read_adr_zero());
			c_flag = tmp & 0x80;
			write_zp(adr, set_nz(tmp << 1));
			ENDOP(5);

		case 0x16:	// ASL zero,X
			tmp = read_zp(adr = read_adr_zero_x());
			c_flag = tmp & 0x80;
			write_zp(adr, set_nz(tmp << 1));
			ENDOP(6);

		case 0x0e:	// ASL abs
			tmp = read_byte(adr = read_adr_abs());
			c_flag = tmp & 0x80;
			write_byte(adr, set_nz(tmp << 1));
			ENDOP(6);

		case 0x1e:	// ASL abs,X
			tmp = read_byte(adr = read_adr_abs_x());
			c_flag = tmp & 0x80;
			write_byte(adr, set_nz(tmp << 1));
			ENDOP(7);

		case 0x4a:	// LSR A
			c_flag = a & 0x01;
			set_nz(a >>= 1);
			ENDOP(2);

		case 0x46:	// LSR zero
			tmp = read_zp(adr = read_adr_zero());
			c_flag = tmp & 0x01;
			write_zp(adr, set_nz(tmp >> 1));
			ENDOP(5);

		case 0x56:	// LSR zero,X
			tmp = read_zp(adr = read_adr_zero_x());
			c_flag = tmp & 0x01;
			write_zp(adr, set_nz(tmp >> 1));
			ENDOP(6);

		case 0x4e:	// LSR abs
			tmp = read_byte(adr = read_adr_abs());
			c_flag = tmp & 0x01;
			write_byte(adr, set_nz(tmp >> 1));
			ENDOP(6);

		case 0x5e:	// LSR abs,X
			tmp = read_byte(adr = read_adr_abs_x());
			c_flag = tmp & 0x01;
			write_byte(adr, set_nz(tmp >> 1));
			ENDOP(7);

		case 0x2a:	// ROL A
			tmp2 = a & 0x80;
			set_nz(a = c_flag ? (a << 1) | 0x01 : a << 1);
			c_flag = tmp2;
			ENDOP(2);

		case 0x26:	// ROL zero
			tmp = read_zp(adr = read_adr_zero());
			tmp2 = tmp & 0x80;
			write_zp(adr, set_nz(c_flag ? (tmp << 1) | 0x01 : tmp << 1));
			c_flag = tmp2;
			ENDOP(5);

		case 0x36:	// ROL zero,X
			tmp = read_zp(adr = read_adr_zero_x());
			tmp2 = tmp & 0x80;
			write_zp(adr, set_nz(c_flag ? (tmp << 1) | 0x01 : tmp << 1));
			c_flag = tmp2;
			ENDOP(6);

		case 0x2e:	// ROL abs
			tmp = read_byte(adr = read_adr_abs());
			tmp2 = tmp & 0x80;
			write_byte(adr, set_nz(c_flag ? (tmp << 1) | 0x01 : tmp << 1));
			c_flag = tmp2;
			ENDOP(6);

		case 0x3e:	// ROL abs,X
			tmp = read_byte(adr = read_adr_abs_x());
			tmp2 = tmp & 0x80;
			write_byte(adr, set_nz(c_flag ? (tmp << 1) | 0x01 : tmp << 1));
			c_flag = tmp2;
			ENDOP(7);

		case 0x6a:	// ROR A
			tmp2 = a & 0x01;
			set_nz(a = (c_flag ? (a >> 1) | 0x80 : a >> 1));
			c_flag = tmp2;
			ENDOP(2);

		case 0x66:	// ROR zero
			tmp = read_zp(adr = read_adr_zero());
			tmp2 = tmp & 0x01;
			write_zp(adr, set_nz(c_flag ? (tmp >> 1) | 0x80 : tmp >> 1));
			c_flag = tmp2;
			ENDOP(5);

		case 0x76:	// ROR zero,X
			tmp = read_zp(adr = read_adr_zero_x());
			tmp2 = tmp & 0x01;
			write_zp(adr, set_nz(c_flag ? (tmp >> 1) | 0x80 : tmp >> 1));
			c_flag = tmp2;
			ENDOP(6);

		case 0x6e:	// ROR abs
			tmp = read_byte(adr = read_adr_abs());
			tmp2 = tmp & 0x01;
			write_byte(adr, set_nz(c_flag ? (tmp >> 1) | 0x80 : tmp >> 1));
			c_flag = tmp2;
			ENDOP(6);

		case 0x7e:	// ROR abs,X
			tmp = read_byte(adr = read_adr_abs_x());
			tmp2 = tmp & 0x01;
			write_byte(adr, set_nz(c_flag ? (tmp >> 1) | 0x80 : tmp >> 1));
			c_flag = tmp2;
			ENDOP(7);


		// Stack group
		case 0x48:	// PHA
			push_byte(a);
			ENDOP(3);

		case 0x68:	// PLA
			set_nz(a = pop_byte());
			ENDOP(4);

		case 0x08:	// PHP
			push_flags(TRUE);
			ENDOP(3);

		case 0x28:	// PLP
			pop_flags();
			if (Interrupt.intr_any && !i_flag)
				cycles_left -= handle_interrupt();
			ENDOP(4);


		// Jump/branch group
		case 0x4c:	// JMP abs
#if PC_IS_POINTER
			jump(read_adr_abs());
#else
			pc = read_word(pc);
#endif
			ENDOP(3);

		case 0x6c:	// JMP (ind)
#if PC_IS_POINTER
			adr = read_adr_abs();
			jump(read_byte(adr) | (read_byte((adr + 1) & 0xff | adr & 0xff00) << 8));
#else
			adr = read_word(pc);
			pc = read_byte(adr) | (read_byte((adr + 1) & 0xff | adr & 0xff00) << 8);
#endif
			ENDOP(5);

		case 0x20:	// JSR abs
#if PC_IS_POINTER
			push_byte((pc-pc_base+1) >> 8); push_byte(pc-pc_base+1);
			jump(read_adr_abs());
#else
			push_byte(pc+1 >> 8); push_byte(pc+1);
			pc = read_word(pc);
#endif
			ENDOP(6);

		case 0x60:	// RTS
#if PC_IS_POINTER
			jump(pop_word() + 1);
#else
			pc = pop_word() + 1;
#endif
			ENDOP(6);

		case 0x40:	// RTI
			pop_flags();
#if PC_IS_POINTER
			jump(pop_word());
#else
			pc = pop_word();
#endif
			if (Interrupt.intr_any && !i_flag)
				cycles_left -= handle_interrupt();
			ENDOP(6);

		case 0x00:	// BRK
#if PC_IS_POINTER
			push_byte((pc-pc_base) >> 8); push_byte(pc-pc_base);
#else
			push_byte(pc >> 8); push_byte(pc);
#endif
			push_flags(TRUE);
			i_flag = TRUE;
#if PC_IS_POINTER
			jump(read_word(0xfffe));
#else
			pc = read_word(0xfffe);
#endif
			ENDOP(7);

#if PC_IS_POINTER
#define Branch(flag) \
	if (flag) { \
		pc += (BYTE)*pc + 1; \
		ENDOP(3); \
	} else { \
		pc++; \
		ENDOP(2); \
	}
#else
#define Branch(flag) \
	if (flag) { \
		pc += (BYTE)read_byte(pc) + 1; \
		ENDOP(3); \
	} else { \
		pc++; \
		ENDOP(2); \
	}
#endif

		case 0xb0:	// BCS rel
			Branch(c_flag);

		case 0x90:	// BCC rel
			Branch(!c_flag);

		case 0xf0:	// BEQ rel
			Branch(!z_flag);

		case 0xd0:	// BNE rel
			Branch(z_flag);

		case 0x70:	// BVS rel
			Branch(v_flag);

		case 0x50:	// BVC rel
			Branch(!v_flag);

		case 0x30:	// BMI rel
			Branch(n_flag);

		case 0x10:	// BPL rel
			Branch(!n_flag);


		// Flags group
		case 0x38:	// SEC
			c_flag = TRUE;
			ENDOP(2);

		case 0x18:	// CLC
			c_flag = FALSE;
			ENDOP(2);

		case 0xf8:	// SED
			d_flag = TRUE;
			ENDOP(2);

		case 0xd8:	// CLD
			d_flag = FALSE;
			ENDOP(2);

		case 0x78:	// SEI
			i_flag = TRUE;
			ENDOP(2);

		case 0x58:	// CLI
			i_flag = FALSE;
			if (Interrupt.intr_any)
				cycles_left -= handle_interrupt();
			ENDOP(2);

		case 0xb8:	// CLV
			v_flag = FALSE;
			ENDOP(2);


		// NOP group
		case 0xea:	// NOP
			ENDOP(2);


#if UNDOCUMENTED_OPCODES
		// NOP group
		case 0x1a:
		case 0x3a:
		case 0x5a:
		case 0x7a:
		case 0xda:
		case 0xfa:
			ENDOP(2);

		case 0x04:	// NOP zero
		case 0x44:
		case 0x64:
		case 0x80:
		case 0x82:
		case 0x89:
		case 0xc2:
		case 0xe2:
			pc++;
			ENDOP(3);

		case 0x14:	// NOP zero,X
		case 0x34:
		case 0x54:
		case 0x74:
		case 0xd4:
		case 0xf4:
			pc++;
			ENDOP(4);

		case 0x0c:	// NOP abs
		case 0x1c:	// NOP abs,X
		case 0x3c:
		case 0x5c:
		case 0x7c:
		case 0xdc:
		case 0xfc:
			pc+=2;
			ENDOP(4);


		// Load A/X group
		case 0xa7:	// LAX zero
			set_nz(a = x = read_byte_zero());
			ENDOP(3);

		case 0xb7:	// LAX zero,Y
			set_nz(a = x = read_byte_zero_y());
			ENDOP(4);

		case 0xaf:	// LAX abs
			set_nz(a = x = read_byte_abs());
			ENDOP(4);

		case 0xbf:	// LAX abs,Y
			set_nz(a = x = read_byte_abs_y());
			ENDOP(4);

		case 0xa3:	// LAX (ind,X)
			set_nz(a = x = read_byte_ind_x());
			ENDOP(6);

		case 0xb3:	// LAX (ind),Y
			set_nz(a = x = read_byte_ind_y());
			ENDOP(5);


		// Store A/X group
		case 0x87:	// SAX zero
			write_byte(read_adr_zero(), a & x);
			ENDOP(3);

		case 0x97:	// SAX zero,Y
			write_byte(read_adr_zero_y(), a & x);
			ENDOP(4);

		case 0x8f:	// SAX abs
			write_byte(read_adr_abs(), a & x);
			ENDOP(4);

		case 0x83:	// SAX (ind,X)
			write_byte(read_adr_ind_x(), a & x);
			ENDOP(6);


		// ASL/ORA group
#define ShiftLeftOr \
	c_flag = tmp & 0x80; \
	tmp <<= 1; \
	set_nz(a |= tmp);

		case 0x07:	// SLO zero
			tmp = read_zp(adr = read_adr_zero());
			ShiftLeftOr;
			write_zp(adr, tmp);
			ENDOP(5);

		case 0x17:	// SLO zero,X
			tmp = read_zp(adr = read_adr_zero_x());
			ShiftLeftOr;
			write_zp(adr, tmp);
			ENDOP(6);

		case 0x0f:	// SLO abs
			tmp = read_byte(adr = read_adr_abs());
			ShiftLeftOr;
			write_byte(adr, tmp);
			ENDOP(6);

		case 0x1f:	// SLO abs,X
			tmp = read_byte(adr = read_adr_abs_x());
			ShiftLeftOr;
			write_byte(adr, tmp);
			ENDOP(7);

		case 0x1b:	// SLO abs,Y
			tmp = read_byte(adr = read_adr_abs_y());
			ShiftLeftOr;
			write_byte(adr, tmp);
			ENDOP(7);

		case 0x03:	// SLO (ind,X)
			tmp = read_byte(adr = read_adr_ind_x());
			ShiftLeftOr;
			write_byte(adr, tmp);
			ENDOP(8);

		case 0x13:	// SLO (ind),Y
			tmp = read_byte(adr = read_adr_ind_y());
			ShiftLeftOr;
			write_byte(adr, tmp);
			ENDOP(8);


		// ROL/AND group
#define RoLeftAnd \
	tmp2 = tmp & 0x80; \
	tmp = c_flag ? (tmp << 1) | 0x01 : tmp << 1; \
	set_nz(a &= tmp); \
	c_flag = tmp2;

		case 0x27:	// RLA zero
			tmp = read_zp(adr = read_adr_zero());
			RoLeftAnd;
			write_zp(adr, tmp);
			ENDOP(5);

		case 0x37:	// RLA zero,X
			tmp = read_zp(adr = read_adr_zero_x());
			RoLeftAnd;
			write_zp(adr, tmp);
			ENDOP(6);

		case 0x2f:	// RLA abs
			tmp = read_byte(adr = read_adr_abs());
			RoLeftAnd;
			write_byte(adr, tmp);
			ENDOP(6);

		case 0x3f:	// RLA abs,X
			tmp = read_byte(adr = read_adr_abs_x());
			RoLeftAnd;
			write_byte(adr, tmp);
			ENDOP(7);

		case 0x3b:	// RLA abs,Y
			tmp = read_byte(adr = read_adr_abs_y());
			RoLeftAnd;
			write_byte(adr, tmp);
			ENDOP(7);

		case 0x23:	// RLA (ind,X)
			tmp = read_byte(adr = read_adr_ind_x());
			RoLeftAnd;
			write_byte(adr, tmp);
			ENDOP(8);

		case 0x33:	// RLA (ind),Y
			tmp = read_byte(adr = read_adr_ind_y());
			RoLeftAnd;
			write_byte(adr, tmp);
			ENDOP(8);


		// LSR/EOR group
#define ShiftRightEor \
	c_flag = tmp & 0x01; \
	tmp >>= 1; \
	set_nz(a ^= tmp);

		case 0x47:	// SRE zero
			tmp = read_zp(adr = read_adr_zero());
			ShiftRightEor;
			write_zp(adr, tmp);
			ENDOP(5);

		case 0x57:	// SRE zero,X
			tmp = read_zp(adr = read_adr_zero_x());
			ShiftRightEor;
			write_zp(adr, tmp);
			ENDOP(6);

		case 0x4f:	// SRE abs
			tmp = read_byte(adr = read_adr_abs());
			ShiftRightEor;
			write_byte(adr, tmp);
			ENDOP(6);

		case 0x5f:	// SRE abs,X
			tmp = read_byte(adr = read_adr_abs_x());
			ShiftRightEor;
			write_byte(adr, tmp);
			ENDOP(7);

		case 0x5b:	// SRE abs,Y
			tmp = read_byte(adr = read_adr_abs_y());
			ShiftRightEor;
			write_byte(adr, tmp);
			ENDOP(7);

		case 0x43:	// SRE (ind,X)
			tmp = read_byte(adr = read_adr_ind_x());
			ShiftRightEor;
			write_byte(adr, tmp);
			ENDOP(8);

		case 0x53:	// SRE (ind),Y
			tmp = read_byte(adr = read_adr_ind_y());
			ShiftRightEor;
			write_byte(adr, tmp);
			ENDOP(8);


		// ROR/ADC group (v_flag !!)
#define RoRightAdc \
	tmp2 = tmp & 0x01; \
	tmp = c_flag ? (tmp >> 1) | 0x80 : tmp >> 1; \
	set_nz(a += tmp + (tmp2 ? 1 : 0)); \
	c_flag = tmp2;

		case 0x67:	// RRA zero
			tmp = read_zp(adr = read_adr_zero());
			RoRightAdc;
			write_zp(adr, tmp);
			ENDOP(5);

		case 0x77:	// RRA zero,X
			tmp = read_zp(adr = read_adr_zero_x());
			RoRightAdc;
			write_zp(adr, tmp);
			ENDOP(6);

		case 0x6f:	// RRA abs
			tmp = read_byte(adr = read_adr_abs());
			RoRightAdc;
			write_byte(adr, tmp);
			ENDOP(6);

		case 0x7f:	// RRA abs,X
			tmp = read_byte(adr = read_adr_abs_x());
			RoRightAdc;
			write_byte(adr, tmp);
			ENDOP(7);

		case 0x7b:	// RRA abs,Y
			tmp = read_byte(adr = read_adr_abs_y());
			RoRightAdc;
			write_byte(adr, tmp);
			ENDOP(7);

		case 0x63:	// RRA (ind,X)
			tmp = read_byte(adr = read_adr_ind_x());
			RoRightAdc;
			write_byte(adr, tmp);
			ENDOP(8);

		case 0x73:	// RRA (ind),Y
			tmp = read_byte(adr = read_adr_ind_y());
			RoRightAdc;
			write_byte(adr, tmp);
			ENDOP(8);


		// DEC/CMP group
#define DecCompare \
	set_nz(a - tmp); \
	c_flag = (a - tmp) < 0x100;

		case 0xc7:	// DCP zero
			tmp = read_zp(adr = read_adr_zero()) - 1;
			DecCompare;
			write_zp(adr, tmp);
			ENDOP(5);

		case 0xd7:	// DCP zero,X
			tmp = read_zp(adr = read_adr_zero_x()) - 1;
			DecCompare;
			write_zp(adr, tmp);
			ENDOP(6);

		case 0xcf:	// DCP abs
			tmp = read_byte(adr = read_adr_abs()) - 1;
			DecCompare;
			write_byte(adr, tmp);
			ENDOP(6);

		case 0xdf:	// DCP abs,X
			tmp = read_byte(adr = read_adr_abs_x()) - 1;
			DecCompare;
			write_byte(adr, tmp);
			ENDOP(7);

		case 0xdb:	// DCP abs,Y
			tmp = read_byte(adr = read_adr_abs_y()) - 1;
			DecCompare;
			write_byte(adr, tmp);
			ENDOP(7);

		case 0xc3:	// DCP (ind,X)
			tmp = read_byte(adr = read_adr_ind_x()) - 1;
			DecCompare;
			write_byte(adr, tmp);
			ENDOP(8);

		case 0xd3:	// DCP (ind),Y
			tmp = read_byte(adr = read_adr_ind_y()) - 1;
			DecCompare;
			write_byte(adr, tmp);
			ENDOP(8);


		// INC/SBC group
#define IncAdc {\
	UWORD tmp3; \
	tmp3 = a - tmp - (c_flag ? 0 : 1); \
	c_flag = tmp3 < 0x100; \
	v_flag = ((a ^ tmp3) & 0x80) && ((a ^ tmp) & 0x80); \
	set_nz(a = tmp3); }

		case 0xe7:	// ISB zero
			tmp = read_zp(adr = read_adr_zero()) - 1;
			IncAdc;
			write_zp(adr, tmp);
			ENDOP(5);

		case 0xf7:	// ISB zero,X
			tmp = read_zp(adr = read_adr_zero_x()) - 1;
			IncAdc;
			write_zp(adr, tmp);
			ENDOP(6);

		case 0xef:	// ISB abs
			tmp = read_byte(adr = read_adr_abs()) - 1;
			IncAdc;
			write_byte(adr, tmp);
			ENDOP(6);

		case 0xff:	// ISB abs,X
			tmp = read_byte(adr = read_adr_abs_x()) - 1;
			IncAdc;
			write_byte(adr, tmp);
			ENDOP(7);

		case 0xfb:	// ISB abs,Y
			tmp = read_byte(adr = read_adr_abs_y()) - 1;
			IncAdc;
			write_byte(adr, tmp);
			ENDOP(7);

		case 0xe3:	// ISB (ind,X)
			tmp = read_byte(adr = read_adr_ind_x()) - 1;
			IncAdc;
			write_byte(adr, tmp);
			ENDOP(8);

		case 0xf3:	// ISB (ind),Y
			tmp = read_byte(adr = read_adr_ind_y()) - 1;
			IncAdc;
			write_byte(adr, tmp);
			ENDOP(8);


		// Complex functions
		case 0x0b:	// ANC #imm
		case 0x2b:
			set_nz(a &= read_byte_imm());
			c_flag = n_flag;
			ENDOP(2);

		case 0x4b:	// ASR #imm
			a &= read_byte_imm();
			c_flag = a & 0x01;
			set_nz(a >>= 1);
			ENDOP(2);

		case 0x6b:	// ARR #imm
			a &= read_byte_imm();
			tmp2 = a & 0x01;
			set_nz(a = (c_flag ? (a >> 1) | 0x80 : a >> 1));
			c_flag = tmp2;
			ENDOP(2);

		case 0x8b:	// ANE #imm
			set_nz(a = (a | 0xee) & read_byte_imm() & x);
			ENDOP(2);

		case 0x9b:	// SHS abs,Y
#if PC_IS_POINTER
			tmp2 = pc[1];
#else
			tmp2 = read_byte(pc+1);
#endif
			write_byte(read_adr_abs_y(), a & x & (tmp2+1));
			sp = a & x;
			ENDOP(5);

		case 0x9c:	// SHY abs,X
#if PC_IS_POINTER
			tmp2 = pc[1];
#else
			tmp2 = read_byte(pc+1);
#endif
			write_byte(read_adr_abs_x(), y & (tmp2+1));
			ENDOP(5);

		case 0x9e:	// SHX abs,Y
#if PC_IS_POINTER
			tmp2 = pc[1];
#else
			tmp2 = read_byte(pc+1);
#endif
			write_byte(read_adr_abs_y(), x & (tmp2+1));
			ENDOP(5);

		case 0x9f:	// SHA abs,Y
#if PC_IS_POINTER
			tmp2 = pc[1];
#else
			tmp2 = read_byte(pc+1);
#endif
			write_byte(read_adr_abs_y(), a & x & (tmp2+1));
			ENDOP(5);

		case 0xab:	// LXA #imm
			set_nz(x = (a &= read_byte_imm()));
			ENDOP(2);

		case 0xbb:	// LAS abs,Y (??)
			set_nz(a = x = read_byte_abs_y() & sp);
			ENDOP(4);

		case 0xcb:	// SBX #imm
			x &= a;
			adr = x - read_byte_imm();
			c_flag = tmp < 0x100;
			set_nz(x = tmp);
			ENDOP(2);
#endif


		// Extensions
		case 0xf2:	// Kernal patches (IEC)
			opcode_f2();
			break;


		default:
			illegal_op();
		}
	}
	return last_cycles;
}


/*
 *  0xf2 extension opcodes (6502/1541: illegal opcode)
 */

void MOS6502_1541::opcode_f2(void)
{
	illegal_op();
}


/*
 *  0xf2 extension opcodes (6510: IEC emulation)
 */

void MOS6510::opcode_f2(void)
{
#if PC_IS_POINTER
	if ((pc-pc_base) < 0xe000) {
#else
	if (pc < 0xe000) {
#endif
		illegal_op();
		return;
	}
	switch (read_byte_imm()) {
		case 0x00:
			RAM[0x90] |= TheIEC->Out(RAM[0x95], RAM[0xa3] & 0x80);
			c_flag = FALSE;
#if PC_IS_POINTER
			jump(0xedac);
#else
			pc = 0xedac;
#endif
			break;
		case 0x01:
			RAM[0x90] |= TheIEC->OutATN(RAM[0x95]);
			c_flag = FALSE;
#if PC_IS_POINTER
			jump(0xedac);
#else
			pc = 0xedac;
#endif
			break;
		case 0x02:
			RAM[0x90] |= TheIEC->OutSec(RAM[0x95]);
			c_flag = FALSE;
#if PC_IS_POINTER
			jump(0xedac);
#else
			pc = 0xedac;
#endif
			break;
		case 0x03:
			RAM[0x90] |= TheIEC->In(&a);
			set_nz(a);
			c_flag = FALSE;
#if PC_IS_POINTER
			jump(0xedac);
#else
			pc = 0xedac;
#endif
			break;
		case 0x04:
			TheIEC->SetATN();
#if PC_IS_POINTER
			jump(0xedfb);
#else
			pc = 0xedfb;
#endif
			break;
		case 0x05:
			TheIEC->RelATN();
#if PC_IS_POINTER
			jump(0xedac);
#else
			pc = 0xedac;
#endif
			break;
		case 0x06:
			TheIEC->Turnaround();
#if PC_IS_POINTER
			jump(0xedac);
#else
			pc = 0xedac;
#endif
			break;
		case 0x07:
			TheIEC->Release();
#if PC_IS_POINTER
			jump(0xedac);
#else
			pc = 0xedac;
#endif
			break;
		default:
			illegal_op();
	}
}
