/*
 *  CPU.h - 6510/6502 emulation (line based)
 *
 *  Frodo (C) 1996 Christian Bauer
 */

#ifndef _CPU_H
#define _CPU_H

#include "Types.h"


// Set this to 1 if you want emulation of undocumented 6502 opcodes
#define UNDOCUMENTED_OPCODES 1

// Set this to 1 if the 6502 PC should be represented by a real pointer
#define PC_IS_POINTER 1


// Interrupt types
enum {
	INT_VICIRQ,
	INT_CIAIRQ,
	INT_NMI,
	INT_RESET
};


class MOS6569;
class MOS6581;
class MOS6526_1;
class MOS6526_2;
class IEC;
class C64;
struct MOS6502State;
struct MOS6510State;


// Abstract base class for 65xx CPU emulation
class MOS6502 {
public:
	MOS6502();

	int EmulateLine(int cycles_left);	// Emulate until cycles_left underflows
	virtual void Reset(void);			// Reset the CPU
	void AsyncReset(void);				// Reset the CPU asynchronously
	void AsyncNMI(void);				// Raise NMI asynchronously (NMI pulse)

	UBYTE *RAM;			// Pointer to main RAM

	union {				// Pending interrupts
		CBOOL intr[4];	// Index: See definitions above
		unsigned long intr_any;
	} Interrupt;
	CBOOL NMIState;		// State of NMI line

protected:
	virtual UBYTE read_byte(UWORD adr)=0;
	UWORD read_word(UWORD adr);
	virtual void write_byte(UWORD adr, UBYTE byte)=0;

	UBYTE read_zp(UWORD adr);
	UWORD read_zp_word(UWORD adr);
	void write_zp(UWORD adr, UBYTE byte);

	UWORD read_adr_abs(void);

	UWORD pop_word(void);
	void pop_flags(void);
	void push_flags(CBOOL b_flag);

	virtual void new_config(void)=0;

#if PC_IS_POINTER
	virtual void jump(UWORD adr)=0;
	virtual void illegal_jump(UWORD adr)=0;
#endif

	UBYTE set_nz(UBYTE byte);

	void do_adc(UBYTE byte);
	void do_sbc(UBYTE byte);

	int handle_interrupt(void);

	virtual void illegal_op(void)=0;
	virtual void opcode_f2(void)=0;

	CBOOL n_flag, v_flag, d_flag, i_flag, z_flag, c_flag;
	UBYTE a, x, y, sp;
#if PC_IS_POINTER
	UBYTE *pc, *pc_base;
#else
	UWORD pc;
#endif
};


// 6502 emulation (1541)
class MOS6502_1541 : public MOS6502 {
public:
	MOS6502_1541();

	virtual void Reset(void);
	void GetState(MOS6502State *state);
	void SetState(MOS6502State *state);
	UBYTE ExtReadByte(UWORD adr);
	void ExtWriteByte(UWORD adr, UBYTE byte);
	void CountVIATimers(int cycles);
	void NewATNState(void);
	void IECInterrupt(void);
	CBOOL InterruptEnabled(void);

	UBYTE *ROM;			// Pointer to ROM

	C64 *TheC64;		// Pointer to C64 object
	MOS6526_2 *TheCIA2;	// Pointer to C64 CIA 2

	CBOOL DATAState;		// State of IEC lines
	CBOOL CLKState;

private:
	virtual UBYTE read_byte(UWORD adr);
	virtual void write_byte(UWORD adr, UBYTE byte);
	virtual void new_config(void);
#if PC_IS_POINTER
	virtual void jump(UWORD adr);
	virtual void illegal_jump(UWORD adr);
#endif
	virtual void illegal_op(void);
	virtual void opcode_f2(void);

	UBYTE via1_prb;		// PRB of VIA 1
	UBYTE via2_prb;		// PRB of VIA 2
	UWORD via1_t1c;		// T1 Counter of VIA 1
	UWORD via1_t1l;		// T1 Latch of VIA 1
	UBYTE via1_ifr;		// IFR of VIA 1
};


// 6510 emulation (C64)
class MOS6510 : public MOS6502 {
public:
	MOS6510();

	virtual void Reset(void);
	void GetState(MOS6510State *state);
	void SetState(MOS6510State *state);
	UBYTE ExtReadByte(UWORD adr);
	void ExtWriteByte(UWORD adr, UBYTE byte);

	int ExtConfig;	// Memory configuration for ExtRead/WriteByte (0..7)

	UBYTE *Basic, *Kernal, *Char, *Color;	// Pointers to ROMs and color RAM

	C64 *TheC64;		// Pointer to C64 object
	MOS6569 *TheVIC;	// Pointer to VIC
	MOS6581 *TheSID;	// Pointer to SID
	MOS6526_1 *TheCIA1;	// Pointer to CIA 1
	MOS6526_2 *TheCIA2;	// Pointer to CIA 2

	IEC *TheIEC;		// Pointer to drive array

private:
	virtual UBYTE read_byte(UWORD adr);
	virtual void write_byte(UWORD adr, UBYTE byte);
	virtual void new_config(void);
#if PC_IS_POINTER
	virtual void jump(UWORD adr);
	virtual void illegal_jump(UWORD adr);
#endif
	virtual void illegal_op(void);
	virtual void opcode_f2(void);

	UBYTE read_emulator_id(UWORD adr);

	CBOOL basic_in, kernal_in, char_in, io_in;
	UBYTE dfff_byte;
};


// 6502 state
struct MOS6502State {
	UBYTE a, x, y;
	UBYTE p;			// Processor flags
	UWORD pc, sp;
};

// 6510 state
struct MOS6510State {
	UBYTE a, x, y;
	UBYTE p;			// Processor flags
	UBYTE ddr, pr;		// Port
	UWORD pc, sp;
};

#endif
