/*
 *  Display_x.i - C64 graphics display, emulator window handling,
 *                X specific stuff
 *
 *  Frodo (C) 1996 Christian Bauer
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>

static Display *display;
static int screen;
static Window rootwin, mywin;

static GC whitegc,blackgc;
static XColor black,white;
static Colormap cmap;

static XImage *img;
static Visual *vis;
static XVisualInfo visualInfo;
static int bitdepth;
static char *bufmem;
static int hsize;
 /* Keyboard and mouse */

static int keystate[256];
static int joystate = 0xFF;

static const long int eventmask = (KeyPressMask|KeyReleaseMask
                                  |FocusChangeMask
                                  |ExposureMask);



/*
  C64 keyboard matrix:

    Bit 7   6   5   4   3   2   1   0
  0    CUD  F5  F3  F1  F7 CLR RET DEL
  1    SHL  E   S   Z   4   A   W   3
  2     X   T   F   C   6   D   R   5
  3     V   U   H   B   8   G   Y   7
  4     N   O   K   M   0   J   I   9
  5     ,   @   :   .   -   L   P   +
  6     /   ^   =  SHR HOM  ;   *   
  7    R/S  Q   C= SPC  2  CTL  <-  1
*/

#define MATRIX(a,b) (((a) << 3) | (b))
#define KEY_F10 512
#define KEY_F11 513
#define KEY_F12 514

#define KEY_FIRE 515
#define KEY_JUP 516
#define KEY_JDN 517
#define KEY_JLF 518
#define KEY_JRT 519


/* Decode KeySyms. This function knows about all keys that are common 
 * between different keyboard languages. */
static int kc_decode (KeySym ks)
{
    switch (ks) {      
     case XK_KP_Divide: return MATRIX(6,7);
     case XK_KP_Multiply: return MATRIX(6,1);
     case XK_KP_Subtract: return MATRIX(5,3);
     case XK_KP_Add: return MATRIX(5,0);
     case XK_KP_Enter: return MATRIX(0,1);
               
     case XK_F10: return KEY_F10;
     case XK_F11: return KEY_F11;
     case XK_F12: return KEY_F12;

     case XK_comma: return MATRIX(5,7);
     case XK_period: return MATRIX(5,4);

     case XK_A: case XK_a: return MATRIX(1,2);
     case XK_B: case XK_b: return MATRIX(3,4);
     case XK_C: case XK_c: return MATRIX(2,4);
     case XK_D: case XK_d: return MATRIX(2,2);
     case XK_E: case XK_e: return MATRIX(1,6);
     case XK_F: case XK_f: return MATRIX(2,5);
     case XK_G: case XK_g: return MATRIX(3,2);
     case XK_H: case XK_h: return MATRIX(3,5);
     case XK_I: case XK_i: return MATRIX(4,1);
     case XK_J: case XK_j: return MATRIX(4,2);
     case XK_K: case XK_k: return MATRIX(4,5);
     case XK_L: case XK_l: return MATRIX(5,2);
     case XK_M: case XK_m: return MATRIX(4,4);
     case XK_N: case XK_n: return MATRIX(4,7);
     case XK_O: case XK_o: return MATRIX(4,6);
     case XK_P: case XK_p: return MATRIX(5,1);
     case XK_Q: case XK_q: return MATRIX(7,6);
     case XK_R: case XK_r: return MATRIX(2,1);
     case XK_S: case XK_s: return MATRIX(1,5);
     case XK_T: case XK_t: return MATRIX(2,6);
     case XK_U: case XK_u: return MATRIX(3,6);
     case XK_V: case XK_v: return MATRIX(3,7);
     case XK_W: case XK_w: return MATRIX(1,1);
     case XK_X: case XK_x: return MATRIX(2,7);
     case XK_Y: case XK_y: return MATRIX(3,1);
     case XK_Z: case XK_z: return MATRIX(1,4);
       
     case XK_BackSpace: return MATRIX(0,0);
     case XK_Control_L: case XK_Control_R: return MATRIX(7,2);
     case XK_Tab: return MATRIX(7,1);
     case XK_Return: return MATRIX(0,1);
     case XK_space: return MATRIX(7,4);
     case XK_Shift_L: return MATRIX(1,7);
     case XK_Shift_R: return MATRIX(6,4);
     case XK_Escape: return MATRIX(7,7);
     case XK_Alt_L: return MATRIX(7,5);
     case XK_Alt_R: return MATRIX(7,5);
     case XK_Insert: return MATRIX(6,3);
       
     case XK_Up: return MATRIX(0,7)| 0x80;
     case XK_Down: return MATRIX(0,7);
     case XK_Left: return MATRIX(0,2) | 0x80;
     case XK_Right: return MATRIX(0,2);

     case XK_F1: return MATRIX(0,4);
     case XK_F2: return MATRIX(0,4) | 0x80;
     case XK_F3: return MATRIX(0,5);
     case XK_F4: return MATRIX(0,5) | 0x80;
     case XK_F5: return MATRIX(0,6);
     case XK_F6: return MATRIX(0,6) | 0x80;
     case XK_F7: return MATRIX(0,3);
     case XK_F8: return MATRIX(0,3) | 0x80;
           
     case XK_0: return MATRIX(4,3);
     case XK_1: return MATRIX(7,0);
     case XK_2: return MATRIX(7,3);
     case XK_3: return MATRIX(1,0);
     case XK_4: return MATRIX(1,3);
     case XK_5: return MATRIX(2,0);
     case XK_6: return MATRIX(2,3);
     case XK_7: return MATRIX(3,0);
     case XK_8: return MATRIX(3,3);
     case XK_9: return MATRIX(4,0);
           
       /* You never know which Keysyms might be missing on some workstation
        * This #ifdef should be enough. */
#if defined(XK_KP_Prior) && defined(XK_KP_Left) && defined(XK_KP_Insert) && defined (XK_KP_End)
     case XK_KP_0: case XK_KP_Insert: return KEY_FIRE;
     case XK_KP_1: case XK_KP_End: return -1;
     case XK_KP_2: case XK_KP_Down: return KEY_JDN;
     case XK_KP_3: case XK_KP_Next: return -1;
     case XK_KP_4: case XK_KP_Left: return KEY_JLF;
     case XK_KP_5: case XK_KP_Begin: return KEY_FIRE;
     case XK_KP_6: case XK_KP_Right: return KEY_JRT;
     case XK_KP_7: case XK_KP_Home: return -1;
     case XK_KP_8: case XK_KP_Up: return KEY_JUP;
     case XK_KP_9: case XK_KP_Prior: return -1;
#else
     case XK_KP_0: return KEY_FIRE;
     case XK_KP_1: return -1;
     case XK_KP_2: return KEY_JDN;
     case XK_KP_3: return -1;
     case XK_KP_4: return KEY_JLF;
     case XK_KP_5: return KEY_FIRE;
     case XK_KP_6: return KEY_JRT;
     case XK_KP_7: return -1;
     case XK_KP_8: return KEY_JUP;
     case XK_KP_9: return -1;
#endif
    }
    return -1;
}

static int decode_us(KeySym ks)
{
    switch(ks) {       /* US specific */       
     case XK_bracketleft: return MATRIX(6,6);
     case XK_bracketright: return MATRIX(6,1);
     case XK_slash: return MATRIX(6,7);
     case XK_semicolon: return MATRIX(6,2);
     case XK_minus: return MATRIX(5,3);
     case XK_equal: return MATRIX(6,5);
    }

    return -1;
}

static int decode_de(KeySym ks)
{
    switch(ks) {
       /* DE specific */
     case XK_Odiaeresis: case XK_odiaeresis: return MATRIX(5,5);
     case XK_Adiaeresis: case XK_adiaeresis: return MATRIX(6,2);
     case XK_Udiaeresis: case XK_udiaeresis: return MATRIX(6,5);
     case XK_plus: case XK_asterisk: return MATRIX(5,0);
     case XK_less: case XK_greater: return MATRIX(6,7);
     case XK_numbersign: return MATRIX(6,0);
     case XK_ssharp: return MATRIX(5,6);
     case XK_apostrophe: return MATRIX(6,1);
     case XK_asciicircum: return MATRIX(6,6);
     case XK_minus: return MATRIX(5,3);
     case XK_Tab: return MATRIX(7,1);
    }

    return -1;
}

static int keycode2c64(XKeyEvent *event)
{
	KeySym ks;
	int as;
	int index = 0;
	
	do {
		ks = XLookupKeysym(event, index);
		as = kc_decode (ks);
       
		if (as == -1) {
			as = decode_de(ks);
		}
		if(-1 != as)
			return as;
		index++;
	} while (ks != NoSymbol);
	return -1;
}

C64Display::C64Display()
{
}

int init_graphics(void)
{
       int i;
       char *display_name = 0;
       XSetWindowAttributes wattr;
       int pixbytes;
       
       display = XOpenDisplay(display_name);
       if (display == 0)  {
               fprintf(stderr, "Can't connect to X server %s\n", XDisplayName(display_name));
               return 0;
       }
       
       screen = XDefaultScreen(display);
       rootwin = XRootWindow(display,screen);
#if 0  
       if (XMatchVisualInfo(display, screen, 16, TrueColor, &visualInfo)) {
       } else if (XMatchVisualInfo(display, screen, 24, TrueColor, &visualInfo)) {
       } else 
#endif
               if (XMatchVisualInfo(display, screen, 8, PseudoColor, &visualInfo)) {
               /* for our HP boxes */
       } else if (XMatchVisualInfo(display, screen, 8, GrayScale, &visualInfo)) {
       } else {
               fprintf(stderr, "Can't obtain appropriate X visual\n");
               return 0;
       }

       vis = visualInfo.visual;
       bitdepth = visualInfo.depth;
       pixbytes = (bitdepth == 24 || bitdepth == 32 ? 4
                   : bitdepth == 12 || bitdepth == 16 ? 2
                   : 1);
       hsize = (DISPLAY_X + 3) & ~3;
       bufmem = (char *)malloc(pixbytes*hsize*DISPLAY_Y);
       
       fprintf(stderr, "Using %d bit visual\n", bitdepth);
       
       cmap = XCreateColormap(display, rootwin, vis, AllocNone);
       XParseColor(display, cmap, "#000000", &black);
       if (!XAllocColor(display, cmap, &black))
               fprintf(stderr, "Whoops??\n");
       XParseColor(display, cmap, "#ffffff", &white);
       if (!XAllocColor(display, cmap, &white))
               fprintf(stderr, "Whoops??\n");
       
       wattr.event_mask = eventmask;
       wattr.background_pixel = black.pixel;
       wattr.backing_store = Always;
       wattr.backing_planes = bitdepth;
       wattr.border_pixmap = None;
       wattr.border_pixel = black.pixel;
       wattr.colormap = cmap;
       
       mywin = XCreateWindow(display, rootwin, 0, 0, DISPLAY_X, DISPLAY_Y, 0,
                             bitdepth, InputOutput, vis,
                             CWEventMask|CWBackPixel|CWBorderPixel|CWBackingStore
                             |CWBackingPlanes|CWColormap,
                             &wattr);
       XMapWindow(display,mywin);
       XStoreName(display, mywin, "Frodo");

       whitegc = XCreateGC(display,mywin,0,0);
       blackgc = XCreateGC(display,mywin,0,0);
    
       XSetForeground(display,blackgc,black.pixel);
       XSetForeground(display,whitegc,white.pixel);

       img = XCreateImage(display, vis, bitdepth, ZPixmap, 0, bufmem, 
                          hsize, DISPLAY_Y, 32, 0);
       
       for(i=0; i<256; i++)
               keystate[i] = 0;
       
       return 1;       
}

void C64Display::Update(void)
{
       XPutImage(display, mywin, blackgc, img, 0, 0, 0, 0, DISPLAY_X, DISPLAY_Y);
}

UBYTE *C64Display::BitmapBase(void)
{
       return (UBYTE *)bufmem;
}

int C64Display::BitmapXMod(void)
{
       return hsize;
}

void C64Display::PollKeyboard(UBYTE *key_matrix, UBYTE *rev_matrix, UBYTE *joystick)
{
       for(;;) {
               XEvent event;
               if (!XCheckMaskEvent(display, eventmask, &event)) 
                       break;
       
               switch(event.type) {
               case KeyPress: {                
                       int kc = keycode2c64((XKeyEvent *)&event);
                       if (kc == -1)
                               break;
                       switch (kc) {
                       case KEY_F10:
                               exit(0);
                               break;
                               
                       case KEY_F11:
                               TheC64->NMI();
                               break;
                       
                       case KEY_F12:
                               TheC64->Reset();
                               break;

                       case KEY_FIRE:
                               joystate &= ~0x10;
                               break;
                               
                       case KEY_JDN:
                               joystate &= ~0x2;
                               break;
                               
                       case KEY_JUP:
                               joystate &= ~0x1;
                               break;
                               
                       case KEY_JLF:
                               joystate &= ~0x4;
                               break;
                               
                       case KEY_JRT:
                               joystate &= ~0x8;
                               break;
                               
                       default:
                               if (keystate[kc])
                                       break;
                               keystate[kc] = 1;
                               int c64_byte, c64_bit, shifted;
                               c64_byte = kc >> 3;
                               c64_bit = kc & 7;
                               shifted = kc & 128;
                               c64_byte &= 7;
                               if (shifted) {
                                       key_matrix[6] &= 0xef;
                                       rev_matrix[4] &= 0xbf;
                               }
                               key_matrix[c64_byte] &= ~(1 << c64_bit);
                               rev_matrix[c64_bit] &= ~(1 << c64_byte);
                               break;
                       }
                       break;
               }
               case KeyRelease: {           
                       int kc = keycode2c64((XKeyEvent *)&event);
                       if (kc == -1)
                               break;
                       switch (kc) {
                       case KEY_FIRE:
                               joystate |= 0x10;
                               break;
                               
                       case KEY_JDN:
                               joystate |= 0x2;
                               break;
                               
                       case KEY_JUP:
                               joystate |= 0x1;
                               break;
                               
                       case KEY_JLF:
                               joystate |= 0x4;
                               break;
                               
                       case KEY_JRT:
                               joystate |= 0x8;
                               break;
                               
                       default:
                               if (!keystate[kc])
                                       break;
                               keystate[kc] = 0;
                               int c64_byte, c64_bit, shifted;
                               c64_byte = kc >> 3;
                               c64_bit = kc & 7;
                               shifted = kc & 128;
                               c64_byte &= 7;
                               if (shifted) {
                                       key_matrix[6] |= 0x10;
                                       rev_matrix[4] |= 0x40;
                               }
                               key_matrix[c64_byte] |= (1 << c64_bit);
                               rev_matrix[c64_bit] |= (1 << c64_byte);
                               break;
                       }
               }
               case FocusIn:
                       XAutoRepeatOff(display);
                       break;
               case FocusOut:
                       XAutoRepeatOn(display);
                       break;
               }
       }
       *joystick = joystate;
}

/*
 *  Check if NumLock is down (for switching the joystick keyboard emulation)
 */

CBOOL C64Display::NumLock(void)
{
       return 0;
}


/*
 *  Allocate C64 colors
 */

void C64Display::InitColors(UBYTE *colors)
{
       int i;
       XColor col;
       char str[20];

       for (i=0; i< 256; i++) {
               sprintf(str, "rgb:%x/%x/%x", palette_red[i & 0x0f], palette_green[i & 0x0f], palette_blue[i & 0x0f]);
               XParseColor(display, cmap, str, &col);
               if (XAllocColor(display, cmap, &col)) {
                       colors[i] = col.pixel;
               } else {
                       fprintf(stderr, "Couldn't get all colors\n");
               }
       }
}

void C64Display::UpdateLEDs(int, int, int, int)
{
}

long int ShowRequester(char *,char *b,char *)
{
       printf("%s\n",b);
       return 1;
}
