/*
 *  Prefs_Be.i - Global preferences, Be specific stuff
 *
 *  Frodo (C) 1996 Christian Bauer
 */

#include <InterfaceKit.h>
#include <StorageKit.h>


// Window background color
const rgb_color light_gray = {208, 208, 208, 0};


// Window thread messages
const ulong MSG_OK = 'okok';
const ulong MSG_CANCEL = 'cncl';
const ulong MSG_SPRITES_ON = 'spon';
const ulong MSG_SPRITE_COLLISIONS = 'scol';
const ulong MSG_JOYSTICK_1_ON = 'joy1';
const ulong MSG_JOYSTICK_2_ON = 'joy2';
const ulong MSG_JOYSTICK_SWAP = 'jswp';
const ulong MSG_LIMIT_SPEED = 'lmit';
const ulong MSG_FAST_RESET = 'frst';
const ulong MSG_CIA_IRQ_HACK = 'cirq';
const ulong MSG_MAP_SLASH = 'mpsl';
const ulong MSG_EMUL_1541_PROC = '15pr';
const ulong MSG_DRVTYPE_8 = 'drt8';
const ulong MSG_DRVTYPE_9 = 'drt9';
const ulong MSG_DRVTYPE_10 = 'drt:';
const ulong MSG_DRVTYPE_11 = 'drt;';
const ulong MSG_GETDRIVE_8 = 'gtd8';
const ulong MSG_GETDRIVE_9 = 'gtd9';
const ulong MSG_GETDRIVE_10 = 'gtd:';
const ulong MSG_GETDRIVE_11 = 'gtd;';
const ulong MSG_SID_TYPE = 'sidt';

const ulong MSG_OPEN = 'open';
const ulong MSG_SAVE = 'save';
const ulong MSG_SAVE_AS = 'svas';
const ulong MSG_REVERT = 'rvrt';


/*
 *  Preferences window class
 */

class NumberView;
class PathView;

class PrefsWindow : public BWindow {
public:
	PrefsWindow(Prefs *p, CBOOL start, char *path);
	virtual void MessageReceived(BMessage *msg);
	void OpenRequested(record_ref ref);
	virtual void SaveRequested(record_ref directory, const char *filename);
	virtual bool FilterKeyDown(ulong *aChar, BView **target);

private:
	BCheckBox *make_checkbox(BRect frame, char *label, ulong what, BView *parent);
	NumberView *make_number_entry(BRect frame, char *label_text, BView *parent);
	void make_drive_label(BRect frame, char *label, BView *parent);
	BPopUpMenu *make_drvtype_popup(BRect frame, ulong what, BView *parent);
	PathView *make_path_entry(BRect frame, BView *parent);
	BPopUpMenu *make_sidtype_popup(BRect frame, char *label_text, ulong what, BView *parent);
	void set_values(void);
	void get_values(void);
	void get_drive(int i, record_ref ref);

	Prefs *prefs;

	BButton *g_ok;
	BButton *g_cancel;
	NumberView *g_normal_cycles;
	NumberView *g_bad_line_cycles;
	NumberView *g_cia_cycles;
	NumberView *g_floppy_cycles;
	NumberView *g_skip_frames;
	BPopUpMenu *g_drive_type[4];
	PathView *g_drive_path[4];
	BPopUpMenu *g_sid_type;
	BCheckBox *g_sprites_on;
	BCheckBox *g_sprite_collisions;
	BCheckBox *g_joystick_1_on;
	BCheckBox *g_joystick_2_on;
	BCheckBox *g_joystick_swap;
	BCheckBox *g_limit_speed;
	BCheckBox *g_fast_reset;
	BCheckBox *g_cia_irq_hack;
	BCheckBox *g_map_slash;
	BCheckBox *g_emul_1541_proc;

	char *prefs_path;
	CBOOL startup;
};


/*
 *  Start preferences editor (asynchronously)
 *  startup = FALSE: Send MSG_PREFS_DONE to application on close
 *  startup = TRUE : Send MSG_STARTUP to application on close if not canceled,
 *                   B_QUIT_REQUESTED otherwise
 *  prefs_name points to the file name of the preferences (which may be changed)
 */

void Prefs::ShowEditor(CBOOL startup, char *prefs_name)
{
	PrefsWindow *win = new PrefsWindow(this, startup, prefs_name);
}


/*
 *  Number-only BTextView
 */

// Class definition
class NumberView : public BTextView {
public:
	NumberView(BRect frame, const char *name, BRect textRect, ulong resizingMode, ulong flags);
	void SetValue(long value);
	long Value(void);
};

// Constructor: Allow only digits
NumberView::NumberView(BRect frame, const char *name, BRect textRect, ulong resizingMode, ulong flags)
 : BTextView(frame, name, textRect, resizingMode, flags)
{
	for (int c=0; c<256; c++)
		if (!isdigit(c) && c != B_BACKSPACE && c != B_LEFT_ARROW && c != B_RIGHT_ARROW) 
			DisallowChar(c);
}

// Set integer value
void NumberView::SetValue(long value)
{
	char str[32];
	sprintf(str, "%ld", value);
	SetText(str);
}

// Get integer value
long NumberView::Value(void)
{
	return atol(Text());
}


/*
 *  Path-entry BTextView
 */

// Class definition
class PathView : public BTextView {
public:
	PathView(BRect frame, const char *name, BRect textRect, ulong resizingMode, ulong flags);
	virtual bool MessageDropped(BMessage *message, BPoint point, BPoint offset);
};

// Constructor: Disable some keys
PathView::PathView(BRect frame, const char *name, BRect textRect, ulong resizingMode, ulong flags)
 : BTextView(frame, name, textRect, resizingMode, flags)
{
	for (int c=0; c<' '; c++)
		if (c != B_BACKSPACE && c != B_LEFT_ARROW && c != B_RIGHT_ARROW) 
			DisallowChar(c);
}

// Message dropped: Look out for refs
bool PathView::MessageDropped(BMessage *message, BPoint point, BPoint offset)
{
	// First look for refs
	if (message->what == B_REFS_RECEIVED) {
		BFile *file = new BFile;
		BDirectory *dir = new BDirectory;
		record_ref ref;
		char path[256];

		ref = message->FindRef("refs");
		if (does_ref_conform(ref, "File") && (file->SetRef(ref) == B_NO_ERROR))
			file->GetPath(path, 256);
		else if (dir->SetRef(ref) == B_NO_ERROR)
			dir->GetPath(path, 256);
		else
			path[0] = 0;

		SetText(path);
		MakeFocus();

		delete dir;
		delete file;
		return TRUE;
	}

	// No refs, then see if the BTextView finds text or characters
	return BTextView::MessageDropped(message, point, offset);
}


/*
 *  Open preferences window
 */

PrefsWindow::PrefsWindow(Prefs *p, CBOOL start, char *path) : BWindow(BRect(0, 0, 400, 323), "Frodo Preferences", B_TITLED_WINDOW, B_NOT_CLOSABLE | B_NOT_RESIZABLE | B_NOT_ZOOMABLE)
{
	int i;
	Lock();
	BRect b = Bounds();
	prefs = p;
	startup = start;
	prefs_path = path;

	// Move window to right position
	MoveTo(80, 80);

	// Light gray background
	BView *top = new BView(BRect(0, 15, b.right, b.bottom), "", B_FOLLOW_NONE, B_WILL_DRAW);
	AddChild(top);
	top->SetViewColor(light_gray);

	// "OK" button
	top->AddChild(g_ok = new BButton(BRect(20, 273, 90, 298), "", startup ? "Start" : "OK", new BMessage(MSG_OK)));
	SetDefaultButton(g_ok);

	// "Cancel" button
	top->AddChild(g_cancel = new BButton(BRect(b.right-90, 273, b.right-20, 298), "", startup ? "Quit" : "Cancel", new BMessage(MSG_CANCEL)));

	// Number entry fields
	g_normal_cycles = make_number_entry(BRect(160, 10, 390, 26), "Cycles per line (CPU)", top);
	g_bad_line_cycles = make_number_entry(BRect(160, 30, 390, 46), "Cycles per Bad Line (CPU)", top);
	g_cia_cycles = make_number_entry(BRect(160, 50, 390, 66), "Cycles per line (CIA)", top);
	g_floppy_cycles = make_number_entry(BRect(160, 70, 390, 86), "Cycles per line (1541)", top);
	g_skip_frames = make_number_entry(BRect(160, 90, 390, 106), "Draw every n-th frame", top);
	g_sid_type = make_sidtype_popup(BRect(160, 110, 390, 126), "SID emulation type", MSG_SID_TYPE, top);

	// Checkboxes
	g_sprites_on = make_checkbox(BRect(10, 10, 160, 21), "Sprite display", MSG_SPRITES_ON, top);
	g_sprite_collisions = make_checkbox(BRect(10, 25, 160, 36), "Sprite collisions", MSG_SPRITE_COLLISIONS, top);
	g_joystick_1_on = make_checkbox(BRect(10, 40, 160, 51), "Joystick on port 1", MSG_JOYSTICK_1_ON, top);
	g_joystick_2_on = make_checkbox(BRect(10, 55, 160, 66), "Joystick on port 2", MSG_JOYSTICK_2_ON, top);
	g_joystick_swap = make_checkbox(BRect(10, 70, 160, 81), "Swap joysticks", MSG_JOYSTICK_SWAP, top);
	g_limit_speed = make_checkbox(BRect(10, 85, 160, 96), "Limit speed", MSG_LIMIT_SPEED, top);
	g_fast_reset = make_checkbox(BRect(10, 100, 160, 111), "Fast Reset", MSG_FAST_RESET, top);
	g_cia_irq_hack = make_checkbox(BRect(10, 115, 160, 126), "Clear CIA ICR on write", MSG_CIA_IRQ_HACK, top);

	// Drive settings
	BBox *drvbox = new BBox(BRect(10, 131, 390, 262));
	top->AddChild(drvbox);
	drvbox->SetViewColor(light_gray);
	drvbox->SetLowColor(light_gray);
	drvbox->SetLabel("Drives");

	make_drive_label(BRect(10, 14, 25, 30), "8", drvbox);
	make_drive_label(BRect(10, 34, 25, 50), "9", drvbox);
	make_drive_label(BRect(10, 54, 25, 70), "10", drvbox);
	make_drive_label(BRect(10, 74, 25, 90), "11", drvbox);

	for (i=0; i<4; i++) {
		g_drive_type[i] = make_drvtype_popup(BRect(30, 14+i*20, 75, 30+i*20), MSG_DRVTYPE_8 + i, drvbox);
		g_drive_path[i] = make_path_entry(BRect(80, 14+i*20, 349, 30+i*20), drvbox);
		drvbox->AddChild(new BButton(BRect(350, 14+i*20, 373, 30+i*20), "", "B", new BMessage(MSG_GETDRIVE_8 + i)));
	}

	g_map_slash = make_checkbox(BRect(10, 94, 300, 110), "Map '/'<->'\\' in filenames", MSG_MAP_SLASH, drvbox);
	g_emul_1541_proc = make_checkbox(BRect(10, 109, 300, 125), "Enable 1541 processor emulation", MSG_EMUL_1541_PROC, drvbox);

	// Set up menus
	BMenuBar *bar = new BMenuBar(BRect(0, 0, b.right, 14), "");
	BMenu *menu = new BMenu("Preferences");
	menu->AddItem(new BMenuItem("Open...", new BMessage(MSG_OPEN), 'O'));
	menu->AddItem(new BMenuItem("Save", new BMessage(MSG_SAVE), 'S'));
	menu->AddItem(new BMenuItem("Save As...", new BMessage(MSG_SAVE_AS), 'A'));
	menu->AddItem(new BMenuItem("Revert", new BMessage(MSG_REVERT)));
	menu->SetTargetForItems(this);
	bar->AddItem(menu);
	AddChild(bar);
	SetKeyMenuBar(bar);

	// Set the values of all controls to reflect the preferences
	set_values();
	g_normal_cycles->MakeFocus();

	// Show the window
	Show();
	Unlock();
}


/*
 *  Create checkbox
 */

BCheckBox *PrefsWindow::make_checkbox(BRect frame, char *label, ulong what, BView *parent)
{
	BCheckBox *checkbox = new BCheckBox(frame, "", label, new BMessage(what));
	parent->AddChild(checkbox);
	checkbox->SetViewColor(light_gray);
	return checkbox;
}


/*
 *  Create number entry field
 */

NumberView *PrefsWindow::make_number_entry(BRect frame, char *label_text, BView *parent)
{
	// Label to the left of the entry field
	BRect label_rect = frame;
	label_rect.right -= 55;

	BStringView *label = new BStringView(label_rect, "", label_text);
	parent->AddChild(label);
	label->SetAlignment(B_ALIGN_RIGHT);
	label->SetViewColor(light_gray);
	label->SetLowColor(light_gray);
	label->SetFontName("Emily");

	// Box around entry field
	BRect frame_rect = frame;
	frame_rect.left = frame_rect.right - 50;

	BBox *box = new BBox(frame_rect);
	parent->AddChild(box);

	// The entry field
	BRect textview_rect = frame_rect;
	textview_rect.OffsetTo(0, 1);
	textview_rect.InsetBy(4, 2);
	BRect text_rect = textview_rect;
	text_rect.OffsetTo(B_ORIGIN);

	NumberView *text = new NumberView(textview_rect, "", text_rect, B_FOLLOW_NONE, B_WILL_DRAW);
	box->AddChild(text);
	text->SetAlignment(B_ALIGN_RIGHT);
	text->SetFontName("Kate");

	return text;
}


/*
 *  Create drive number label
 */

void PrefsWindow::make_drive_label(BRect frame, char *label, BView *parent)
{
	BStringView *view = new BStringView(frame, "", label);
	parent->AddChild(view);
	view->SetAlignment(B_ALIGN_RIGHT);
	view->SetViewColor(light_gray);
	view->SetLowColor(light_gray);
	view->SetFontName("Emily");
}


/*
 *  Create drive type popup menu
 */

BPopUpMenu *PrefsWindow::make_drvtype_popup(BRect frame, ulong what, BView *parent)
{
	BMenuItem *item;

	BPopUpMenu *popup = new BPopUpMenu("");
	item = new BMenuItem("Dir", new BMessage(what));
	popup->AddItem(item);
	item = new BMenuItem("D64", new BMessage(what));
	popup->AddItem(item);
	popup->SetTargetForItems(this);

	BMenuBar *bar = new BMenuBar(frame, "", B_FOLLOW_NONE, B_ITEMS_IN_ROW, FALSE);
	bar->AddItem(popup);
	parent->AddChild(bar);

	return popup;
}


/*
 *  Create path entry field
 */

PathView *PrefsWindow::make_path_entry(BRect frame, BView *parent)
{
	// Box around entry field
	BBox *box = new BBox(frame);
	parent->AddChild(box);

	// The entry field
	BRect textview_rect = frame;
	textview_rect.OffsetTo(0, 1);
	textview_rect.InsetBy(4, 2);
	BRect text_rect = textview_rect;
	text_rect.OffsetTo(B_ORIGIN);

	PathView *text = new PathView(textview_rect, "", text_rect, B_FOLLOW_NONE, B_WILL_DRAW);
	box->AddChild(text);
	text->SetFontName("Kate");
	text->SetMaxChars(255);

	return text;
}


/*
 *  Create SID type popup
 */

BPopUpMenu *PrefsWindow::make_sidtype_popup(BRect frame, char *label_text, ulong what, BView *parent)
{
	// Label to the left of the entry field
	BRect label_rect = frame;
	label_rect.right -= 55;

	BStringView *label = new BStringView(label_rect, "", label_text);
	parent->AddChild(label);
	label->SetAlignment(B_ALIGN_RIGHT);
	label->SetViewColor(light_gray);
	label->SetLowColor(light_gray);
	label->SetFontName("Emily");

	// Popup menu
	BRect popup_rect = frame;
	popup_rect.left = popup_rect.right - 50;

	BMenuItem *item;

	BPopUpMenu *popup = new BPopUpMenu("");
	item = new BMenuItem("None", new BMessage(what));
	popup->AddItem(item);
	item = new BMenuItem("Digital", new BMessage(what));
	popup->AddItem(item);
	popup->SetTargetForItems(this);

	BMenuBar *bar = new BMenuBar(popup_rect, "", B_FOLLOW_NONE, B_ITEMS_IN_ROW, FALSE);
	bar->AddItem(popup);
	parent->AddChild(bar);

	return popup;
}


/*
 *  Set the values of the controls
 */

void PrefsWindow::set_values(void)
{
	prefs->Check();

	g_normal_cycles->SetValue(prefs->NormalCycles);
	g_bad_line_cycles->SetValue(prefs->BadLineCycles);
	g_cia_cycles->SetValue(prefs->CIACycles);
	g_floppy_cycles->SetValue(prefs->FloppyCycles);
	g_skip_frames->SetValue(prefs->SkipFrames);

	for (int i=0; i<4; i++) {
		g_drive_type[i]->ItemAt(prefs->DriveType[i])->SetMarked(TRUE);
		g_drive_path[i]->SetText(prefs->DrivePath[i]);
	}

	g_sid_type->ItemAt(prefs->SIDType)->SetMarked(TRUE);

	g_sprites_on->SetValue(prefs->SpritesOn ? B_CONTROL_ON : B_CONTROL_OFF);
	g_sprite_collisions->SetValue(prefs->SpriteCollisions ? B_CONTROL_ON : B_CONTROL_OFF);
	g_joystick_1_on->SetValue(prefs->Joystick1On ? B_CONTROL_ON : B_CONTROL_OFF);
	g_joystick_2_on->SetValue(prefs->Joystick2On ? B_CONTROL_ON : B_CONTROL_OFF);
	g_joystick_swap->SetValue(prefs->JoystickSwap ? B_CONTROL_ON : B_CONTROL_OFF);
	g_limit_speed->SetValue(prefs->LimitSpeed ? B_CONTROL_ON : B_CONTROL_OFF);
	g_fast_reset->SetValue(prefs->FastReset ? B_CONTROL_ON : B_CONTROL_OFF);
	g_cia_irq_hack->SetValue(prefs->CIAIRQHack ? B_CONTROL_ON : B_CONTROL_OFF);
	g_map_slash->SetValue(prefs->MapSlash ? B_CONTROL_ON : B_CONTROL_OFF);
	g_emul_1541_proc->SetValue(prefs->Emul1541Proc ? B_CONTROL_ON : B_CONTROL_OFF);
}


/*
 *  Get the values of the controls
 */

void PrefsWindow::get_values(void)
{
	prefs->NormalCycles = g_normal_cycles->Value();
	prefs->BadLineCycles = g_bad_line_cycles->Value();
	prefs->CIACycles = g_cia_cycles->Value();
	prefs->FloppyCycles = g_floppy_cycles->Value();
	prefs->SkipFrames = g_skip_frames->Value();

	for (int i=0; i<4; i++)
		strcpy(prefs->DrivePath[i], g_drive_path[i]->Text());

	prefs->Check();
}


/*
 *  Message from controls/menus received
 */

void PrefsWindow::MessageReceived(BMessage *msg)
{
	switch (msg->what) {
		case MSG_OK:
			get_values();
			if (startup)
				be_app->PostMessage(MSG_STARTUP);
			else {
				BMessage *msg = new BMessage(MSG_PREFS_DONE);
				msg->AddBool("canceled", FALSE);
				msg->AddObject("prefs", (BObject *)prefs);
				be_app->PostMessage(msg);
			}
			Quit();

		case MSG_CANCEL:
			if (startup)
				be_app->PostMessage(B_QUIT_REQUESTED);
			else {
				BMessage *msg = new BMessage(MSG_PREFS_DONE);
				msg->AddBool("canceled", TRUE);
				msg->AddObject("prefs", (BObject *)prefs);
				be_app->PostMessage(msg);
			}
			Quit();

		case MSG_SPRITES_ON:
			prefs->SpritesOn = !prefs->SpritesOn;
			break;

		case MSG_SPRITE_COLLISIONS:
			prefs->SpriteCollisions = !prefs->SpriteCollisions;
			break;

		case MSG_JOYSTICK_1_ON:
			prefs->Joystick1On = !prefs->Joystick1On;
			break;

		case MSG_JOYSTICK_2_ON:
			prefs->Joystick2On = !prefs->Joystick2On;
			break;

		case MSG_JOYSTICK_SWAP:
			prefs->JoystickSwap = !prefs->JoystickSwap;
			break;

		case MSG_LIMIT_SPEED:
			prefs->LimitSpeed = !prefs->LimitSpeed;
			break;

		case MSG_FAST_RESET:
			prefs->FastReset = !prefs->FastReset;
			break;

		case MSG_CIA_IRQ_HACK:
			prefs->CIAIRQHack = !prefs->CIAIRQHack;
			break;

		case MSG_SID_TYPE:
			prefs->SIDType = msg->FindLong("index");
			break;

		case MSG_MAP_SLASH:
			prefs->MapSlash = !prefs->MapSlash;
			break;

		case MSG_EMUL_1541_PROC:
			prefs->Emul1541Proc = !prefs->Emul1541Proc;
			break;

		case MSG_DRVTYPE_8:
		case MSG_DRVTYPE_9:
		case MSG_DRVTYPE_10:
		case MSG_DRVTYPE_11:
			prefs->DriveType[msg->what & 3] = msg->FindLong("index");
			break;

		case MSG_GETDRIVE_8:
		case MSG_GETDRIVE_9:
		case MSG_GETDRIVE_10:
		case MSG_GETDRIVE_11:
			// Frodo will send B_REFS_RECEIVED with a "GetDrive" member to us
			BMessage *sendmsg = new BMessage(B_REFS_RECEIVED);
			sendmsg->AddLong("GetDrive", msg->what & 3);
			if (prefs->DriveType[msg->what & 3] == DRVTYPE_DIR)
				be_app->RunFilePanel(NULL, NULL, NULL, TRUE, sendmsg);
			else
				be_app->RunFilePanel(NULL, NULL, NULL, FALSE, sendmsg);
			break;

		case MSG_OPEN:
			// Frodo will send B_REFS_RECEIVED to us
			be_app->RunFilePanel();
			break;

		case B_REFS_RECEIVED:
			// Either open a prefs file or get a drive path
			int i=msg->FindLong("GetDrive");
			if (msg->Error() != B_NO_ERROR)
				OpenRequested(msg->FindRef("refs"));
			else
				get_drive(i, msg->FindRef("refs"));
			break;

		case MSG_SAVE:
			get_values();
			prefs->Save(prefs_path);
			break;

		case MSG_SAVE_AS:
			RunSavePanel(prefs_path);
			break;

		case MSG_REVERT:
			get_values();	// Useful if Load() is unsuccessful
			prefs->Load(prefs_path);
			set_values();
			break;

		default:
			inherited::MessageReceived(msg);
	}
}


/*
 *  Open panel done, open prefs requested
 */

void PrefsWindow::OpenRequested(record_ref ref)
{
	BFile *file = new BFile(ref);

	get_values();	// Useful if Load() is unsuccessful
	file->GetPath(prefs_path, 256);
	prefs->Load(prefs_path);
	set_values();

	be_app->CloseFilePanel();

	delete file;
}


/*
 *  Save panel done, save prefs requested
 */

void PrefsWindow::SaveRequested(record_ref directory, const char *filename)
{
	BDirectory *dir = new BDirectory(directory);

	dir->GetPath(prefs_path, 256);
	strncat(prefs_path, "/", 256);
	strncat(prefs_path, filename, 256);

	get_values();
	if (prefs->Save(prefs_path))
		CloseSavePanel();
	else
		ShowRequester("Couldn't save preferences.", "Too bad");

	delete dir;
}


/*
 *  Open panel done, get drive path requested
 */

void PrefsWindow::get_drive(int i, record_ref ref)
{
	if (prefs->DriveType[i] == DRVTYPE_DIR) {
		BDirectory *dir = new BDirectory(ref);
		if (dir->Error() == B_NO_ERROR) {
			dir->GetPath(prefs->DrivePath[i], 256);
			set_values();
		}
	} else {
		BFile *file = new BFile(ref);
		if (file->Error() == B_NO_ERROR) {
			file->GetPath(prefs->DrivePath[i], 256);
			set_values();
		}
	}
}


/*
 *  Intercept ESC key (works as clicking the Cancel button)
 */

bool PrefsWindow::FilterKeyDown(ulong *aChar, BView **target)
{
	if (*aChar == B_ESCAPE) {
		// Flash Cancel button
		g_cancel->SetValue(B_CONTROL_ON);
		snooze(100000.0);
		PostMessage(MSG_CANCEL);
	}
	return TRUE;
}
