/*
 *  SAM.h - Simple Assembler and Monitor With Integrated System Explorer
 *
 *  Frodo (C) 1996 Christian Bauer
 */

#include <ctype.h>
#include <stdio.h>
#include <signal.h>

#include "SAM.h"
#include "Types.h"
#include "C64.h"
#include "CPU.h"
#include "VIC.h"
#include "SID.h"
#include "CIA.h"


// Pointers to chips
MOS6510 *TheCPU;
MOS6502_1541 *TheCPU1541;
MOS6569 *TheVIC;
MOS6581 *TheSID;
MOS6526_1 *TheCIA1;
MOS6526_2 *TheCIA2;

// 6510/6502 registers
MOS6510State R64;
MOS6502State R1541;

CBOOL access_1541;	// FALSE: accessing C64, TRUE: accessing 1541

// Access to 6510/6502 address space
inline UBYTE SAMReadByte(UWORD adr)
{
	if (access_1541)
		return TheCPU1541->ExtReadByte(adr);
	else
		return TheCPU->ExtReadByte(adr);
}

inline void SAMWriteByte(UWORD adr, UBYTE byte)
{
	if (access_1541)
		TheCPU1541->ExtWriteByte(adr, byte);
	else
		TheCPU->ExtWriteByte(adr, byte);
}


// Streams for input, output and error messages
FILE *fin, *fout, *ferr;

// Input line
#define INPUT_LENGTH 80
char input[INPUT_LENGTH];
char *in_ptr;

UWORD address, end_address;


// Input tokens
enum Token {
	T_NULL,		// Invalid token
	T_END,		// End of line
	T_NUMBER,	// Hexadecimal number
	T_STRING,	// String enclosed in ""
	T_LPAREN,	// '('
	T_RPAREN,	// ')'
	T_ADD,		// '+'
	T_SUB,		// '-'
	T_MUL,		// '*'
	T_DIV,		// '/'
	T_COMMA,	// ','
	T_IMMED,	// '#'
	T_X,		// 'x'
	T_Y,		// 'y'
	T_PC,		// 'pc'
	T_SP,		// 'sp'

	T_A,		// 'a'	(get_reg_token() only)
	T_DR,		// 'dr'	(get_reg_token() only)
	T_PR		// 'pr'	(get_reg_token() only)
};

enum Token the_token;
UWORD the_number;				// Contains the number if the_token==T_NUMBER
char the_string[INPUT_LENGTH];	// Contains the string if the_token==T_STRING


// Addressing modes
enum {
	A_IMPL,
	A_ACCU,		// A
	A_IMM,		// #zz
	A_REL,		// Branches
	A_ZERO,		// zz
	A_ZEROX,	// zz,x
	A_ZEROY,	// zz,y
	A_ABS,		// zzzz
	A_ABSX,		// zzzz,x
	A_ABSY,		// zzzz,y
	A_IND,		// (zzzz)
	A_INDX,		// (zz,x)
	A_INDY		// (zz),y
};

// Mnemonics
enum {
	M_ADC, M_AND, M_ASL, M_BCC, M_BCS, M_BEQ, M_BIT, M_BMI, M_BNE, M_BPL,
	M_BRK, M_BVC, M_BVS, M_CLC, M_CLD, M_CLI, M_CLV, M_CMP, M_CPX, M_CPY,
	M_DEC, M_DEX, M_DEY, M_EOR, M_INC, M_INX, M_INY, M_JMP, M_JSR, M_LDA,
	M_LDX, M_LDY, M_LSR, M_NOP, M_ORA, M_PHA, M_PHP, M_PLA, M_PLP, M_ROL,
	M_ROR, M_RTI, M_RTS, M_SBC, M_SEC, M_SED, M_SEI, M_STA, M_STX, M_STY,
	M_TAX, M_TAY, M_TSX, M_TXA, M_TXS, M_TYA,

	M_ILLEGAL,  // Undocumented opcodes start here

	M_IANC, M_IANE, M_IARR, M_IASR, M_IDCP, M_IISB, M_IJAM, M_INOP, M_ILAS,
	M_ILAX, M_ILXA, M_IRLA, M_IRRA, M_ISAX, M_ISBC, M_ISBX, M_ISHA, M_ISHS,
	M_ISHX, M_ISHY, M_ISLO, M_ISRE,

	M_MAXIMUM  // Highest element
};

// Mnemonic for each opcode
const char mnemonic[256] = {
	M_BRK , M_ORA , M_IJAM, M_ISLO, M_INOP, M_ORA, M_ASL , M_ISLO,	// 00
	M_PHP , M_ORA , M_ASL , M_IANC, M_INOP, M_ORA, M_ASL , M_ISLO,
	M_BPL , M_ORA , M_IJAM, M_ISLO, M_INOP, M_ORA, M_ASL , M_ISLO,	// 10
	M_CLC , M_ORA , M_INOP, M_ISLO, M_INOP, M_ORA, M_ASL , M_ISLO,
	M_JSR , M_AND , M_IJAM, M_IRLA, M_BIT , M_AND, M_ROL , M_IRLA,	// 20
	M_PLP , M_AND , M_ROL , M_IANC, M_BIT , M_AND, M_ROL , M_IRLA,
	M_BMI , M_AND , M_IJAM, M_IRLA, M_INOP, M_AND, M_ROL , M_IRLA,	// 30
	M_SEC , M_AND , M_INOP, M_IRLA, M_INOP, M_AND, M_ROL , M_IRLA,
	M_RTI , M_EOR , M_IJAM, M_ISRE, M_INOP, M_EOR, M_LSR , M_ISRE,	// 40
	M_PHA , M_EOR , M_LSR , M_IASR, M_JMP , M_EOR, M_LSR , M_ISRE,
	M_BVC , M_EOR , M_IJAM, M_ISRE, M_INOP, M_EOR, M_LSR , M_ISRE,	// 50
	M_CLI , M_EOR , M_INOP, M_ISRE, M_INOP, M_EOR, M_LSR , M_ISRE,
	M_RTS , M_ADC , M_IJAM, M_IRRA, M_INOP, M_ADC, M_ROR , M_IRRA,	// 60
	M_PLA , M_ADC , M_ROR , M_IARR, M_JMP , M_ADC, M_ROR , M_IRRA,
	M_BVS , M_ADC , M_IJAM, M_IRRA, M_INOP, M_ADC, M_ROR , M_IRRA,	// 70
	M_SEI , M_ADC , M_INOP, M_IRRA, M_INOP, M_ADC, M_ROR , M_IRRA,
	M_INOP, M_STA , M_INOP, M_ISAX, M_STY , M_STA, M_STX , M_ISAX,	// 80
	M_DEY , M_INOP, M_TXA , M_IANE, M_STY , M_STA, M_STX , M_ISAX,
	M_BCC , M_STA , M_IJAM, M_ISHA, M_STY , M_STA, M_STX , M_ISAX,	// 90
	M_TYA , M_STA , M_TXS , M_ISHS, M_ISHY, M_STA, M_ISHX, M_ISHA,
	M_LDY , M_LDA , M_LDX , M_ILAX, M_LDY , M_LDA, M_LDX , M_ILAX,	// a0
	M_TAY , M_LDA , M_TAX , M_ILXA, M_LDY , M_LDA, M_LDX , M_ILAX,
	M_BCS , M_LDA , M_IJAM, M_ILAX, M_LDY , M_LDA, M_LDX , M_ILAX,	// b0
	M_CLV , M_LDA , M_TSX , M_ILAS, M_LDY , M_LDA, M_LDX , M_ILAX,
	M_CPY , M_CMP , M_INOP, M_IDCP, M_CPY , M_CMP, M_DEC , M_IDCP,	// c0
	M_INY , M_CMP , M_DEX , M_ISBX, M_CPY , M_CMP, M_DEC , M_IDCP,
	M_BNE , M_CMP , M_IJAM, M_IDCP, M_INOP, M_CMP, M_DEC , M_IDCP,	// d0
	M_CLD , M_CMP , M_INOP, M_IDCP, M_INOP, M_CMP, M_DEC , M_IDCP,
	M_CPX , M_SBC , M_INOP, M_IISB, M_CPX , M_SBC, M_INC , M_IISB,	// e0
	M_INX , M_SBC , M_NOP , M_ISBC, M_CPX , M_SBC, M_INC , M_IISB,
	M_BEQ , M_SBC , M_IJAM, M_IISB, M_INOP, M_SBC, M_INC , M_IISB,	// f0
	M_SED , M_SBC , M_INOP, M_IISB, M_INOP, M_SBC, M_INC , M_IISB
};

// Addressing mode for each opcode
const char adr_mode[256] = {
	A_IMPL, A_INDX, A_IMPL, A_INDX, A_ZERO , A_ZERO , A_ZERO , A_ZERO,	// 00
	A_IMPL, A_IMM , A_ACCU, A_IMM , A_ABS  , A_ABS  , A_ABS  , A_ABS,
	A_REL , A_INDY, A_IMPL, A_INDY, A_ZEROX, A_ZEROX, A_ZEROX, A_ZEROX,	// 10
	A_IMPL, A_ABSY, A_IMPL, A_ABSY, A_ABSX , A_ABSX , A_ABSX , A_ABSX,
	A_ABS , A_INDX, A_IMPL, A_INDX, A_ZERO , A_ZERO , A_ZERO , A_ZERO,	// 20
	A_IMPL, A_IMM , A_ACCU, A_IMM , A_ABS  , A_ABS  , A_ABS  , A_ABS,
	A_REL , A_INDY, A_IMPL, A_INDY, A_ZEROX, A_ZEROX, A_ZEROX, A_ZEROX,	// 30
	A_IMPL, A_ABSY, A_IMPL, A_ABSY, A_ABSX , A_ABSX , A_ABSX , A_ABSX,
	A_IMPL, A_INDX, A_IMPL, A_INDX, A_ZERO , A_ZERO , A_ZERO , A_ZERO,	// 40
	A_IMPL, A_IMM , A_ACCU, A_IMM , A_ABS  , A_ABS  , A_ABS  , A_ABS,
	A_REL , A_INDY, A_IMPL, A_INDY, A_ZEROX, A_ZEROX, A_ZEROX, A_ZEROX,	// 50
	A_IMPL, A_ABSY, A_IMPL, A_ABSY, A_ABSX , A_ABSX , A_ABSX , A_ABSX,
	A_IMPL, A_INDX, A_IMPL, A_INDX, A_ZERO , A_ZERO , A_ZERO , A_ZERO,	// 60
	A_IMPL, A_IMM , A_ACCU, A_IMM , A_IND  , A_ABS  , A_ABS  , A_ABS,
	A_REL , A_INDY, A_IMPL, A_INDY, A_ZEROX, A_ZEROX, A_ZEROX, A_ZEROX,	// 70
	A_IMPL, A_ABSY, A_IMPL, A_ABSY, A_ABSX , A_ABSX , A_ABSX , A_ABSX,
	A_IMM , A_INDX, A_IMM , A_INDX, A_ZERO , A_ZERO , A_ZERO , A_ZERO,	// 80
	A_IMPL, A_IMM , A_IMPL, A_IMM , A_ABS  , A_ABS  , A_ABS  , A_ABS,
	A_REL , A_INDY, A_IMPL, A_INDY, A_ZEROX, A_ZEROX, A_ZEROY, A_ZEROY,	// 90
	A_IMPL, A_ABSY, A_IMPL, A_ABSY, A_ABSX , A_ABSX , A_ABSY , A_ABSY,
	A_IMM , A_INDX, A_IMM , A_INDX, A_ZERO , A_ZERO , A_ZERO , A_ZERO,	// a0
	A_IMPL, A_IMM , A_IMPL, A_IMM , A_ABS  , A_ABS  , A_ABS  , A_ABS,
	A_REL , A_INDY, A_IMPL, A_INDY, A_ZEROX, A_ZEROX, A_ZEROY, A_ZEROY,	// b0
	A_IMPL, A_ABSY, A_IMPL, A_ABSY, A_ABSX , A_ABSX , A_ABSY , A_ABSY,
	A_IMM , A_INDX, A_IMM , A_INDX, A_ZERO , A_ZERO , A_ZERO , A_ZERO,	// c0
	A_IMPL, A_IMM , A_IMPL, A_IMM , A_ABS  , A_ABS  , A_ABS  , A_ABS,
	A_REL , A_INDY, A_IMPL, A_INDY, A_ZEROX, A_ZEROX, A_ZEROX, A_ZEROX,	// d0
	A_IMPL, A_ABSY, A_IMPL, A_ABSY, A_ABSX , A_ABSX , A_ABSX , A_ABSX,
	A_IMM , A_INDX, A_IMM , A_INDX, A_ZERO , A_ZERO , A_ZERO , A_ZERO,	// e0
	A_IMPL, A_IMM , A_IMPL, A_IMM , A_ABS  , A_ABS  , A_ABS  , A_ABS,
	A_REL , A_INDY, A_IMPL, A_INDY, A_ZEROX, A_ZEROX, A_ZEROX, A_ZEROX,	// f0
	A_IMPL, A_ABSY, A_IMPL, A_ABSY, A_ABSX , A_ABSX , A_ABSX , A_ABSX
};

// Chars for each mnemonic
const char mnem_1[] = "aaabbbbbbbbbbcccccccdddeiiijjllllnopppprrrrssssssstttttt?aaaadijnlllrrsssssssss";
const char mnem_2[] = "dnscceimnprvvllllmppeeeonnnmsdddsorhhlloottbeeetttaasxxy?nnrscsaoaaxlrabbhhhhlr";
const char mnem_3[] = "cdlcsqtielkcscdivpxycxyrcxypraxyrpaapaplrisccdiaxyxyxasa?cerrpbmpsxaaaxcxasxyoe";

// Instruction length for each addressing mode
const char adr_length[] = {1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 2, 2};


// Prototypes
void error(char *s);
void init_abort(void);
void exit_abort(void);
void handle_abort(int sig);
CBOOL aborted(void);

void read_line(void);			// Scanner
char get_char(void);
void put_back(char c);
enum Token get_token(void);
enum Token get_reg_token(void);
UWORD get_number(void);
enum Token get_string(char *str);

CBOOL expression(UWORD *number);	// Parser
CBOOL term(UWORD *number);
CBOOL factor(UWORD *number);
CBOOL address_args(void);
CBOOL range_args(int def_range);
CBOOL instr_args(UWORD *number, char *mode);

void help(void);				// Routines for commands
void registers(void);
void display_registers(void);
void memory_dump(void);
void ascii_dump(void);
char conv_from_64(char c);
void screen_dump(void);
char conv_from_scode(char c);
void binary_dump(void);
void sprite_dump(void);
void byte_to_bin(UBYTE byte, char *str);
void disassemble(void);
int disass_line(UWORD adr, UBYTE op, UBYTE lo, UBYTE hi);
void assemble(void);
char find_mnemonic(char op1, char op2, char op3);
CBOOL find_opcode(char mnem, char mode, UBYTE *opcode);
void mem_config(void);
void fill(void);
void compare(void);
void transfer(void);
void modify(void);
void print_expr(void);
void redir_output(void);
void int_vectors(void);
void view_state(void);
void view_cia_state(void);
void dump_cia_ints(UBYTE i);
void view_sid_state(void);
void dump_sid_waveform(UBYTE wave);
void view_vic_state(void);
void dump_spr_flags(UBYTE f);
void dump_vic_ints(UBYTE i);
void view_1541_state(void);
void load_data(void);
void save_data(void);


/*
 *  Open and handle SAM
 */

void SAM(C64 *the_c64)
{
	CBOOL done = FALSE;
	char c;

	TheCPU = the_c64->TheCPU;
	TheCPU1541 = the_c64->TheCPU1541;
	TheVIC = the_c64->TheVIC;
	TheSID = the_c64->TheSID;
	TheCIA1 = the_c64->TheCIA1;
	TheCIA2 = the_c64->TheCIA2;

	// Get CPU registers and current memory configuration
	TheCPU->GetState(&R64);
	TheCPU->ExtConfig = (~R64.ddr | R64.pr) & 7;
	TheCPU1541->GetState(&R1541);

#ifdef AMIGA
	if (!(fin = fout = ferr = fopen("CON:0/0/640/480/SAM", "w+")))
		return;
#else
	fin = stdin;
	fout = stdout;
	ferr = stdout;
#endif

	access_1541 = FALSE;
	address = R64.pc;

	fprintf(ferr, "\n *** SAM - Simple Assembler and Monitor ***\n ***         Press 'h' for help         ***\n\n");
	init_abort();
	display_registers();

	while (!done) {
		if (access_1541)
			fprintf(ferr, "1541> ");
		else
			fprintf(ferr, "C64> ");
		fflush(ferr);
		read_line();
		while ((c = get_char()) == ' ') ;

		switch (c) {
			case 'a':		// Assemble
				get_token();
				assemble();
				break;

			case 'b':		// Binary dump
				get_token();
				binary_dump();
				break;

			case 'c':		// Compare
				get_token();
				compare();
				break;

			case 'd':		// Disassemble
				get_token();
				disassemble();
				break;

			case 'e':       // Interrupt vectors
				int_vectors();
				break;

			case 'f':		// Fill
				get_token();
				fill();
				break;

			case 'h':		// Help
				help();
				break;

			case 'i':		// ASCII dump
				get_token();
				ascii_dump();
				break;

			case 'k':		// Memory configuration
				get_token();
				mem_config();
				break;

			case 'l':		// Load data
				get_token();
				load_data();
				break;

			case 'm':		// Memory dump
				get_token();
				memory_dump();
				break;

			case 'n':		// Screen code dump
				get_token();
				screen_dump();
				break;

			case 'o':		// Redirect output
				get_token();
				redir_output();
				break;

			case 'p':		// Sprite dump
				get_token();
				sprite_dump();
				break;

			case 'r':		// Registers
				get_reg_token();
				registers();
				break;

			case 's':		// Save data
				get_token();
				save_data();
				break;

			case 't':		// Transfer
				get_token();
				transfer();
				break;

			case 'v':		// View machine state
				view_state();
				break;

			case 'x':		// Exit
				done = TRUE;
				break;

			case ':':		// Change memory
				get_token();
				modify();
				break;

			case '1':		// Switch to 1541 mode
				access_1541 = TRUE;
				break;

			case '6':		// Switch to C64 mode
				access_1541 = FALSE;
				break;

			case '?':		// Compute expression
				get_token();
				print_expr();
				break;

			case '\n':		// Blank line
				break;

			default:		// Unknown command
				error("Unknown command");
				break;
		}
	}

	exit_abort();

#ifdef AMIGA
	fclose(fin);
#endif
	if (fout != ferr)
		fclose(fout);

	// Set CPU registers
	TheCPU->SetState(&R64);
	TheCPU1541->SetState(&R1541);
}


/*
 *  Print error message
 */

void error(char *s)
{
	fprintf(ferr, "*** %s\n", s);
}


/*
 *  CTRL-C pressed?
 */

CBOOL WasAborted;
struct sigaction my_sa;

void init_abort(void)
{
	WasAborted = FALSE;
	my_sa.sa_handler = handle_abort;
	my_sa.sa_flags = 0;
	sigemptyset(&my_sa.sa_mask);
	sigaction(SIGINT, &my_sa, NULL);
}

void exit_abort(void)
{
	my_sa.sa_handler = SIG_DFL;
	sigaction(SIGINT, &my_sa, NULL);
}

#ifdef __beos__
void handle_abort(int sig, void *data, struct vregs *regs)
#else
void handle_abort(int sig)
#endif
{
	WasAborted = TRUE;
}

CBOOL aborted(void)
{
	CBOOL ret = WasAborted;

	WasAborted = FALSE;
	return ret;
}


/*
 *  Read a line from the keyboard
 */

void read_line(void)
{
	fgets(in_ptr = input, INPUT_LENGTH, fin);
}


/*
 *  Read a character from the input line
 */

char get_char(void)
{
	return *in_ptr++;
}


/*
 *  Stuff back a character into the input line
 */

void put_back(char c)
{
	*(--in_ptr) = c;
}


/*
 *  Scanner: Get a token from the input line
 */

enum Token get_token(void)
{
	char c;

	// Skip spaces
	while ((c = get_char()) == ' ') ;

	switch (tolower(c)) {
		case '\n':
			return the_token = T_END;
		case '(':
			return the_token = T_LPAREN;
		case ')':
			return the_token = T_RPAREN;
		case '+':
			return the_token = T_ADD;
		case '-':
			return the_token = T_SUB;
		case '*':
			return the_token = T_MUL;
		case '/':
			return the_token = T_DIV;
		case ',':
			return the_token = T_COMMA;
		case '#':
			return the_token = T_IMMED;
		case 'x':
			return the_token = T_X;
		case 'y':
			return the_token = T_Y;
		case 'p':
			if (tolower(get_char()) == 'c')
				return the_token = T_PC;
			else {
				error("Unrecognized token");
				return the_token = T_NULL;
			}
		case 's':
			if (tolower(get_char()) == 'p')
				return the_token = T_SP;
			else {
				error("Unrecognized token");
				return the_token = T_NULL;
			}
		case '0': case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
		case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
			put_back(c);
			the_number = get_number();
			return the_token = T_NUMBER;
		case '"':
			return the_token = get_string(the_string);
		default:
			error("Unrecognized token");
			return the_token = T_NULL;
	}
}

enum Token get_reg_token(void)
{
	char c;

	// Skip spaces
	while ((c = get_char()) == ' ') ;

	switch (tolower(c)) {
		case '\n':
			return the_token = T_END;
		case 'a':
			return the_token = T_A;
		case 'd':
			if (tolower(get_char()) == 'r')
				return the_token = T_DR;
			else {
				error("Unrecognized token");
				return the_token = T_NULL;
			}
		case 'p':
			if ((c = tolower(get_char())) == 'c')
				return the_token = T_PC;
			else if (c == 'r')
				return the_token = T_PR;
			else {
				error("Unrecognized token");
				return the_token = T_NULL;
			}
		case 's':
			if (tolower(get_char()) == 'p')
				return the_token = T_SP;
			else {
				error("Unrecognized token");
				return the_token = T_NULL;
			}
		case 'x':
			return the_token = T_X;
		case 'y':
			return the_token = T_Y;
		default:
			error("Unrecognized token");
			return the_token = T_NULL;
	}
}

UWORD get_number(void)
{
	char c;
	UWORD i = 0;

	while (((c = tolower(get_char())) >= '0') && (c <= '9') || (c >= 'a') && (c <= 'f'))
		if (c < 'a')
			i = (i << 4) + (c - '0');
		else
			i = (i << 4) + (c - 'a' + 10);

	put_back(c);
	return i;
}

enum Token get_string(char *str)
{
	char c;

	while ((c = get_char()) != '\n') {
		if (c == '"') {
			*str = 0;
			return T_STRING;
		}
		*str++ = c;
	}

	error("Unterminated string");
	return T_NULL;
}


/*
 *  expression = term {(ADD | SUB) term}
 *  TRUE: OK, FALSE: Error
 */

CBOOL expression(UWORD *number)
{
	UWORD accu, trm;

	if (!term(&accu))
		return FALSE;

	for (;;)
		switch (the_token) {
			case T_ADD:
				get_token();
				if (!term(&trm))
					return FALSE;
				accu += trm;
				break;

			case T_SUB:
				get_token();
				if (!term(&trm))
					return FALSE;
				accu -= trm;
				break;

			default:
				*number = accu;
				return TRUE;
		}
}


/*
 *  term = factor {(MUL | DIV) factor}
 *  TRUE: OK, FALSE: Error
 */

CBOOL term(UWORD *number)
{
	UWORD accu, fact;

	if (!factor(&accu))
		return FALSE;

	for (;;)
		switch (the_token) {
			case T_MUL:
				get_token();
				if (!factor(&fact))
					return FALSE;
				accu *= fact;
				break;

			case T_DIV:
				get_token();
				if (!factor(&fact))
					return FALSE;
				if (fact == 0) {
					error("Division by 0");
					return FALSE;
				}
				accu /= fact;
				break;

			default:
				*number = accu;
				return TRUE;
		}
}


/*
 *  factor = NUMBER | PC | SP | LPAREN expression RPAREN
 *  TRUE: OK, FALSE: Error
 */

CBOOL factor(UWORD *number)
{
	switch (the_token) {
		case T_NUMBER:
			*number = the_number;
			get_token();
			return TRUE;

		case T_PC:
			get_token();
			*number = access_1541 ? R1541.pc : R64.pc;
			return TRUE;

		case T_SP:
			get_token();
			*number = access_1541 ? R1541.sp : R64.sp;
			return TRUE;

		case T_LPAREN:
			get_token();
			if (expression(number))
				if (the_token == T_RPAREN) {
					get_token();
					return TRUE;
				} else {
					error("Missing ')'");
					return FALSE;
				}
			else {
				error("Error in expression");
				return FALSE;
			}

		case T_END:
			error("Required argument missing");
			return FALSE;

		default:
			error("'pc', 'sp', '(' or number expected");
			return FALSE;
	}
}


/*
 *  address_args = [expression] END
 *
 *  Read start to "address"
 *
 *  TRUE: OK, FALSE: Error
 */

CBOOL address_args(void)
{
	if (the_token == T_END)
		return TRUE;
	else {
		if (!expression(&address))
			return FALSE;
		return the_token == T_END;
	}
}


/*
 *  range_args = [expression] [[COMMA] expression] END
 *
 *  Read start address to "address", end address to "end_address"
 *
 *  TRUE: OK, FALSE: Error
 */

CBOOL range_args(int def_range)
{
	end_address = address + def_range;

	if (the_token == T_END)
		return TRUE;
	else {
		if (!expression(&address))
			return FALSE;
		end_address = address + def_range;
		if (the_token == T_END)
			return TRUE;
		else {
			if (the_token == T_COMMA) get_token();
			if (!expression(&end_address))
				return FALSE;
			return the_token == T_END;
		}
	}
}


/*
 *  instr_args = END
 *             | IMMED NUMBER END
 *             | NUMBER [COMMA (X | Y)] END
 *             | LPAREN NUMBER (RPAREN [COMMA Y] | COMMA X RPAREN) END
 *
 *  Read arguments of a 6510 instruction, determine address and addressing mode
 *
 *  TRUE: OK, FALSE: Error
 */

CBOOL instr_args(UWORD *number, char *mode)
{
	switch (the_token) {

		case T_END:
			*mode = A_IMPL;
			return TRUE;

		case T_IMMED:
			get_token();
			if (the_token == T_NUMBER) {
				*number = the_number;
				*mode = A_IMM;
				get_token();
				return the_token == T_END;
			} else {
				error("Number expected");
				return FALSE;
			}

		case T_NUMBER:
			*number = the_number;
			get_token();
			switch (the_token) {

				case T_END:
					if (*number < 0x100)
						*mode = A_ZERO;
					else
						*mode = A_ABS;
					return TRUE;

				case T_COMMA:
					get_token();
					switch (the_token) {

						case T_X:
							get_token();
							if (*number < 0x100)
								*mode = A_ZEROX;
							else
								*mode = A_ABSX;
							return the_token == T_END;

						case T_Y:
							get_token();
							if (*number < 0x100)
								*mode = A_ZEROY;
							else
								*mode = A_ABSY;
							return the_token == T_END;

						default:
							error("Illegal index register");
							return FALSE;
					}

				default:
					return FALSE;
			}

		case T_LPAREN:
			get_token();
			if (the_token == T_NUMBER) {
				*number = the_number;
				get_token();
				switch (the_token) {

					case T_RPAREN:
						get_token();
						switch (the_token) {

							case T_END:
								*mode = A_IND;
								return TRUE;

							case T_COMMA:
								get_token();
								if (the_token == T_Y) {
									*mode = A_INDY;
									get_token();
									return the_token == T_END;
								} else {
									error("Only 'y' index register allowed");
									return FALSE;
								}

							default:
								error("Illegal characters after ')'");
								return FALSE;
						}

					case T_COMMA:
						get_token();
						if (the_token == T_X) {
							get_token();
							if (the_token == T_RPAREN) {
								*mode = A_INDX;
								get_token();
								return the_token == T_END;
							} else {
								error("')' expected");
								return FALSE;
							}
						} else {
							error("Only 'x' index register allowed");
							return FALSE;
						}

					default:
						error("')' or ',' expected");
						return FALSE;
				}
			} else {
				error("Number expected");
				return FALSE;
			}

		default:
			error("'(', '#' or number expected");
			return FALSE;
	}
}


/*
 *  Display help
 *  ?
 */

void help(void)
{
	fprintf(fout, "a [start]           Assemble\n"
				"b [start] [end]     Binary dump\n"
				"c start end dest    Compare memory\n"
				"d [start] [end]     Disassemble\n"
				"e                   Show interrupt vectors\n"
				"f start end byte    Fill memory\n"
				"i [start] [end]     ASCII/PETSCII dump\n"
				"k [config]          Show/set C64 memory configuration\n"
				"l start \"file\"      Load data\n"
				"m [start] [end]     Memory dump\n"
				"n [start] [end]     Screen code dump\n"
				"o [\"file\"]          Redirect output\n"
				"p [start] [end]     Sprite dump\n"
				"r [reg value]       Show/set CPU registers\n"
				"s start end \"file\"  Save data\n"
				"t start end dest    Transfer memory\n"
				"vc1                 View CIA 1 state\n"
				"vc2                 View CIA 2 state\n"
				"vf                  View 1541 state\n"
				"vs                  View SID state\n"
				"vv                  View VIC state\n"
				"x                   Return to Frodo\n"
				": addr {byte}       Modify memory\n"
				"1541                Switch to 1541\n"
				"64                  Switch to C64\n"
				"? expression        Calculate expression\n");
}


/*
 *  Display/change 6510 registers
 *  r [reg value]
 */

void registers(void)
{
	enum Token the_reg;
	UWORD value;

	if (the_token != T_END)
		switch (the_reg = the_token) {
			case T_A:
			case T_X:
			case T_Y:
			case T_PC:
			case T_SP:
			case T_DR:
			case T_PR:
				get_token();
				if (!expression(&value))
					return;

				switch (the_reg) {
					case T_A:
						if (access_1541)
							R1541.a = value;
						else
							R64.a = value;
						break;
					case T_X:
						if (access_1541)
							R1541.x = value;
						else
							R64.x = value;
						break;
					case T_Y:
						if (access_1541)
							R1541.y = value;
						else
							R64.y = value;
						break;
					case T_PC:
						if (access_1541)
							R1541.pc = value;
						else
							R64.pc = value;
						break;
					case T_SP:
						if (access_1541)
							R1541.sp = (value & 0xff) | 0x0100;
						else
							R64.sp = (value & 0xff) | 0x0100;
						break;
					case T_DR:
						if (!access_1541)
							R64.ddr = value;
						break;
					case T_PR:
						if (!access_1541)
							R64.pr = value;
						break;
				}
				break;

			default:
				return;
		}

	display_registers();
}

void display_registers(void)
{
	if (access_1541) {
		fprintf(fout, " PC  A  X  Y   SP  NVDIZC  Instruction\n");
		fprintf(fout, "%04lx %02lx %02lx %02lx %04lx %c%c%c%c%c%c ",
			R1541.pc, R1541.a, R1541.x, R1541.y, R1541.sp,
			R1541.p & 0x80 ? '1' : '0', R1541.p & 0x40 ? '1' : '0', R1541.p & 0x08 ? '1' : '0',
			R1541.p & 0x04 ? '1' : '0', R1541.p & 0x02 ? '1' : '0', R1541.p & 0x01 ? '1' : '0');
		disass_line(R1541.pc, SAMReadByte(R1541.pc), SAMReadByte(R1541.pc+1), SAMReadByte(R1541.pc+2));
	} else {
		fprintf(fout, " PC  A  X  Y   SP  DR PR NVDIZC  Instruction\n");
		fprintf(fout, "%04lx %02lx %02lx %02lx %04lx %02lx %02lx %c%c%c%c%c%c ",
			R64.pc, R64.a, R64.x, R64.y, R64.sp, R64.ddr, R64.pr,
			R64.p & 0x80 ? '1' : '0', R64.p & 0x40 ? '1' : '0', R64.p & 0x08 ? '1' : '0',
			R64.p & 0x04 ? '1' : '0', R64.p & 0x02 ? '1' : '0', R64.p & 0x01 ? '1' : '0');
		disass_line(R64.pc, SAMReadByte(R64.pc), SAMReadByte(R64.pc+1), SAMReadByte(R64.pc+2));
	}
}


/*
 *  Memory dump
 *  m [start] [end]
 */

#define MEMDUMP_BPL 16  // Bytes per line

void memory_dump(void)
{
	CBOOL done = FALSE;
	short i;
	UBYTE mem[MEMDUMP_BPL + 2];
	UBYTE byte;

	mem[MEMDUMP_BPL] = 0;

	if (!range_args(16 * MEMDUMP_BPL - 1))  // 16 lines unless end address specified
		return;

	do {
		fprintf(fout, "%04lx:", address);
		for (i=0; i<MEMDUMP_BPL; i++, address++) {
			if (address == end_address) done = TRUE;

			fprintf(fout, " %02lx", byte = SAMReadByte(address));
			if ((byte >= ' ') && (byte <= '~'))
				mem[i] = conv_from_64(byte);
			else
				mem[i] = '.';
		}
		fprintf(fout, "  '%s'\n", mem);
	} while (!done && !aborted());
}


/*
 *  ASCII dump
 *  i [start] [end]
 */

#define ASCIIDUMP_BPL 64  // Bytes per line

void ascii_dump(void)
{
	CBOOL done = FALSE;
	short i;
	UBYTE mem[ASCIIDUMP_BPL + 2];
	UBYTE byte;

	mem[ASCIIDUMP_BPL] = 0;

	if (!range_args(16 * ASCIIDUMP_BPL - 1))  // 16 lines unless end address specified
		return;

	do {
		fprintf(fout, "%04lx:", address);
		for (i=0; i<ASCIIDUMP_BPL; i++, address++) {
			if (address == end_address) done = TRUE;

			byte = SAMReadByte(address);
			if ((byte >= ' ') && (byte <= '~'))
				mem[i] = conv_from_64(byte);
			else
				mem[i] = '.';
		}
		fprintf(fout, " '%s'\n", mem);
	} while (!done && !aborted());
}


/*
 *  Convert PETSCII->ASCII
 */

char conv_from_64(char c)
{
	if ((c >= 'A') && (c <= 'Z') || (c >= 'a') && (c <= 'z'))
		return c ^ 0x20;
	else
		return c;
}


/*
 *  Screen code dump
 *  n [start] [end]
 */

#define SCRDUMP_BPL 64  // Bytes per line

void screen_dump(void)
{
	CBOOL done = FALSE;
	short i;
	UBYTE mem[SCRDUMP_BPL + 2];
	UBYTE byte;

	mem[SCRDUMP_BPL] = 0;

	if (!range_args(16 * SCRDUMP_BPL - 1))  // 16 Zeilen unless end address specified
		return;

	do {
		fprintf(fout, "%04lx:", address);
		for (i=0; i<SCRDUMP_BPL; i++, address++) {
			if (address == end_address) done = TRUE;

			byte = SAMReadByte(address);
			if (byte < 90)
				mem[i] = conv_from_scode(byte);
			else
				mem[i] = '.';
		}
		fprintf(fout, " '%s'\n", mem);
	} while (!done && !aborted());
}


/*
 *  Convert screen code->ASCII
 */

char conv_from_scode(char c)
{
	c &= 0x7f;

	if (c <= 31)
		return c + 64;
	else
		if (c >= 64)
			return c + 32;
		else
			return c;
}


/*
 *  Binary dump
 *  b [start] [end]
 */

void binary_dump(void)
{
	CBOOL done = FALSE;
	char bin[10];

	bin[8] = 0;

	if (!range_args(7))  // 8 lines unless end address specified
		return;

	do {
		if (address == end_address) done = TRUE;

		byte_to_bin(SAMReadByte(address), bin);
		fprintf(fout, "%04lx: %s\n", address++, bin);
	} while (!done && !aborted());
}


/*
 *  Sprite data dump
 *  p [start] [end]
 */

void sprite_dump(void)
{
	CBOOL done = FALSE;
	short i;
	char bin[10];

	bin[8] = 0;

	if (!range_args(21 * 3 - 1))  // 21 lines unless end address specified
		return;

	do {
		fprintf(fout, "%04lx: ", address);
		for (i=0; i<3; i++, address++) {
			if (address == end_address) done = TRUE;

			byte_to_bin(SAMReadByte(address), bin);
			fprintf(fout, "%s", bin);
		}
		fputc('\n', fout);
	} while (!done && !aborted());
}


/*
 *  Convert byte to binary representation
 */

void byte_to_bin(UBYTE byte, char *str)
{
	short i;

	for (i=0; i<8; i++, byte<<=1)
		if (byte & 0x80)
			str[i] = '#';
		else
			str[i] = '.';
}


/*
 *  Disassemble
 *  d [start] [end]
 */

void disassemble(void)
{
	CBOOL done = FALSE;
	short i;
	UBYTE op[3];
	UWORD adr;

	if (!range_args(31))  // 32 bytes unless end address specified
		return;

	do {
		fprintf(fout, "%04lx:", adr = address);
		for (i=0; i<3; i++, adr++) {
			if (adr == end_address) done = TRUE;
			op[i] = SAMReadByte(adr);
		}
		address += disass_line(address, op[0], op[1], op[2]);
	} while (!done && !aborted());
}


/*
 *  Disassemble one instruction, return length
 */

int disass_line(UWORD adr, UBYTE op, UBYTE lo, UBYTE hi)
{
	char mode = adr_mode[op], mnem = mnemonic[op];

	// Display instruction bytes in hex
	switch (adr_length[mode]) {
		case 1:
			fprintf(fout, " %02lx       ", op);
			break;

		case 2:
			fprintf(fout, " %02lx %02lx    ", op, lo);
			break;

		case 3:
			fprintf(fout, " %02lx %02lx %02lx ", op, lo, hi);
			break;
	}

	// Tag undocumented opcodes with an asterisk
	if (mnem > M_ILLEGAL)
		fputc('*', fout);
	else
		fputc(' ', fout);

	// Print mnemonic
	fprintf(fout, "%c%c%c ", mnem_1[mnem], mnem_2[mnem], mnem_3[mnem]);

	// Pring argument
	switch (mode) {
		case A_IMPL:
			break;

		case A_ACCU:
			fprintf(fout, "a");
			break;

		case A_IMM:
			fprintf(fout, "#%02lx", lo);
			break;

		case A_REL:
			fprintf(fout, "%04lx", ((adr + 2) + (BYTE)lo) & 0xffff);
			break;

		case A_ZERO:
			fprintf(fout, "%02lx", lo);
			break;

		case A_ZEROX:
			fprintf(fout, "%02lx,x", lo);
			break;

		case A_ZEROY:
			fprintf(fout, "%02lx,y", lo);
			break;

		case A_ABS:
			fprintf(fout, "%04lx", (hi << 8) | lo);
			break;

		case A_ABSX:
			fprintf(fout, "%04lx,x", (hi << 8) | lo);
			break;

		case A_ABSY:
			fprintf(fout, "%04lx,y", (hi << 8) | lo);
			break;

		case A_IND:
			fprintf(fout, "(%04lx)", (hi << 8) | lo);
			break;

		case A_INDX:
			fprintf(fout, "(%02lx,x)", lo);
			break;

		case A_INDY:
			fprintf(fout, "(%02lx),y", lo);
			break;
	}

	fputc('\n', fout);
	return adr_length[mode];
}


/*
 *  Assemble
 *  a [start]
 */

void assemble(void)
{
	CBOOL done = FALSE;
	char c1, c2, c3;
	char mnem, mode;
	UBYTE opcode;
	UWORD arg;
	WORD rel;

	// Read parameters
	if (!address_args())
		return;

	do {
		fprintf(fout, "%04lx> ", address);
		read_line();

		c1 = tolower(get_char());
		c2 = tolower(get_char());
		c3 = tolower(get_char());

		if (c1 != '\n') {

			if ((mnem = find_mnemonic(c1, c2, c3)) != M_ILLEGAL) {

				get_token();
				if (instr_args(&arg, &mode)) {

					// Convert A_IMPL -> A_ACCU if necessary
					if ((mode == A_IMPL) && find_opcode(mnem, A_ACCU, &opcode))
						mode = A_ACCU;

					// Handle relative addressing seperately
					if (((mode == A_ABS) || (mode == A_ZERO)) && find_opcode(mnem, A_REL, &opcode)) {
						mode = A_REL;
						rel = arg - (address + 2) & 0xffff;
						if ((rel < -128) || (rel > 127)) {
							error("Branch too long");
							continue;
						} else
							arg = rel & 0xff;
					}

					if (find_opcode(mnem, mode, &opcode)) {

						// Print disassembled line
						fprintf(fout, "\v%04lx:", address);
						disass_line(address, opcode, arg & 0xff, arg >> 8);

						switch (adr_length[mode]) {
							case 1:
								SAMWriteByte(address++, opcode);
								break;

							case 2:
								SAMWriteByte(address++, opcode);
								SAMWriteByte(address++, arg);
								break;

							case 3:
								SAMWriteByte(address++, opcode);
								SAMWriteByte(address++, arg & 0xff);
								SAMWriteByte(address++, arg >> 8);
								break;

							default:
								error("Internal error");
								break;
						}

					} else
						error("Addressing mode not supported by instruction");

				} else
					error("Unrecognized addressing mode");

			} else
				error("Unknown instruction");

		} else			// Input is terminated with a blank line
			done = TRUE;
	} while (!done);
}


/*
 *  Find mnemonic code to three letters
 *  M_ILLEGAL: No matching mnemonic found
 */

char find_mnemonic(char op1, char op2, char op3)
{
	int i;

	for (i=0; i<M_MAXIMUM; i++)
		if ((mnem_1[i] == op1) && (mnem_2[i] == op2) && (mnem_3[i] == op3))
			return i;

	return M_ILLEGAL;
}


/*
 *  Determine opcode of an instruction given mnemonic and addressing mode
 *  TRUE: OK, FALSE: Mnemonic can't have specified addressing mode
 */

CBOOL find_opcode(char mnem, char mode, UBYTE *opcode)
{
	int i;

	for (i=0; i<256; i++)
		if ((mnemonic[i] == mnem) && (adr_mode[i] == mode)) {
			*opcode = i;
			return TRUE;
		}

	return FALSE;
}


/*
 *  Show/set memory configuration
 *  k [config]
 */

void mem_config(void)
{
	UWORD con;

	if (the_token != T_END)
		if (!expression(&con))
			return;
		else
			TheCPU->ExtConfig = con;
	else
		con = TheCPU->ExtConfig;

	fprintf(fout, "Configuration: %ld\n", con & 7);
	fprintf(fout, "A000-BFFF: %s\n", (con & 3) == 3 ? "Basic" : "RAM");
	fprintf(fout, "D000-DFFF: %s\n", (con & 3) ? ((con & 4) ? "I/O" : "Char") : "RAM");
	fprintf(fout, "E000-FFFF: %s\n", (con & 2) ? "Kernal" : "RAM");
}


/*
 *  Fill
 *  k start end byte
 */

void fill(void)
{
	CBOOL done = FALSE;
	UWORD adr, end_adr, value;

	if (!expression(&adr))
		return;
	if (!expression(&end_adr))
		return;
	if (!expression(&value))
		return;

	do {
		if (adr == end_adr) done = TRUE;

		SAMWriteByte(adr++, value);
	} while (!done);
}


/*
 *  Compare
 *  c start end dest
 */

void compare(void)
{
	CBOOL done = FALSE;
	UWORD adr, end_adr, dest;
	int num = 0;

	if (!expression(&adr))
		return;
	if (!expression(&end_adr))
		return;
	if (!expression(&dest))
		return;

	do {
		if (adr == end_adr) done = TRUE;

		if (SAMReadByte(adr) != SAMReadByte(dest)) {
			fprintf(fout, "%04lx ", adr);
			num++;
			if (!(num & 7))
				fputc('\n', fout);
		}
		adr++; dest++;
	} while (!done && !aborted());

	if (num & 7)
		fputc('\n', fout);
	fprintf(fout, "%ld byte(s) different\n", num);
}


/*
 *  Copy
 *  t start end dest
 */

void transfer(void)
{
	CBOOL done = FALSE;
	UWORD adr, end_adr, dest;

	if (!expression(&adr))
		return;
	if (!expression(&end_adr))
		return;
	if (!expression(&dest))
		return;

	if (dest < adr)
		do {
			if (adr == end_adr) done = TRUE;
			SAMWriteByte(dest++,SAMReadByte(adr++));
		} while (!done);
	else {
		dest += end_adr - adr;
		do {
			if (adr == end_adr) done = TRUE;
			SAMWriteByte(dest--,SAMReadByte(end_adr--));
		} while (!done);
	}
}


/*
 *  Change memory
 *  : addr {byte}
 */

void modify(void)
{
	UWORD adr, val;

	if (!expression(&adr))
		return;

	while (the_token != T_END)
		if (expression(&val))
			SAMWriteByte(adr++, val);
		else
			return;
}


/*
 *  Compute and display expression
 *  ? expression
 */

void print_expr(void)
{
	UWORD val;

	if (!expression(&val))
		return;

	fprintf(fout, "Hex: %04lx\nDec: %lu\n", val, val);
}


/*
 *  Redirect output
 *  o [file]
 */

void redir_output(void)
{
	// Close old file
	if (fout != ferr) {
		fclose(fout);
		fout = ferr;
		return;
	}

	// No argument given?
	if (the_token == T_END)
		return;

	// Otherwise open file
	if (the_token == T_STRING) {
		if (!(fout = fopen(the_string, "w")))
			error("Unable to open file");
	} else
		error("'\"' around file name expected");
}


/*
 *  Display interrupt vectors
 */

void int_vectors(void)
{
	fprintf(fout, "        IRQ  BRK  NMI\n");
	fprintf(fout, "%d  : %04lx %04lx %04lx\n",
		access_1541 ? 6502 : 6510,
		SAMReadByte(0xffff) << 8 | SAMReadByte(0xfffe),
		SAMReadByte(0xffff) << 8 | SAMReadByte(0xfffe),
		SAMReadByte(0xfffb) << 8 | SAMReadByte(0xfffa));

	if (!access_1541 && TheCPU->ExtConfig & 2)
		fprintf(fout, "Kernal: %04lx %04lx %04lx\n",
			SAMReadByte(0x0315) << 8 | SAMReadByte(0x0314),
			SAMReadByte(0x0317) << 8 | SAMReadByte(0x0316),
			SAMReadByte(0x0319) << 8 | SAMReadByte(0x0318));
}


/*
 *  Display state of custom chips
 */

void view_state(void)
{
	switch (get_char()) {
		case 'c':		// CIA
			view_cia_state();
			break;

		case 's':		// SID
			view_sid_state();
			break;

		case 'v':		// VIC
			view_vic_state();
			break;

		case 'f':		// Floppy
			view_1541_state();
			break;

		default:
			error("Unknown command");
			break;
	}
}

void view_cia_state(void)
{
	MOS6526State cs;

	switch (get_char()) {
		case '1':
			TheCIA1->GetState(&cs);
			break;
		case '2':
			TheCIA2->GetState(&cs);
			break;
		default:
			error("Unknown command");
			return;
	}

	fprintf(fout, "Timer A  : %s\n", cs.cra & 1 ? "On" : "Off");
	fprintf(fout, " Counter : %04lx  Latch: %04lx\n", (cs.ta_hi << 8) | cs.ta_lo, cs.latcha);
	fprintf(fout, " Run mode: %s\n", cs.cra & 8 ? "One-shot" : "Continuous");
	fprintf(fout, " Input   : %s\n", cs.cra & 0x20 ? "CNT" : "Phi2");
	fprintf(fout, " Output  : ");
	if (cs.cra & 2)
		if (cs.cra & 4)
			fprintf(fout, "PB6 Toggle\n\n");
		else
			fprintf(fout, "PB6 Pulse\n\n");
	else
		fprintf(fout, "None\n\n");

	fprintf(fout, "Timer B  : %s\n", cs.crb & 1 ? "On" : "Off");
	fprintf(fout, " Counter : %04lx  Latch: %04lx\n", (cs.tb_hi << 8) | cs.tb_lo, cs.latchb);
	fprintf(fout, " Run mode: %s\n", cs.crb & 8 ? "One-shot" : "Continuous");
	fprintf(fout, " Input   : ");
	if (cs.crb & 0x40)
		if (cs.crb & 0x20)
			fprintf(fout, "Timer A underflow (CNT high)\n");
		else
			fprintf(fout, "Timer A underflow\n");
	else
		if (cs.crb & 0x20)
			fprintf(fout, "CNT\n");
		else
			fprintf(fout, "Phi2\n");
	fprintf(fout, " Output  : ");
	if (cs.crb & 2)
		if (cs.crb & 4)
			fprintf(fout, "PB7 Toggle\n\n");
		else
			fprintf(fout, "PB7 Pulse\n\n");
	else
		fprintf(fout, "None\n\n");

	fprintf(fout, "TOD         : %lx%lx:%lx%lx:%lx%lx.%lx %s\n",
		(cs.tod_hr >> 4) & 1, cs.tod_hr & 0x0f,
		(cs.tod_min >> 4) & 7, cs.tod_min & 0x0f,
		(cs.tod_sec >> 4) & 7, cs.tod_sec & 0x0f,
		cs.tod_10ths & 0x0f, cs.tod_hr & 0x80 ? "PM" : "AM");
	fprintf(fout, "Alarm       : %lx%lx:%lx%lx:%lx%lx.%lx %s\n",
		(cs.alm_hr >> 4) & 1, cs.alm_hr & 0x0f,
		(cs.alm_min >> 4) & 7, cs.alm_min & 0x0f,
		(cs.alm_sec >> 4) & 7, cs.alm_sec & 0x0f,
		cs.alm_10ths & 0x0f, cs.alm_hr & 0x80 ? "PM" : "AM");
	fprintf(fout, "TOD input   : %s\n", cs.cra & 0x80 ? "50Hz" : "60Hz");
	fprintf(fout, "Write to    : %s registers\n\n", cs.crb & 0x80 ? "Alarm" : "TOD");

	fprintf(fout, "Serial data : %02lx\n", cs.sdr);
	fprintf(fout, "Serial mode : %s\n\n", cs.cra & 0x40 ? "Output" : "Input");

	fprintf(fout, "Pending int.: ");
	dump_cia_ints(cs.int_data);
	fprintf(fout, "Enabled int.: ");
	dump_cia_ints(cs.int_mask);
}

void dump_cia_ints(UBYTE i)
{
	if (i & 0x1f) {
		if (i & 1) fprintf(fout, "TA ");
		if (i & 2) fprintf(fout, "TB ");
		if (i & 4) fprintf(fout, "Alarm ");
		if (i & 8) fprintf(fout, "Serial ");
		if (i & 0x10) fprintf(fout, "Flag");
	} else
		fprintf(fout, "None");
	fputc('\n', fout);
}

void view_sid_state(void)
{
	MOS6581State ss;

	TheSID->GetState(&ss);

	fprintf(fout, "Voice 1\n");
	fprintf(fout, " Frequency  : %04lx\n", (ss.freq_hi_1 << 8) | ss.freq_lo_1);
	fprintf(fout, " Pulse Width: %04lx\n", ((ss.pw_hi_1 & 0x0f) << 8) | ss.pw_lo_1);
	fprintf(fout, " Env. (ADSR): %lx %lx %lx %lx\n", ss.AD_1 >> 4, ss.AD_1 & 0x0f, ss.SR_1 >> 4, ss.SR_1 & 0x0f);
	fprintf(fout, " Waveform   : ");
	dump_sid_waveform(ss.ctrl_1);
	fprintf(fout, " Gate       : %s  Ring mod.: %s\n", ss.ctrl_1 & 0x01 ? "On " : "Off", ss.ctrl_1 & 0x04 ? "On" : "Off");
	fprintf(fout, " Test bit   : %s  Synchron.: %s\n", ss.ctrl_1 & 0x08 ? "On " : "Off", ss.ctrl_1 & 0x02 ? "On" : "Off");
	fprintf(fout, " Filter     : %s\n", ss.res_filt & 0x01 ? "On" : "Off");

	fprintf(fout, "\nVoice 2\n");
	fprintf(fout, " Frequency  : %04lx\n", (ss.freq_hi_2 << 8) | ss.freq_lo_2);
	fprintf(fout, " Pulse Width: %04lx\n", ((ss.pw_hi_2 & 0x0f) << 8) | ss.pw_lo_2);
	fprintf(fout, " Env. (ADSR): %lx %lx %lx %lx\n", ss.AD_2 >> 4, ss.AD_2 & 0x0f, ss.SR_2 >> 4, ss.SR_2 & 0x0f);
	fprintf(fout, " Waveform   : ");
	dump_sid_waveform(ss.ctrl_2);
	fprintf(fout, " Gate       : %s  Ring mod.: %s\n", ss.ctrl_2 & 0x01 ? "On " : "Off", ss.ctrl_2 & 0x04 ? "On" : "Off");
	fprintf(fout, " Test bit   : %s  Synchron.: %s\n", ss.ctrl_2 & 0x08 ? "On " : "Off", ss.ctrl_2 & 0x02 ? "On" : "Off");
	fprintf(fout, " Filter     : %s\n", ss.res_filt & 0x02 ? "On" : "Off");

	fprintf(fout, "\nVoice 3\n");
	fprintf(fout, " Frequency  : %04lx\n", (ss.freq_hi_3 << 8) | ss.freq_lo_3);
	fprintf(fout, " Pulse Width: %04lx\n", ((ss.pw_hi_3 & 0x0f) << 8) | ss.pw_lo_3);
	fprintf(fout, " Env. (ADSR): %lx %lx %lx %lx\n", ss.AD_3 >> 4, ss.AD_3 & 0x0f, ss.SR_3 >> 4, ss.SR_3 & 0x0f);
	fprintf(fout, " Waveform   : ");
	dump_sid_waveform(ss.ctrl_3);
	fprintf(fout, " Gate       : %s  Ring mod.: %s\n", ss.ctrl_3 & 0x01 ? "On " : "Off", ss.ctrl_3 & 0x04 ? "On" : "Off");
	fprintf(fout, " Test bit   : %s  Synchron.: %s\n", ss.ctrl_3 & 0x08 ? "On " : "Off", ss.ctrl_3 & 0x02 ? "On" : "Off");
	fprintf(fout, " Filter     : %s  Mute     : %s\n", ss.res_filt & 0x04 ? "On" : "Off", ss.mode_vol & 0x80 ? "Yes" : "No");

	fprintf(fout, "\nFilters/Volume\n");
	fprintf(fout, " Frequency: %04lx\n", (ss.fc_hi << 3) | (ss.fc_lo & 0x07));
	fprintf(fout, " Resonance: %lx\n", ss.res_filt >> 4);
	fprintf(fout, " Mode     : ");
	if (ss.mode_vol & 0x70) {
		if (ss.mode_vol & 0x10) fprintf(fout, "Low-pass ");
		if (ss.mode_vol & 0x20) fprintf(fout, "Band-pass ");
		if (ss.mode_vol & 0x40) fprintf(fout, "High-pass");
	} else
		fprintf(fout, "None");
	fprintf(fout, "\n Volume   : %lx\n", ss.mode_vol & 0x0f);
}

void dump_sid_waveform(UBYTE wave)
{
	if (wave & 0xf0) {
		if (wave & 0x10) fprintf(fout, "Triangle ");
		if (wave & 0x20) fprintf(fout, "Sawtooth ");
		if (wave & 0x40) fprintf(fout, "Rectangle ");
		if (wave & 0x80) fprintf(fout, "Noise");
	} else
		fprintf(fout, "None");
	fputc('\n', fout);
}

void view_vic_state(void)
{
	MOS6569State vs;
	short i;

	TheVIC->GetState(&vs);

	fprintf(fout, "Raster line       : %04lx\n", vs.raster | ((vs.ctrl1 & 0x80) << 1));
	fprintf(fout, "IRQ raster line   : %04lx\n\n", vs.irq_raster);

	fprintf(fout, "X scroll          : %ld\n", vs.ctrl2 & 7);
	fprintf(fout, "Y scroll          : %ld\n", vs.ctrl1 & 7);
	fprintf(fout, "Horizontal border : %ld columns\n", vs.ctrl2 & 8 ? 40 : 38);
	fprintf(fout, "Vertical border   : %ld rows\n\n", vs.ctrl1 & 8 ? 25 : 24);

	fprintf(fout, "Display mode      : ");
	switch (((vs.ctrl1 >> 4) & 6) | ((vs.ctrl2 >> 4) & 1)) {
		case 0:
			fprintf(fout, "Standard text\n");
			break;
		case 1:
			fprintf(fout, "Multicolor text\n");
			break;
		case 2:
			fprintf(fout, "Standard bitmap\n");
			break;
		case 3:
			fprintf(fout, "Multicolor bitmap\n");
			break;
		case 4:
			fprintf(fout, "ECM text\n");
			break;
		case 5:
			fprintf(fout, "Invalid text (ECM+MCM)\n");
			break;
		case 6:
			fprintf(fout, "Invalid bitmap (ECM+BMM)\n");
			break;
		case 7:
			fprintf(fout, "Invalid bitmap (ECM+BMM+ECM)\n");
			break;
	}
	fprintf(fout, "Sequencer state   : %s\n", vs.display_state ? "Display" : "Idle");
	fprintf(fout, "Bad line state    : %s\n", vs.bad_line ? "Yes" : "No");
	fprintf(fout, "Bad lines enabled : %s\n", vs.bad_line_enable ? "Yes" : "No");
	fprintf(fout, "Video counter     : %04lx\n", vs.vc);
	fprintf(fout, "Video counter base: %04lx\n", vs.vcbase);
	fprintf(fout, "Row counter       : %ld\n\n", vs.rc);

	fprintf(fout, "VIC bank          : %04lx-%04lx\n", vs.bank_base, vs.bank_base + 0x3fff);
	fprintf(fout, "Video matrix base : %04lx\n", vs.matrix_base);
	fprintf(fout, "Character base    : %04lx\n", vs.char_base);
	fprintf(fout, "Bitmap base       : %04lx\n\n", vs.bitmap_base);

	fprintf(fout, "         Spr.0  Spr.1  Spr.2  Spr.3  Spr.4  Spr.5  Spr.6  Spr.7\n");
	fprintf(fout, "Enabled: "); dump_spr_flags(vs.me);
	fprintf(fout, "Data   : %04lx   %04lx   %04lx   %04lx   %04lx   %04lx   %04lx   %04lx\n",
		vs.sprite_base[0], vs.sprite_base[1], vs.sprite_base[2], vs.sprite_base[3],
		vs.sprite_base[4], vs.sprite_base[5], vs.sprite_base[6], vs.sprite_base[7]);
	fprintf(fout, "MC     : %02lx     %02lx     %02lx     %02lx     %02lx     %02lx     %02lx     %02lx\n",
		vs.mc[0], vs.mc[1], vs.mc[2], vs.mc[3], vs.mc[4], vs.mc[5], vs.mc[6], vs.mc[7]);

	fprintf(fout, "Mode   : ");
	for (i=0; i<8; i++)
		if (vs.mmc & (1<<i))
			fprintf(fout, "Multi  ");
		else
			fprintf(fout, "Std.   ");

	fprintf(fout, "\nX-Exp. : "); dump_spr_flags(vs.mxe);
	fprintf(fout, "Y-Exp. : "); dump_spr_flags(vs.mye);

	fprintf(fout, "Prio.  : ");
	for (i=0; i<8; i++)
		if (vs.mdp & (1<<i))
			fprintf(fout, "Back   ");
		else
			fprintf(fout, "Fore   ");

	fprintf(fout, "\n\nPending interrupts: ");
	dump_vic_ints(vs.irq_flag);
	fprintf(fout, "Enabled interrupts: ");
	dump_vic_ints(vs.irq_mask);
}

void dump_spr_flags(UBYTE f)
{
	short i;

	for (i=0; i<8; i++)
		if (f & (1<<i))
			fprintf(fout, "Yes    ");
		else
			fprintf(fout, "No     ");

	fputc('\n', fout);
}

void dump_vic_ints(UBYTE i)
{
	if (i & 0x1f) {
		if (i & 1) fprintf(fout, "Raster ");
		if (i & 2) fprintf(fout, "Spr-Data ");
		if (i & 4) fprintf(fout, "Spr-Spr ");
		if (i & 8) fprintf(fout, "Lightpen");
	} else
		fprintf(fout, "None");
	fputc('\n', fout);
}

void view_1541_state(void)
{
	fprintf(fout, "Pending 1541 jobs:\n");
	for (int i=0; i<5; i++) {
		fprintf(fout, " Buffer %04x: Job Code %02x, Track %d, Sector %d\n",
			(i+3) << 8, TheCPU1541->RAM[i], TheCPU1541->RAM[i*2+6], TheCPU1541->RAM[i*2+7]);
	}
}


/*
 *  Load data
 *  l start "file"
 */

void load_data(void)
{
	UWORD adr;
	FILE *file;
	int fc;

	if (!expression(&adr))
		return;
	if (the_token == T_END) {
		error("Missing file name");
		return;
	}
	if (the_token != T_STRING) {
		error("'\"' around file name expected");
		return;
	}

	if (!(file = fopen(the_string, "rb")))
		error("Unable to open file");
	else {
		while ((fc = fgetc(file)) != EOF)
			SAMWriteByte(adr++, fc);
		fclose(file);
	}
}


/*
 *  Save data
 *  s start end "file"
 */

void save_data(void)
{
	CBOOL done = FALSE;
	UWORD adr, end_adr;
	FILE *file;

	if (!expression(&adr))
		return;
	if (!expression(&end_adr))
		return;
	if (the_token == T_END) {
		error("Missing file name");
		return;
	}
	if (the_token != T_STRING) {
		error("'\"' around file name expected");
		return;
	}

	if (!(file = fopen(the_string, "wb")))
		error("Unable to create file");
	else {
		do {
			if (adr == end_adr) done = TRUE;

			fputc(SAMReadByte(adr++), file);
		} while (!done);
		fclose(file);
	}
}
