/*
 *  SID_Be.i - 6581 emulation, Be specific stuff
 *
 *  Frodo (C) 1996 Christian Bauer
 *

 *
 * Incompatibilities:
 * ------------------
 *
 *  - No filters
 *  - The test bits don't work the way they should
 *  - Envelopes are only calculated once every 1/60 seconds. This
 *    should be done more often.
 *  - Combined waveforms don't respect pulse width
 */

#include <math.h>
#include <MediaKit.h>


/**
 **  Renderer for digital SID emulation (SIDTYPE_DIGITAL)
 **/

const ULONG SAMPLE_FREQ = 44100;	// Sample output frequency in Hz
const ULONG SID_FREQ = 985248;		// SID frequency in Hz
const ULONG STREAM_FREQ = 60;		// Frequency at which stream_func is called in Hz
const ULONG SID_CYCLES = SID_FREQ/SAMPLE_FREQ;	// # of SID clocks per sample frame

// SID waveforms (some of them :-)
enum {
	WAVE_NONE,
	WAVE_TRI,
	WAVE_SAW,
	WAVE_TRISAW,
	WAVE_RECT,
	WAVE_TRIRECT,
	WAVE_SAWRECT,
	WAVE_TRISAWRECT,
	WAVE_NOISE
};

// EG states
enum {
	EG_IDLE,
	EG_ATTACK,
	EG_DECAY,
	EG_SUSTAIN,
	EG_RELEASE
};

// Structure for one voice
struct DRVoice {
	int wave;		// Selected waveform
	int eg_state;	// Current state of EG
	DRVoice *mod_by;	// Voice that modulates this one
	DRVoice *mod_to;	// Voice that is modulated by this one

	ULONG count;	// Counter for waveform generator
	ULONG add;		// Added to counter in every frame

	UWORD freq;		// SID frequency value
	UWORD pw;		// SID pulse-width value

	UWORD a_add;	// EG parameters
	UWORD d_sub;
	UWORD s_level;	// (linear sustain level)
	UWORD r_sub;
	ULONG eg_level;	// Linear EG level

	ULONG noise;	// Last noise generator output value

	CBOOL gate;		// EG gate bit
	CBOOL ring;		// Ring modulation bit
	CBOOL test;		// Test bit

					// The following bit is set for the modulating
					// voice, not for the modulated one (as the SID bits)
	CBOOL sync;		// Sync modulation bit
};

// Renderer class
class DigitalRenderer : public SIDRenderer {
public:
	DigitalRenderer();
	virtual ~DigitalRenderer();

	virtual void Reset(void);
	virtual void WriteRegister(UWORD adr, UBYTE byte);
	virtual void Pause(void);
	virtual void Resume(void);

private:
	static bool stream_invoc(void *arg, char *buf, long count);
	bool stream_func(UBYTE *buf, long count);

	BAudioSubscriber *the_sub;		// Pointer to subscriber
	CBOOL ready;					// Flag: Subscriber has subscribed
	CBOOL in_stream;				// Flag: Subscriber has entered stream
	UBYTE volume;					// Master volume
	CBOOL v3_mute;					// Voice 3 muted

	static UWORD TriTable[0x200];	// Tables for certain waveforms
	static UWORD TriSawTable[0x100];
	static UWORD TriRectTable[0x100];
	static UWORD SawRectTable[0x100];
	static UWORD TriSawRectTable[0x100];
	static UBYTE MixTable[0x300];	// Table for mixing the three voices
	static ULONG ATable[16];		// Increment/decrement values for all A/D/R settings
	static ULONG DTable[16];
	static UWORD EGLin2Exp[0x100];	// For conversion of linear<->exponential EG level
	static UWORD EGExp2Lin[0x100];
	static const ULONG ARate[16];	// Attack times in milliseconds
	static const ULONG DRate[16];	// Decay/release times in milliseconds

	DRVoice voice[3];				// Data for 3 voices
};

// Static data members
UWORD DigitalRenderer::TriTable[0x200];
UWORD DigitalRenderer::TriSawTable[0x100] = {
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0808,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x1818, 0x3C3C,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x1C1C,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x8080, 0x0000, 0x8080, 0x8080,
	0xC0C0, 0xC0C0, 0xC0C0, 0xC0C0, 0xC0C0, 0xC0C0, 0xC0C0, 0xE0E0,
	0xF0F0, 0xF0F0, 0xF0F0, 0xF0F0, 0xF8F8, 0xF8F8, 0xFCFC, 0xFEFE
};
UWORD DigitalRenderer::TriRectTable[0x100] = {
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0xFFFF, 0xFCFC, 0xF8F8, 0xF0F0, 0xF4F4, 0xF0F0, 0xF0F0, 0xE0E0,
	0xECEC, 0xE0E0, 0xE0E0, 0xC0C0, 0xE0E0, 0xC0C0, 0xC0C0, 0xC0C0,
	0xDCDC, 0xC0C0, 0xC0C0, 0xC0C0, 0xC0C0, 0xC0C0, 0x8080, 0x8080,
	0xC0C0, 0x8080, 0x8080, 0x8080, 0x8080, 0x8080, 0x0000, 0x0000,
	0xBEBE, 0xA0A0, 0x8080, 0x8080, 0x8080, 0x8080, 0x8080, 0x0000,
	0x8080, 0x8080, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x8080, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x7E7E, 0x7070, 0x6060, 0x0000, 0x4040, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
};
UWORD DigitalRenderer::SawRectTable[0x100] = {
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x8080,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x8080, 0x8080,
	0x0000, 0x8080, 0x8080, 0x8080, 0x8080, 0x8080, 0xB0B0, 0xBEBE,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x8080,
	0x0000, 0x0000, 0x0000, 0x8080, 0x8080, 0x8080, 0x8080, 0xC0C0,
	0x0000, 0x8080, 0x8080, 0x8080, 0x8080, 0x8080, 0x8080, 0xC0C0,
	0x8080, 0x8080, 0xC0C0, 0xC0C0, 0xC0C0, 0xC0C0, 0xC0C0, 0xDCDC,
	0x8080, 0x8080, 0x8080, 0xC0C0, 0xC0C0, 0xC0C0, 0xC0C0, 0xC0C0,
	0xC0C0, 0xC0C0, 0xC0C0, 0xE0E0, 0xE0E0, 0xE0E0, 0xE0E0, 0xECEC,
	0xC0C0, 0xE0E0, 0xE0E0, 0xE0E0, 0xE0E0, 0xF0F0, 0xF0F0, 0xF4F4,
	0xF0F0, 0xF0F0, 0xF8F8, 0xF8F8, 0xF8F8, 0xFCFC, 0xFEFE, 0xFFFF
};
UWORD DigitalRenderer::TriSawRectTable[0x100] = {
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x8080, 0x8080,
	0x8080, 0x8080, 0x8080, 0x8080, 0x8080, 0x8080, 0xC0C0, 0xC0C0,
	0xC0C0, 0xC0C0, 0xE0E0, 0xE0E0, 0xE0E0, 0xF0F0, 0xF8F8, 0xFCFC
};
UBYTE DigitalRenderer::MixTable[0x300];
ULONG DigitalRenderer::ATable[16];
ULONG DigitalRenderer::DTable[16];
UWORD DigitalRenderer::EGLin2Exp[0x100];
UWORD DigitalRenderer::EGExp2Lin[0x100];
const ULONG DigitalRenderer::ARate[16] = {
	2, 8, 16, 24, 38, 56, 68, 80, 100, 250, 500, 800, 1000, 3000, 5000, 8000
};
const ULONG DigitalRenderer::DRate[16] = {
	6, 24, 48, 72, 114, 168, 204, 240, 300, 750, 1500, 2400, 3000, 9000, 15000, 24000
};


/*
 *  Constructor, open subscriber
 */

DigitalRenderer::DigitalRenderer()
{
	int i;

	voice[0].mod_by = &voice[2];
	voice[1].mod_by = &voice[0];
	voice[2].mod_by = &voice[1];
	voice[0].mod_to = &voice[1];
	voice[1].mod_to = &voice[2];
	voice[2].mod_to = &voice[0];

	in_stream = FALSE;

	the_sub = new BAudioSubscriber("Frodo SID emulation");
	ready = the_sub->Subscribe(B_DAC_STREAM, the_sub->ID(), FALSE) == B_NO_ERROR;
	if (ready) {
		the_sub->SetDACSampleInfo(1, 1, B_BIG_ENDIAN, B_LINEAR_SAMPLES);
		the_sub->SetSamplingRate(SAMPLE_FREQ);
		the_sub->EnterStream(NULL, TRUE, this, stream_invoc, NULL, TRUE);
		the_sub->SetStreamBuffers(SAMPLE_FREQ / STREAM_FREQ, 4);	// Must be called after EnterStream()
		in_stream = TRUE;
	}

	Reset();

	// Calculate triangle table
	for (i=0; i<0x100; i++) {
		TriTable[i] = (i << 8) | i;
		TriTable[0x1ff-i] = (i << 8) | i;
	}

	// Calculate mixing table (3 voices unsigned -> 1 voice signed)
	for (i=0; i<0x300; i++)
		MixTable[i] = (i / 3) ^ 0x80;

	// Calculate EG increments/decrements table
	for (i=0; i<16; i++) {
		ATable[i] = 0xffff * 1000 / STREAM_FREQ / (float)ARate[i];
		if (ATable[i] > 0xffff)
			ATable[i] = 0xffff;
		DTable[i] = 0xffff * 1000 / STREAM_FREQ / (float)DRate[i];
		if (DTable[i] > 0xffff)
			DTable[i] = 0xffff;
	}

	// Calculate linear<->exponential table
	const float gamma=2;
	for (i=0; i<0x100; i++) {
		EGLin2Exp[i] = pow((float)i / 0x100, gamma) * 0xffff;
		EGExp2Lin[i] = pow((float)i / 0x100, 1/gamma) * 0xffff;
	}
}


/*
 *  Destructor
 */

DigitalRenderer::~DigitalRenderer()
{
	if (ready) {
		if (in_stream) {
			the_sub->ExitStream();
			in_stream = FALSE;
		}
		the_sub->SetStreamBuffers(B_PAGE_SIZE, 8);
		the_sub->Unsubscribe();
		ready = FALSE;
	}
	delete the_sub;
}


/*
 *  Reset emulation
 */

void DigitalRenderer::Reset(void)
{
	volume = 0;
	v3_mute = FALSE;

	for (int v=0; v<3; v++) {
		voice[v].wave = WAVE_NONE;
		voice[v].eg_state = EG_IDLE;
		voice[v].count = voice[v].add = voice[v].freq = voice[v].pw = 0;
		voice[v].eg_level = voice[v].a_add = voice[v].d_sub = voice[v].r_sub = 0;
		voice[v].gate = voice[v].ring = voice[v].test = FALSE;
	}
}


/*
 *  Write to register
 */

void DigitalRenderer::WriteRegister(UWORD adr, UBYTE byte)
{
	int v = adr/7;	// Voice number

	if (!ready)
		return;

	switch (adr) {
		case 0:
		case 7:
		case 14:
			voice[v].freq = (voice[v].freq & 0xff00) | byte;
			voice[v].add = (float)voice[v].freq * SID_FREQ / SAMPLE_FREQ;
			break;

		case 1:
		case 8:
		case 15:
			voice[v].freq = (voice[v].freq & 0xff) | (byte << 8);
			voice[v].add = (float)voice[v].freq * SID_FREQ / SAMPLE_FREQ;
			break;

		case 2:
		case 9:
		case 16:
			voice[v].pw = (voice[v].pw & 0x0f00) | byte;
			break;

		case 3:
		case 10:
		case 17:
			voice[v].pw = (voice[v].pw & 0xff) | ((byte & 0xf) << 8);
			break;

		case 4:
		case 11:
		case 18:
			voice[v].wave = (byte >> 4) & 0xf;
			if ((byte & 1) != voice[v].gate)
				if (byte & 1)	// Gate turned on
					voice[v].eg_state = EG_ATTACK;
				else			// Gate turned off
					if (voice[v].eg_state != EG_IDLE)
						voice[v].eg_state = EG_RELEASE;
			voice[v].gate = byte & 1;
			voice[v].ring = byte & 4;
			voice[v].test = byte & 8;
			voice[v].mod_by->sync = byte & 2;
			break;

		case 5:
		case 12:
		case 19:
			voice[v].a_add = ATable[byte >> 4];
			voice[v].d_sub = DTable[byte & 0xf];
			break;

		case 6:
		case 13:
		case 20:
			int s = byte >> 4;
			voice[v].s_level = EGExp2Lin[(s << 4) | s];
			voice[v].r_sub = DTable[byte & 0xf];
			break;

		case 24:
			volume = byte & 0xf;
			v3_mute = byte & 0x80;
			break;
	}
}


/*
 *  Pause sound output
 */

void DigitalRenderer::Pause(void)
{
	if (in_stream) {
		the_sub->ExitStream();
		in_stream = FALSE;
	}
}


/*
 *  Resume sound output
 */

void DigitalRenderer::Resume(void)
{
	if (!in_stream) {
		the_sub->EnterStream(NULL, TRUE, this, stream_invoc, NULL, TRUE);
		in_stream = TRUE;
	}
}


/*
 *  Stream function 
 */

bool DigitalRenderer::stream_invoc(void *arg, char *buf, long count)
{
	return ((DigitalRenderer *)arg)->stream_func((UBYTE *)buf, count);
}


bool DigitalRenderer::stream_func(UBYTE *buf, long count)
{
	UWORD output[3];
	ULONG eg_level[3];
	DRVoice *v;

	// Envelope generator
	for (int i=0; i<3; i++) {
		v = &voice[i];

		switch (v->eg_state) {
			case EG_ATTACK:
				v->eg_level += v->a_add;
				if (v->eg_level > 0xffff) {
					v->eg_level = 0xffff;
					v->eg_state = EG_DECAY;
				}
				break;
			case EG_DECAY:
				v->eg_level -= v->d_sub;
				if (v->eg_level < v->s_level || v->eg_level > 0xffff) {
					v->eg_level = v->s_level;
					v->eg_state = EG_SUSTAIN;
				}
				break;
			case EG_SUSTAIN:
				v->eg_level = v->s_level;
				break;
			case EG_RELEASE:
				v->eg_level -= v->r_sub;
				if (v->eg_level > 0xffff) {
					v->eg_level = 0;
					v->eg_state = EG_IDLE;
				}
				break;
			case EG_IDLE:
				v->eg_level = 0;
		}
		eg_level[i] = (EGLin2Exp[(v->eg_level >> 8) & 0xff] * volume) >> 12;
	}

	if (v3_mute)
		eg_level[2] = 0;

	while (count--) {
		for (int j=0; j<3; j++) {
			v = &voice[j];

			// Waveform generator
			v->count += v->add;

			if (v->sync && (v->count > 0x1000000))
				v->mod_to->count = 0;

			v->count &= 0xffffff;

			switch (v->wave) {
				case WAVE_TRI:
					if (v->ring)
						output[j] = TriTable[(v->count ^ (v->mod_by->count & 0x800000)) >> 15];
					else
						output[j] = TriTable[v->count >> 15];
					break;
				case WAVE_SAW:
					output[j] = v->count >> 8;
					break;
				case WAVE_RECT:
					if (v->count > (v->pw << 12))
						output[j] = 0xffff;
					else
						output[j] = 0;
					break;
				case WAVE_TRISAW:
					output[j] = TriSawTable[v->count >> 16];
					break;
				case WAVE_TRIRECT:
					output[j] = TriRectTable[v->count >> 16];
					break;
				case WAVE_SAWRECT:
					output[j] = SawRectTable[v->count >> 16];
					break;
				case WAVE_TRISAWRECT:
					output[j] = TriSawRectTable[v->count >> 16];
					break;
				case WAVE_NOISE:
					if (v->count > 0x100000) {
						output[j] = v->noise = (rand() & 0xff) << 8;
						v->count &= 0xfffff;
					} else
						output[j] = v->noise;
					break;
				default:
					output[j] = 0x8000;
			}
		}

		*buf++ += MixTable[(output[0] * eg_level[0] + output[1] * eg_level[1] + output[2] * eg_level[2]) >> 16];
	}
	return TRUE;
}
