/*
 *  main_Be.i - Main program, BeOS specific stuff
 *
 *  Frodo (C) 1996 Christian Bauer
 */

#include <AppKit.h>
#include <StorageKit.h>
#include <InterfaceKit.h>

// Constants
const BRect AboutFrame = BRect(0, 0, 383, 99);

// Global variables
CBOOL FromShell;
BBitmap *AboutBitmap;


/*
 *  Create application object and start it
 */

int main(int argc, char **argv)
{	
	Frodo *the_app;

	srand(system_time());
	FromShell = (argc != 0);	// !! This doesn't work...

	the_app = new Frodo();
	the_app->Run();
	delete the_app;

	return 0;
}


/*
 *  Constructor: Initialize member variables
 */

Frodo::Frodo() : BApplication(APP_SIGNATURE)
{
	TheC64 = NULL;
	AboutBitmap = NULL;
	strcpy(prefs_path, "Frodo Prefs");
	prefs_showing = FALSE;
}


/*
 *  Process command line arguments
 */

void Frodo::ArgvReceived(int argc, char **argv)
{
	if (argc == 2)
		strcpy(prefs_path, argv[1]);
}


/*
 *  Process Browser arguments
 */

void Frodo::RefsReceived(BMessage *message)
{
	// If preferences editor is on screen, forward message to all windows and
	//  hope that the preferences editor gets it
	if (prefs_showing) {
		int i=0;
		BWindow *window;
		while (window = WindowAt(i++)) {
			BMessage *msg = new BMessage(message);
			window->PostMessage(msg);
		}

	// Preferences editor not open, set preferences path unless C64 is running
	} else if (!TheC64) {
		BFile *file = new BFile(message->FindRef("refs"));
		file->GetPath(prefs_path, 255);
		delete file;
	}
}


/*
 *  Arguments processed, prepare emulation and show preferences editor window
 */

void Frodo::ReadyToRun(void)
{
	// Find application directory
	app_info the_info;
	BDirectory *dir = new BDirectory;
	BFile *file = new BFile;
	if (GetAppInfo(&the_info) == B_NO_ERROR) {
		file->SetRef(the_info.ref);
		file->GetParent(dir);
	} else {
		ShowRequester("Can't find application directory.", "Quit");
		PostMessage(B_QUIT_REQUESTED);
		return;
	}
	delete file;

	// Set cwd to application directory
	dir->GetPath(AppDirPath, sizeof(AppDirPath));
	chdir(AppDirPath);
	delete dir;

	// Set up "about" window bitmap
	AboutBitmap = new BBitmap(AboutFrame, B_COLOR_8_BIT);
	FILE *logofile = fopen("Frodo Logo", "rb");
	if (logofile != NULL) {
		fread(AboutBitmap->Bits(), 384*100, 1, logofile);
		fclose(logofile);
	}

	// Create application main menu
	BPopUpMenu *menu = new BPopUpMenu("", FALSE, FALSE);
	BMenuItem *item;
	item = new BMenuItem("About Frodo...", new BMessage(B_ABOUT_REQUESTED));
	menu->AddItem(item);
	item = new BMenuItem("Reset C64", new BMessage(MSG_RESET));
	menu->AddItem(item);
	item = new BMenuItem("Preferences...", new BMessage(MSG_PREFS), 'P');
	menu->AddItem(item);
	item = new BMenuItem("SAM...", new BMessage(MSG_SAM), 'M');
	menu->AddItem(item);
	item = new BMenuItem("Quit Frodo", new BMessage(B_QUIT_REQUESTED), 'Q');
	menu->AddItem(item);
	menu->SetTargetForItems(be_app);
	SetMainMenu(menu);

	// Load preferences
	ThePrefs.Load(prefs_path);

	// Show preferences editor (sends MSG_STARTUP on close)
	prefs_showing = TRUE;
	ThePrefs.ShowEditor(TRUE, prefs_path);
}


/*
 *  Handle incoming messages
 */

void Frodo::MessageReceived(BMessage *msg)
{
	switch (msg->what) {
		case MSG_STARTUP: // Start the emulation

			// Preferences editor is not longer on screen
			prefs_showing = FALSE;

			// Create everything
			TheC64 = new C64;

			// Load ROMs
			if (!load_rom_files()) {
				PostMessage(B_QUIT_REQUESTED);
				return;
			}

			// Run the 6510
			TheC64->Run();
			break;

		case MSG_PREFS:	// Show preferences editor
			if (TheC64 && !prefs_showing) {
				TheC64->Pause();
				Prefs *prefs = new Prefs(ThePrefs);
				prefs_showing = TRUE;
				prefs->ShowEditor(FALSE, prefs_path);	// Sends MSG_PREFS_DONE on close
			}
			break;

		case MSG_PREFS_DONE: // Preferences editor closed
			Prefs *prefs = (Prefs *)(msg->FindObject("prefs"));
			if (!msg->FindBool("canceled")) {
				TheC64->NewPrefs(prefs);
				ThePrefs = *prefs;
			}
			delete prefs;
			prefs_showing = FALSE;
			TheC64->Resume();
			break;

		case MSG_RESET:	// Reset C64
			if (TheC64)
				TheC64->Reset();
			break;

		case MSG_NMI:	// NMI
			if (TheC64)
				TheC64->NMI();
			break;

		case MSG_SAM:	// Invoke SAM
			if (TheC64 && !prefs_showing && FromShell) {
				TheC64->Pause();
				SAM(TheC64);
				TheC64->Resume();
			}
			break;

		default:
			inherited::MessageReceived(msg);
	}
}


/*
 *  Quit requested (either by menu or by closing the C64 display window)
 */

bool Frodo::QuitRequested(void)
{
	if (BApplication::QuitRequested()) {

		// Stop emulation
		if (TheC64) {
			TheC64->Quit();
			delete TheC64;
		}

		if (AboutBitmap)
			delete AboutBitmap;

		return TRUE;
	}
	return FALSE;
}


/*
 *  Set main menu enable/disable states
 */

void Frodo::MenusWillShow(void)
{
	BPopUpMenu *menu = MainMenu();
	menu->FindItem(MSG_RESET)->SetEnabled(TheC64 != NULL);
	menu->FindItem(MSG_PREFS)->SetEnabled(!prefs_showing);
	menu->FindItem(MSG_SAM)->SetEnabled((TheC64 != NULL) && !prefs_showing && FromShell);
}


/*
 *  Display "about" window
 */

class AboutWindow : public BWindow {
public:
	AboutWindow();
};

class AboutView : public BView {
public:
	AboutView();
	virtual void AttachedToWindow(void);
	virtual void Draw(BRect update);
	virtual void MouseDown(BPoint point);
};

AboutWindow::AboutWindow() : BWindow(AboutFrame, NULL, B_BORDERED_WINDOW, B_NOT_RESIZABLE | B_NOT_ZOOMABLE | B_WILL_ACCEPT_FIRST_CLICK)
{
	AboutView *view;

	Lock();
	MoveTo(100, 100);
	view = new AboutView;
	AddChild(view);
	view->MakeFocus();
	Unlock();
	Show();
}

AboutView::AboutView() : BView(AboutFrame, "", B_FOLLOW_NONE, B_WILL_DRAW) {}

void AboutView::AttachedToWindow(void)
{
	SetHighColor(0, 0, 0);
}

void AboutView::Draw(BRect update)
{
	DrawBitmap(AboutBitmap, update, update);

	SetFontName("Emily");
	SetFontSize(12);
	SetDrawingMode(B_OP_OVER);
	MovePenTo(215, 20);
	DrawString(VERSION_STRING);

	SetFontName("Erich");
	SetFontSize(9);
	MovePenTo(215, 40);
	DrawString("by Christian Bauer");
	MovePenTo(215, 52);
	DrawString("<cbauer@iphcip1.physik.uni-mainz.de>");
	MovePenTo(215, 75);
	DrawString(" Copyright 1994-1996");
	MovePenTo(215, 87);
	DrawString("Freely distributable.");
}

void AboutView::MouseDown(BPoint point)
{
	Window()->PostMessage(B_QUIT_REQUESTED);
}

void Frodo::AboutRequested(void)
{
	AboutWindow *win = new AboutWindow;
}
